(******************************************************************************

  ProtoVeriPhy

  Module: distance/SystemInvariants.thy

  Copyright (C) 2007-2012 Benedikt Schmidt <benedikt.schmidt@inf.ethz.ch>
  License: BSD (see COPYRIGHT)

******************************************************************************)

header {* Protocol-independent Invariants of the System *}

theory SystemInvariants imports System begin

subsection {* Some Simple Lemmas *}

lemma createEv_no_Recv [simp,intro]: "Recv A m \<noteq> createEv fid pev m'"
  apply (cases pev)
by (auto simp add: createEv.psimps)

text {* These hold for all protocols *}

text {* prefix closed *}

lemma (in PROTOCOL) prefix_closed_sys_H:
  "\<lbrakk> (a#as) \<in> sys \<rbrakk> \<Longrightarrow> tl (a#as) \<in> sys"
  apply (rule_tac x="(a#as)" in sys.induct)
  apply (auto)
done

lemma (in PROTOCOL) prefix_closed_sys: "\<lbrakk> (a#as) \<in> sys \<rbrakk> \<Longrightarrow> as \<in> sys"
  apply (subst tl.simps [THEN sym])
  apply (rule prefix_closed_sys_H)
by (auto)

text {* time in traces increases (not strictly) monotonically *}

lemma (in PROTOCOL) tracetime_non_negative:
  assumes A: "tr \<in> sys" and B: "(t,ev) \<in> set tr"
  shows   "0 \<le> t" using A B
proof (induct tr  arbitrary: t ev rule: sys.induct)
  case Nil thus ?case by auto
next
  case (Con tr trecv M B j tab t ev)
  have mt: "maxtime tr \<ge> 0" by auto
  show ?case
  proof cases
    assume "(t,ev) = (trecv, Recv (Rx B j) M)"
    hence teq: "trecv = t" by auto
    thus ?case using prems(4-)
      apply (case_tac tr)
      apply auto
      apply (subgoal_tac "0 \<le> a")
      apply auto
      done
  next
    assume "(t,ev) \<noteq> (trecv, Recv (Rx B j) M)"
    with prems have "(t,ev) \<in> set tr" by auto
    thus ?case using Con.hyps by auto
  qed
next
  case (Fake tr ts X I j)
  have mt: "maxtime tr \<ge> 0" by auto
  show ?case
  proof cases
    assume P: "\<exists>I j L. (t,ev) = (ts, Send (Tx (Intruder I) j) X L)"
    with P obtain I j L where "(t,ev) = (ts, Send (Tx (Intruder I) j) X L)"
      by auto
    hence teq: "ts = t" by auto
    with prems mt show ?case by arith
  next
    assume "\<not> (\<exists>I j L. (t,ev) = (ts, Send (Tx (Intruder I) j) X L))"
    with prems have "(t,ev) \<in> set tr" by auto
    thus ?case using Fake.hyps by auto
  qed
next
  case (Proto tr t' step m pEv A')
  have mt: "maxtime tr \<ge> 0" by auto
  show ?case
  proof cases
    assume "(t,ev) = (t', createEv A' pEv m)"
    hence teq: "t' = t" by auto
    with prems mt show ?case by arith
  next
    assume "(t,ev) \<noteq> (t', createEv A' pEv m)"
    with prems have "(t,ev) \<in> set tr" by auto
    thus ?case using Proto.hyps by auto
  qed
qed

lemma (in PROTOCOL) tracetime_increases:
  assumes A: "tr \<in> sys" and B: "tr=(t,ev)#trtl"
  shows   "t \<ge> maxtime trtl" using A B
proof (induct tr arbitrary: t ev trtl rule: sys.induct)
  case Nil thus ?case by auto
next
  case (Con tr trecv)
  hence "t=trecv" and "tr=trtl" by auto
  with prems show ?case by auto
next
  case (Fake tr ts X I j)
  hence "t=ts" and "tr=trtl" by auto
  with prems show ?case by auto
next
  case (Proto tr t' step m pEv A')
  hence "t=t'" and "tr=trtl" by auto
  with prems show ?case by auto
qed

lemma (in PROTOCOL) maxtime_cons:
  "c \<le> maxtime (tr) ==> c \<le> maxtime (ev # tr)"
  apply auto
done

text {* a suffix of a trace removing all events after a certain event is still a valid trace *}

lemma (in PROTOCOL) proto_before_event:
  "\<lbrakk> tr \<in> sys; e \<in> set tr \<rbrakk> \<Longrightarrow> (beforeEvent e tr) \<in> sys"
apply (induct tr)
apply (force)
apply (rule_tac P="%tr. (beforeEvent e tr) \<in> sys" in sys.induct)
by (auto simp add: beforeEvent.simps)


lemma (in PROTOCOL) not_beforeEvent_later:
  assumes A: "(ta, eva) \<notin> set (beforeEvent (tb, evb) tr)" and
          B: "(ta, eva) \<in> set tr" and C: "(tb, evb) \<in> set tr" and p: "tr \<in> sys"
  shows "tb \<le> ta" using A B C p
proof (induct tr rule: trace_induct)
  case 1
  from prems show ?thesis by (force)
next
  case (2 t ev xs)
  show ?thesis
  proof cases
    assume "(t,ev)=(tb,evb) \<and> (tb,evb) \<notin> set xs"
    with prems show ?thesis by clarsimp
  next
    assume n: "\<not> ((t,ev)=(tb,evb) \<and> (tb,evb) \<notin> set xs)"
    with prems have "(ta, eva) \<notin> set (beforeEvent (tb, evb) xs)" by auto
    with n `(tb, evb) \<in> set ((t, ev)#xs)`
      have bxs: "(tb,evb) \<in> set xs" by auto
    show ?thesis
    proof cases
      assume "(t,ev)=(ta,eva)"
      with `(t,ev)=(ta,eva)` prems have "((ta,eva)#xs) \<in> sys" by auto
      with bxs have "ta \<ge> maxtime xs" apply - apply (rule tracetime_increases) by auto
      then also have "\<dots> \<ge> tb" using `(tb, evb) \<in> set xs`
        apply - apply (erule maxtime_geq_elem) by auto
      thus ?thesis .
    next
      assume "(t,ev)\<noteq>(ta,eva)"
      with prems have "(ta,eva) \<in> set tr" by auto
      with prems have "xs \<in> sys" by (auto intro: prefix_closed_sys)
      with prems show ?thesis by auto
    qed
  qed
qed

lemma (in PROTOCOL) beforeEvent_earlier:
  assumes "tr \<in> sys" and "ta < tb" and "(tb,b) \<in> set tr" and  "(ta,a) \<in> set tr"
  shows "(ta,a) \<in> set (beforeEvent (tb,b) tr)" using prems
  apply (rotate_tac 1)
  apply (erule contrapos_pp)
  apply (subgoal_tac "ta \<ge> tb")
  apply force
  apply (erule not_beforeEvent_later)
  apply auto
  apply (insert prems, auto)
done

lemma (in PROTOCOL) beforeEvent_cons_event_delayed: (* move to protocol *)
  assumes a: "tr \<in> sys" and
          b: "e \<in> set tr"
  shows "(e#beforeEvent e tr) \<in> sys" using a b
proof (induct tr rule: sys.induct)
  case Nil thus ?case by auto
next
  case (Fake tr ti Y I j)
  show ?case proof cases
    assume "e = (ti, Send (Tx (Intruder I) j) Y [])"
    thus ?case using prems apply auto by (rule sys.Fake, auto)
  next
    assume "e \<noteq> (ti, Send (Tx (Intruder I) j) Y [])" thus ?case using prems by auto
  qed
next
  case (Con tr trecv X D j)
  let ?ev = "(trecv, Recv (Rx D j) X)"
  show ?case proof cases
    assume "e = ?ev" thus ?case using prems apply auto by (rule sys.Con, auto)
  next
    assume "e \<noteq> ?ev" thus ?thesis using prems by auto
  qed
next
  case (Proto tr t step m pEv A)
  let ?ev="(t, createEv A pEv m)"
  show ?case proof cases
    assume "e = ?ev"
    thus ?thesis using prems apply auto by (rule sys.Proto, auto)
  next
    assume "e \<noteq> ?ev" thus ?case using prems by auto
  qed
qed

lemma (in PROTOCOL) beforeEvent_maxtime:
  assumes del: "tr \<in> sys" and
          ev: "(tev,ev) \<in> set tr"
  shows "maxtime (beforeEvent (tev,ev) tr) \<le> tev" using del ev
  apply (induct tr rule: sys.induct)
  apply auto
done

lemma beforeEvent_prefix:
  assumes a: "ev \<in> set (e#beforeEvent e tr)" and
          b: "e \<in> set tr"
  shows "ev \<in> set tr" using a b
  apply (induct tr, auto split: split_if_asm)
done

lemma view_elem_ex:
  "(t,ev) \<in> (set (view A tr)) \<Longrightarrow> \<exists> t'. (t',ev) \<in> (set tr)"
by (auto simp add: view_def split: split_if_asm)

lemma view_elem_at_ex:
  "\<lbrakk>  (t,ev) \<in> set tr; occursAt ev = Honest A \<rbrakk> \<Longrightarrow>
   \<exists> t'. (t',ev) \<in> (set (view A tr))"
apply (induct tr)
by (auto simp add: view_def split: split_if_asm)

definition 
  timetrans :: "[friendid, 'msg trace] => 'msg trace" where
  "timetrans A tr = [(clocktime A t,ev) . (t, ev) \<leftarrow> tr]"

lemma send_a_view_a_u:
  "((t, Send (Tu (Honest A)) m L) \<in> set (view A tr)) \<equiv>
   ((t, Send (Tu (Honest A)) m L) \<in> set (timetrans A tr))"
apply (rule  HOL.eq_reflection)
apply (auto simp add: view_def occursAt.simps timetrans_def split: split_if_asm)
done


lemma recv_a_view_a_u:
  "((t, Recv (Ru (Honest A)) m) \<in> set (view A tr)) \<equiv>
   ((t, Recv (Ru (Honest A)) m) \<in> set (timetrans A tr))"
apply (rule  HOL.eq_reflection)
apply (auto simp add: view_def occursAt.simps timetrans_def split: split_if_asm)
done

lemma send_a_view_a_r:
  "((t, Send (Tr (Honest A)) m L) \<in> set (view A tr)) \<equiv>
   ((t, Send (Tr (Honest A)) m L) \<in> set (timetrans A tr))"
apply (rule  HOL.eq_reflection)
apply (auto simp add: view_def occursAt.simps timetrans_def split: split_if_asm)
done

lemma recv_a_view_a_r:
  "((t, Recv (Rec (Honest A)) m) \<in> set (view A tr)) \<equiv>
     ((t, Recv (Rec (Honest A)) m) \<in> set (timetrans A tr))"
apply (rule  HOL.eq_reflection)
apply (auto simp add: view_def occursAt.simps timetrans_def split: split_if_asm)
done

lemma view_subset_timetrans:
  "set (view A tr) \<subseteq> set (timetrans A tr)"
  apply (auto simp add: timetrans_def view_def)
done

lemma timetrans_snd [simp]:
  "snd`set (timetrans A tr) = snd`set tr"
  apply (auto simp add: timetrans_def view_def)
  apply (rule_tac x="(a,b)" in rev_image_eqI, auto)
done

lemma  trace_weaken:
  "\<exists>tb. (tb,ev) \<in> set tr ==> \<exists>tb. (tb,ev) \<in> set (tev#tr)"
by auto

lemma (in INITSTATE) usedI_timetrans [simp]:
  "usedI (timetrans A tr) = usedI tr"
  apply auto
  apply (rule usedI_mono_snd)
  apply (rule timetrans_snd [THEN equalityD1], auto)
  apply (rule usedI_mono_snd)
  apply (rule timetrans_snd [THEN equalityD2], auto)
done

text {* a receive is always preceded by the corresponding send *}

lemma (in PROTOCOL) send_before_recv [rule_format, intro]:
  assumes rang: "tr \<in> sys" and
          recv: "(tb, Recv RB M) \<in> set tr" and
          comp: "X \<in> components {M}"
  shows "\<exists> A i tsend L M'.
           \<exists> Y \<in> components {M'}.
             (tsend, Send (Tx A i) M' L) \<in> set tr \<and>
             distort X Y \<in> LowHam \<and>
             cdistM (Tx A i) RB \<noteq> None \<and>
             tsend \<le> tb - cdist (Tx A i) RB"
  using rang recv
proof (induct rule: sys.induct)
     case Nil thus ?case by auto
next case Fake thus ?case using prems apply -
  apply auto
  apply (rule_tac x ="A" in exI)
  apply (rule_tac x ="i" in exI)
  apply (rule_tac x ="tsend" in exI)
  apply (rule_tac x ="L" in exI)
  apply (rule_tac x ="M'" in exI)
  by auto
next case (Proto step m t' pEv A' tr) thus ?case using prems
  apply auto
  apply (rule_tac x ="A" in exI)
  apply (rule_tac x ="i" in exI)
  apply (rule_tac x ="tsend" in exI)
  apply (rule_tac x ="L" in exI)
  apply (rule_tac x ="M'" in exI)
  by auto
next
  case (Con tr trecv N B j tab)
  show ?case
  proof cases
    assume "(tb, Recv RB M) = (trecv, Recv (Rx B j) N)"
    hence "RB=Rx B j" "tb = trecv" "M=N" by auto
    thus ?thesis using prems(3-)
      apply -
      apply (erule_tac x=X in ballE) prefer 2
      apply force
      apply (elim exE bexE)
      apply (rule_tac x ="A" in exI)
      apply (rule_tac x ="i" in exI)
      apply (rule_tac x ="tsend" in exI)
      apply (rule_tac x ="L" in exI)
      apply (rule_tac x ="M'" in exI)
      apply auto
      apply (auto  simp add: cdist_def)
      done
  next
    assume "(tb, Recv RB M) \<noteq> (trecv, Recv (Rx B j) N)"
    hence "(tb, Recv RB M) \<in> set tr" using prems by auto
    with Con.hyps(2) show ?thesis by auto
  qed
qed

lemma (in PROTOCOL) send_before_recv_notime [ intro]:
  assumes rang: "tr \<in> sys" and
          recv: "(tb, Recv RB M) \<in> set tr" and
          comp: "X \<in> components {M}"
  shows "\<exists> A i tsend L M'.
           \<exists> Y \<in> components {M'}.
             (tsend, Send (Tx A i) M' L) \<in> set tr \<and> distort X Y \<in> LowHam"
  using rang recv comp
  apply -
  apply (drule send_before_recv)
  apply simp
  apply auto
  apply (intro exI)
  apply auto
done

end