(******************************************************************************

  ProtoVeriPhy

  Module: distance/TimeSyncSPS.thy

  Copyright (C) 2012 Benedikt Schmidt <benedikt.schmidt@inf.ethz.ch>
  License: BSD (see COPYRIGHT)

******************************************************************************)

header {* Security Analysis of the SPS time synchronization protocol *}

theory TimeSyncSPS
imports SystemCoffset SystemOrigination MessageTheoryFree
begin


locale INITSTATE_SYMKEYS_NN = INITSTATE_SYMKEYS + INITSTATE_NONONCE

text {* Here we define locale describing the necessary initial state
for the protocol, i.e., agents have pairwise shared keys and their
knowledge initially does not contain nonces of other agents. *}

definition
  initStateSPS :: "agent \<Rightarrow> msg set" where
  "initStateSPS A == Key`((symKey A)`UNIV)"

interpretation INITSTATE_SYMKEYS_NN Crypt Nonce MPair Hash Number parts subterms DM
               initStateSPS Key
  apply (unfold_locales, auto simp add: initStateSPS_def dest: injective_symKey)
done


text {* The minimal delay it takes for protocol message to be transmitted to one of the
        honest nodes. I.e. how long does it minimally take from the moment the first bit
        has been received until the complete message is received. *}

consts
  min_delay :: real
  bound_delay :: real

specification (min_delay)
  min_delay_pos: "min_delay > (0::real)"
by (rule_tac x=1 in exI, arith)

specification (bound_delay)
  bound_delay_pos: "bound_delay > (0::real)"
  bound_delay_min: "bound_delay \<ge> min_delay"
proof -
  from min_delay_pos obtain x where "min_delay \<le> x" by fast
  from this and min_delay_pos show ?thesis by fast
qed

definition
  offset :: "real \<Rightarrow> real \<Rightarrow> real \<Rightarrow> real \<Rightarrow> real"
 where
  "offset t1 t2 t3 t4 \<equiv> ((t2 - t1) - (t4 - t3))/2"

text {* The protocols we consider below are challenge response
protocols, i.e., Alice sends a message mA at t1 that is received by
Bob at t2, Bob then replies with a message mB sent by Bob at t3
received by Alice at t4. Assuming that it takes d seconds for a
message to travel from Alice to Bob (and similarly d seconds for a
message into the other direction) we have t2 = t1 + d + o, where o
denotes the offset of the clocks. similarly we have t4 = t3 + d -
o. Thus the above expression computes o given t1, t2, t3, and t4.*}

definition
  delay :: "real \<Rightarrow> real \<Rightarrow> real \<Rightarrow> real \<Rightarrow> real"
 where
  "delay t1 t2 t3 t4 \<equiv> ((t2 - t1) + (t4 - t3))/2"

text {* Using again the dependencies between t1 and t2 (and those
relating t3 and t4) it is easy to see that the function delay computes
d.*}


subsection {* Direct and Parameterized Definition and their Equality *}

text {* The protocol has the simple form
A t1\<rightarrow>t2 B: {A,B,NA}
B t3\<rightarrow>t4 B: Hash[symKey AB]{B,A,NA,t2,t3} 
(details about the message length have been omitted).
*}

subsubsection {* Step Functions for Authenticated Ranging *}

definition
  sps1 :: "msg step"
 where
 "sps1 tr A t =
    (UN NA B. {ev. ev = (\<lbrace>Agent (Honest A),Agent B,Nonce (Honest A) NA\<rbrace>,SendEv 0 []) \<and>
                   Nonce (Honest A) NA \<notin> usedI tr})"

definition
  sps2 :: "msg step"
 where
 "sps2 tr B tsend =
    (UN NA A A' trec_start trec_complete.
       {ev. ev = (Hash[Key (symKey (Honest B) A)]
                    \<lbrace>Agent (Honest B), Agent A, Nonce A' NA,
                     Real trec_complete, Real tsend\<rbrace>,SendEv 0 []) \<and>
            (trec_start, Recv (Rx (Honest B) 0) \<lbrace>Agent A, Agent (Honest B), Nonce A' NA\<rbrace>) \<in> set tr \<and>
            trec_complete \<le> tsend \<and>
            trec_complete \<ge> trec_start + min_delay})"

definition
  sps3 :: "msg step"
 where
 "sps3 tr A tclaim =
    (UN NA B t1 t2 t3 t4_start t4_complete.
       {ev. ev = (\<lbrace>Agent B, Real (offset t1 t2 t3 t4_complete)\<rbrace>,ClaimEv) \<and>
            (t4_start, Recv (Rx (Honest A) 0)
                   Hash[Key (symKey B (Honest A))]
                     \<lbrace>Agent B, Agent (Honest A), Nonce (Honest A) NA, Real t2, Real t3\<rbrace>)
            \<in> set tr \<and>
            
            (t1, Send (Tr (Honest A)) \<lbrace>Agent (Honest A),Agent B,Nonce (Honest A) NA\<rbrace> []) \<in> set tr \<and>
            t4_complete \<le> tclaim \<and>
            t4_complete \<ge> t4_start + min_delay \<and>
            (delay t1 t2 t3 t4_complete) \<le> bound_delay })"

definition
  sps_proto :: "msg proto" where
  "sps_proto = {sps1,sps2,sps3}"

lemmas sps_defs = sps_proto_def sps1_def sps2_def sps3_def

text {* The initial state contains no nonces and signatures. However,
        the initial state includes required public and private
        keys and keys that are never sent over the network. *}

locale PROTOCOL_SPS =
  PROTOCOL_SYMKEYS_NOKEYS+PROTOCOL_NONONCE+INITSTATE_SYMKEYS_NN

interpretation PROTOCOL_SPS Crypt Nonce MPair Hash Number parts subterms DM initStateSPS Key sps_proto
  apply (unfold_locales)
  apply (auto simp add: sps_defs messagesProtoTr_def messagesProtoTrHonest_def
              split: event.split split_if dest: parts.fst_set)
done

text {* Agents only look at their own views and all messages are derivable. *}

lemma notin_analz_union_and: "x \<notin> analz (A \<union> B) \<Longrightarrow> (x \<notin> analz A) \<and> (x \<notin> analz B)"
proof (erule contrapos_pp, simp, auto)
  assume " x \<in> analz A" thus "x \<in> analz (A \<union> B)"
    proof (rule rev_subsetD)
      show "analz A \<subseteq> analz (A \<union> B)" by (rule analz_mono, simp)
      qed
 next
  assume "x \<in> analz B" thus "x \<in> analz (A \<union> B)"
     proof (rule rev_subsetD)
        show "analz B \<subseteq> analz (A \<union> B)" by (rule analz_mono, simp)
      qed
qed


interpretation PROTOCOL_EXECUTABLE Crypt Nonce MPair Hash Number parts subterms DM initStateSPS  sps_proto sys Key
  apply (unfold_locales)
  apply (auto simp add: sps_defs DM_analz_synth initStateSPS_def
                        messagesProtoTr_def messagesProtoTrHonest_def
              split: event.split split_if)

  apply (unfold MACM_def)
  apply (auto intro!: synth.Crypt synth.MPair)
  apply (rule synth.Hash)
  apply (auto intro!: synth.Crypt synth.MPair)
  apply (subgoal_tac "Nonce A' NA \<in> analz (knows (Honest A) tr)")
  apply (drule notin_analz_union_and)
  apply force
  apply (drule view_elem_ex, erule exE, drule Recv_imp_knows_A)
  apply (rule analz.Snd)
  apply (rule analz.Snd)
  apply (rule analz.Inj)
  apply force
  apply (subgoal_tac "Nonce A' NA \<in> analz (knows (Honest A) tr)")
  apply (drule notin_analz_union_and)
  apply force
  apply (drule view_elem_ex, erule exE, drule Recv_imp_knows_A)
  apply (rule analz.Snd)
  apply (rule analz.Snd)
  apply (rule analz.Inj)
  apply force
  apply (auto simp add: nonce_view_fresh [simplified sps_proto_def]
                        nonce_view_used [simplified sps_proto_def])
  apply (auto simp add: recv_a_view_a_r send_a_view_a_r)
  apply (rule_tac x=NA in exI)
  apply (rule_tac x=t1 in exI)
  apply (rule_tac x=t2 in exI)
  apply (rule_tac x=t3 in exI)
  apply (rule_tac x=t4_start in exI)
  apply (rule_tac x=t4_complete in exI)
  apply auto
  apply (rule_tac x=NA in exI)
  apply (rule_tac x=t1 in exI)
  apply (rule_tac x=t2 in exI)
  apply (rule_tac x=t3 in exI)
  apply (rule_tac x=t4_start in exI)
  apply (rule_tac x=t4_complete in exI)
  apply auto
done

subsubsection {* Direct Definition for TimeSync SPS as a set of
inductive rules defining the set of valid traces.*}

inductive_set
  SPS :: "(msg trace) set"

 where
   Nil [intro] : "[] \<in> SPS"

 | Fake: 
   "\<lbrakk> tr \<in> SPS; t \<ge> maxtime tr;
      X \<in> DM (Intruder I) (knowsI (Intruder I) tr)\<rbrakk>
   \<Longrightarrow> (t, Send (Tx (Intruder I) j) X [])  # tr \<in> SPS"

 | Con : 
   "\<lbrakk> tr \<in> SPS; trecv \<ge> maxtime tr;
      (tsend, Send (Tx A i) m L) \<in> set tr;
      cdistM (Tx A i) (Rx B j) = Some tab;
      trecv \<ge> tsend + tab \<rbrakk>
    \<Longrightarrow> (trecv, Recv (Rx B j) m) # tr \<in> SPS"

 -- {* Protocol rules for SPS, correspond to step functions *}
 | SPS1:  
   "\<lbrakk> tr \<in> SPS; t >=  maxtime tr ;
      \<not> (usedI tr (Nonce (Honest A) NA)) \<rbrakk> 
        (* cannot use set notation here: infer_arities: bad term (Isabelle Bug?) *)
   \<Longrightarrow> (t, Send (Tx (Honest A) 0) \<lbrace>Agent (Honest A),Agent B,Nonce (Honest A) NA\<rbrace> []) # tr \<in> SPS"

 | SPS2:
   "\<lbrakk> tr \<in> SPS;  tsend \<ge>  maxtime tr;
      trec_complete \<ge> trec_start + min_delay;
      tsend \<ge> trec_complete;
      (trec_start, Recv (Rx (Honest B) 0) \<lbrace>Agent A, Agent (Honest B), Nonce A' NA\<rbrace>) \<in> set tr \<rbrakk>
   \<Longrightarrow> (tsend,
        Send (Tx (Honest B) 0)
             Hash[Key (symKey (Honest B) A)]
                 \<lbrace>Agent (Honest B), Agent A, Nonce A' NA,
                  Real (trec_complete + coffset B),
                  Real (tsend + coffset B)\<rbrace>
             [])
      # tr \<in> SPS"

 | SPS3:
   "\<lbrakk> tr \<in> SPS; t \<ge> maxtime tr;
      delay (t1 + coffset A) t2 t3 (t4_complete + coffset A) \<le> bound_delay;
      t4_complete \<ge> t4_start + min_delay;
      t \<ge> t4_complete;
      (t4_start, Recv (Rx (Honest A) 0)
                   Hash[Key (symKey B (Honest A))]
                     \<lbrace>Agent B, Agent (Honest A), Nonce (Honest A) NA, Real t2, Real t3\<rbrace>)
      \<in> set tr;
      (t1, Send (Tr (Honest A)) \<lbrace>Agent (Honest A),Agent B,Nonce (Honest A) NA\<rbrace> []) \<in> set tr\<rbrakk>
  \<Longrightarrow> (t, Claim (Honest A) \<lbrace>Agent B,
                           Real (offset (t1 + coffset A) t2 t3 (t4_complete + coffset A))\<rbrace>) # tr
      \<in> SPS"

text {* obtain a simpler induction rule for protocol since it is
executable and deltaonly *}

lemmas proto_induct =
  sys.induct [simplified derivable_removable remove_occursAt]

subsubsection {* Equality for direct and Parameterized Definition *}

text {* We now show that both inductive definitions define the
        same set of traces. *}


(* MOVE *)
lemma (in PROTOCOL_EXECUTABLE) sys_Proto_exec:
  "\<lbrakk> sys = sys_param; tr \<in> sys_param; maxtime tr \<le> t;
     step \<in> proto; (m, pEv) \<in> step (timetrans A tr) A (clocktime A t)\<rbrakk>
   \<Longrightarrow> (t, createEv A pEv m) # tr \<in> sys"
by (rule sys.Proto,simp+,subst events_occur_at,simp+,rule messages_derivableI,simp,subst events_occur_at,simp+)
 

lemma abstr_equal: "SPS = sys"
proof safe
  fix tr
  assume r: "tr \<in> sys"
  show "tr \<in> SPS" using r
  proof (induct tr rule: proto_induct)
    case 1 with prems show ?case by assumption
  next
    case 2 with prems show ?case by (intro SPS.Nil)
  next
    case 4 with prems show ?case by (intro SPS.Con)
  next
    case 3 with prems show ?case by (intro SPS.Fake)
  next
    case (5 tr t step m pEv A)
    thus ?case
      apply (auto simp add: sps_defs)
      apply (rule SPS.SPS1)
      apply force+
      apply (simp add: mem_def, unfold timetrans_def)
      apply auto
      apply (drule_tac trec_start="a" and tsend=t and B=A and
	               trec_complete="trec_complete - coffset A" in SPS.SPS2)
      apply force+
      apply (drule_tac t=t and A=A and
	               t4_complete="t4_complete - coffset A" in SPS.SPS3)
      apply (auto simp add: offset_def delay_def)
      done
  qed
next
  fix tr
  assume r: "tr \<in> SPS"
  show "tr \<in> sys" using r
  proof(induct tr rule: SPS.induct)
    case Nil
    with prems show ?case by (intro sys.Nil)
  next
    case (Fake tr ts X I j)
    with prems show ?case by (intro sys.Fake)
  next
    case (Con tr tsend C i X D j dCD trecv)
    with prems show ?case by (intro sys.Con)
  next
    case (SPS1 tr ts C NC D)
    with prems have "(ts,createEv C (SendEv 0 []) \<lbrace>Agent (Honest C), Agent D, Nonce (Honest C) NC\<rbrace>) # tr \<in> sys"
      apply -
      apply (rule_tac step=sps1 in sys_Proto_exec)
      apply (force simp add: sps_defs)+
      apply (simp add: sps_defs)
      apply (simp add: mem_def)
      done
    thus ?case by (auto simp add: createEv.psimps)
  next
    case (SPS2 tr tsend trec_complete trec_start B A A' NA)
    with prems have
      "(tsend, 
        createEv B
                 (SendEv 0 [])
                 (Hash[Key (symKey (Honest B) A)]
                    \<lbrace>Agent (Honest B), Agent A, Nonce A' NA,
                     Real (trec_complete + coffset B), Real (tsend  + coffset B)\<rbrace>)) # tr \<in> sys"
      apply -
      apply (rule_tac step=sps2 in sys_Proto_exec)
      apply (force simp add: sps_defs)+
      apply (simp add: sps_defs)
      apply (rule_tac x="trec_start + coffset B" in exI)
      by (auto simp add: in_timetrans)
    thus ?case by (auto simp add: createEv.psimps)
  next
    case (SPS3 tr t t1 A t2 t3 t4_complete t4_start B NA)
    with prems have
      "(t,
        createEv A ClaimEv
                 (\<lbrace>Agent B, Real (offset (t1 + coffset A) t2 t3 (t4_complete + coffset A))\<rbrace>)) # tr \<in> sys"
      apply - 
      apply (rule_tac step=sps3 in sys_Proto_exec)
      apply (simp add: sps_defs)+
      apply (rule_tac x=NA in exI)
      apply (auto simp add: in_timetrans)
      apply (rule_tac x="t1 + coffset A" in exI)
      apply (rule_tac x=t2 in exI)
      apply (rule_tac x=t3 in exI)
      apply (rule_tac x="t4_start + coffset A" in exI)
      apply (rule_tac x="t4_complete + coffset A" in exI)
      apply (intro conjI) defer
      apply force+
      done
    thus ?case by (simp add: createEv.psimps)
    qed
qed

lemmas [simp,intro] = abstr_equal [THEN sym]

subsection {* A "possibility property": The protocol completes succesfully in certain scenarios*}

lemma delay_offset_remove[simp]:
  "delay (ta + ca) (tb + cb) (tc + cb) (td + ca) = delay ta tb tc td"
by (unfold delay_def, simp)


lemma SPS_possibility:
 assumes bound_ok: "delay 0
                          (cdist (Tr (Honest A)) (Rec (Honest B)) + min_delay)
                          (cdist (Tr (Honest A)) (Rec (Honest B)) + min_delay)
                          (cdist (Tr (Honest A)) (Rec (Honest B)) + min_delay + cdist (Tr (Honest B)) (Rec (Honest A)) + min_delay)  \<le> bound_delay"
 assumes distab: "cdistM (Tr (Honest A)) (Rec (Honest B)) \<noteq> None"
 assumes distba: "cdistM (Tr (Honest B)) (Rec (Honest A)) \<noteq> None"
 shows "\<exists>ta dt. \<exists>tr \<in> SPS. (ta,Claim (Honest A) \<lbrace>Agent (Honest B), Real dt\<rbrace>)
        \<in> set tr"
proof -
  let ?dab = "cdist (Tr (Honest A)) (Rec (Honest B))" and
      ?dba = "cdist (Tr (Honest B)) (Rec (Honest A))" and
      ?fa = "Honest A" and
      ?fb = "Honest B" and
      ?na = "Nonce (Honest A) NA"
  let ?msg1 = "\<lbrace>Agent ?fa, Agent ?fb, ?na\<rbrace>"
  let ?msg2 = "Hash[Key (symKey ?fb ?fa)]
                    \<lbrace>Agent ?fb, Agent ?fa, ?na,
                     Real ((?dab + min_delay) + coffset B), Real ((?dab + min_delay)  + coffset B)\<rbrace>"
  let ?t1 = 0 and
      ?t2 = "?dab+min_delay + coffset B"
  let ?t3 = "?t2" and
      ?t4 = "?dab+min_delay+(?dba+min_delay)"
  let ?done = "Claim ?fa \<lbrace>Agent ?fb, Real (offset (?t1 + coffset A) ?t2 ?t3 (?t4 + coffset A))\<rbrace>"
  from distab have dabnoneg: "?dab \<ge> 0"
    by (auto simp add: cdist_def intro: cdistnoneg_some)
  from distba have dbanoneg: "?dba \<ge> 0"
    by (auto intro: cdistnoneg_some simp add: cdist_def)
  from dabnoneg dbanoneg have snoneg: "?dab+?dba \<ge> 0" by auto
  have "(0,Send (Tr ?fa) ?msg1 [])#[] \<in> SPS" (is "?tr1 \<in> SPS")
    apply - apply (rule SPS.SPS1)
    apply (auto simp add: usedI_def)
    by (simp add: mem_def [THEN Pure.symmetric, where x1="Nonce (Honest A) NA"])
  with distab dabnoneg have "(0+?dab,Recv (Rec ?fb) ?msg1)#?tr1 \<in> SPS" (is "?tr2 \<in> SPS")
     apply -
     apply (rule_tac tab="cdist (Tr (Honest A)) (Rec (Honest B))" in SPS.Con)
     by (auto simp add: cdist_def)
  with dabnoneg have "(?dab + min_delay,Send (Tr ?fb) ?msg2 [])#?tr2 \<in> SPS"
    (is "?tr3 \<in> SPS") apply -
     by (rule_tac tr="?tr2" in SPS.SPS2,insert min_delay_pos,auto)
  hence "((?dab+min_delay)+?dba,Recv (Rec ?fa) ?msg2)#?tr3 \<in> SPS"
    (is "?tr4 \<in> SPS")
    using snoneg distba dbanoneg apply -
    apply (insert min_delay_pos)
    apply (rule SPS.Con)
    by (auto simp add: cdist_def)
  hence "(?dab + min_delay + ?dba + min_delay, ?done)#?tr4 \<in> SPS" (is "?tr5 \<in> SPS")
    apply -
    apply (insert min_delay_pos)
    apply (erule_tac t4_start="(?dab+min_delay)+?dba" and tr="?tr4" in SPS.SPS3)
    apply (insert dbanoneg snoneg)
    apply (insert bound_ok)
    by (auto simp add: delay_offset_remove [where ta=0 and ca="coffset A" and cb="coffset B", simplified])
  -- {* This is the witness trace: @{term [display] ?tr5} *}
  thus ?thesis
    apply -
    apply (rule_tac x="?dab + min_delay+(?dba+min_delay)" in exI)
    apply (rule_tac x="offset (?t1 + coffset A) ?t2 ?t3 (?t4 + coffset A)" in exI)
    apply (intro bexI) by auto
qed

subsection {* Security Proof *}

definition
  abs :: "real \<Rightarrow> real"
 where
  "abs x \<equiv> if 0 \<le> x then x else - x"


lemma minus_eq: "a = b ==> - (a ::real) = - b" by simp

lemma claim_implies:
  assumes sps: "tr \<in> SPS"
  assumes claim: "(t,Claim (Honest A) \<lbrace>Agent B, Real off\<rbrace>) \<in> set tr"
  shows "\<exists> t1 t2B t3B t4_start t4 NA t.
            (delay (t1 + coffset A) t2B t3B (t4 + coffset A)) \<le> bound_delay \<and>
            t4 \<ge> t4_start + min_delay \<and>
            t \<ge> t4 \<and>
            off = offset (t1 + coffset A) t2B t3B (t4 + coffset A) \<and>
            (t4_start,
             Recv (Rx (Honest A) 0)
                      Hash[Key (symKey B (Honest A))]
                          \<lbrace>Agent B, Agent (Honest A), Nonce (Honest A) NA, Real t2B, Real t3B\<rbrace>)
            \<in> set tr \<and>
            (t1, Send (Tr (Honest A)) \<lbrace>Agent (Honest A),Agent B,Nonce (Honest A) NA\<rbrace> [])
            \<in> set tr"
  using sps claim
  apply (induct rule: SPS.induct) prefer 6
  apply (case_tac "(t1,Claim (Honest A) \<lbrace>Agent B, msg.Real off\<rbrace>) =
                   (ta, Claim (Honest Aa)
                        \<lbrace>Agent Ba, msg.Real (offset (t1 + coffset Aa) t2 t3 (t4_complete + coffset Aa))\<rbrace>)")
  apply (rule_tac x=t1 in exI)
  apply (rule_tac x=t2 in exI)
  apply (rule_tac x=t3 in exI)
  apply (rule_tac x=t4_start in exI)
  apply (rule_tac x=t4_complete in exI)
  apply (rule_tac x=NA in exI)
  apply (rule_tac x=ta in exI)
  apply simp_all
  apply force+
done

lemma nonce_send_honest:
  assumes send: "(ta, Send (Tx (Honest A) i) \<lbrace>Agent (Honest A),Agent B,Nonce (Honest A) NA\<rbrace> L) \<in> set tr" and
          sps: "tr \<in> SPS"
  shows   "Nonce (Honest A) NA
           \<notin> usedI (beforeEvent (ta, Send (Tx (Honest A) i) \<lbrace>Agent (Honest A),Agent B,Nonce (Honest A) NA\<rbrace> L) tr)"
  using sps send
proof (induct rule: SPS.induct)
  case Nil thus ?case by simp
next
  case (Con tr tc C mc D tab)
  hence "(ta, Send (Tx (Honest A) i) \<lbrace>Agent (Honest A),Agent B,Nonce (Honest A) NA\<rbrace> L) \<in> set tr" by simp
  with Con.hyps prems show ?case by simp
next
  case (Fake tr mintr I tsend)
  hence "(ta, Send (Tx (Honest A) i)  \<lbrace>Agent (Honest A),Agent B,Nonce (Honest A) NA\<rbrace> L) \<in> set tr" by simp
  with Fake.hyps prems show ?case by simp
next
  case (SPS2 tr tsend trec_complete trec_start C D D' ND)
  hence "(ta, Send (Tx (Honest A) i)  \<lbrace>Agent (Honest A),Agent B,Nonce (Honest A) NA\<rbrace> L) \<in> set tr" by simp
  with SPS2.hyps prems show ?case by simp
next
  case (SPS3 tr t t1 D t2 t3 t4_complete t4_start C ND)
  hence "(ta, Send (Tx (Honest A) i) \<lbrace>Agent (Honest A),Agent B,Nonce (Honest A) NA\<rbrace>   L) \<in> set tr" by simp
  with SPS3.hyps prems show ?case by simp
next
  case (SPS1 tr t D ND C)
  with SPS1.hyps prems show ?case  by (auto, simp add: mem_def) 
qed
  
lemma prefix_SPS:
  assumes "ys \<in> SPS" and "ys = x # xs" shows "xs \<in> SPS" using assms
by (induct "ys" rule: SPS.induct, simp_all)

lemma monotone_SPS:
   assumes "ys \<in> SPS" and "ys = (tb, evb) # tr" and "(ta, eva) \<in> set tr" shows "ta \<le> tb" using assms
by (induct "ys" rule: SPS.induct, auto intro: maxtime_geq_elem)

lemma earlier_implies_beforeEvent:
  assumes "tr \<in> SPS"
  assumes "(ta, eva) \<in> set tr"
  assumes "(tb, evb) \<in> set tr"
  assumes "tb < ta"
  shows   "(tb, evb) \<in> set (beforeEvent (ta, eva) tr)" using assms
  apply -
  apply (induct tr)
  apply simp
  apply (frule_tac x=a and xs=tr in prefix_SPS)
  apply simp
  apply (case_tac "a =  (ta, eva)")
  apply simp
  apply (auto dest: monotone_SPS)
done

lemma used_beforeEvent_later:
  assumes sps: "tr \<in> SPS"
  assumes "(ta, Send (Tx A i) ma La) \<in> set tr"
  assumes "x \<notin> used (beforeEvent (ta, Send (Tx A i) ma La) tr)"
  assumes "(tb,  Recv (Rx B j) mb) \<in> set tr"
  assumes "x \<in> subterms {mb}"
  shows   "ta \<le> tb" using assms
  apply -
  apply (frule send_before_recv [simplified abstr_equal [THEN sym]])
  apply force
  apply (elim exE)
  apply auto
  apply (subgoal_tac "tsend \<le> tb") prefer 2
  apply (erule order_trans)
  apply (subgoal_tac "cdist (Tx Aa ia) (Rx B j) \<ge> 0")
  apply force
  apply (subgoal_tac "y = cdist (Tx Aa ia) (Rx B j)")
  apply (rule cdistnoneg_some)
  apply force
  apply (force simp add: cdist_def)
  apply (rule_tac y=tsend in order_trans)
  prefer 2
  apply assumption
  apply (rotate_tac 2)
  apply (erule contrapos_pp)
  apply (subgoal_tac "tsend < ta") defer
  apply force
  apply (frule_tac ta=ta and tb=tsend in earlier_implies_beforeEvent)
  apply force
  apply force
  apply force
  apply (simp)
  apply (rule Send_imp_used_parts)
  apply auto
done

lemma mac_implies:
  assumes sps: "tr \<in> SPS"
  assumes      "(tsend, Send (Tx (Honest E) i) me Le) \<in> set tr"
  assumes      "(E = A \<or> E = B)"
  assumes      "Hash
                    \<lbrace>Key (symKey (Honest B) (Honest A)),
                     Agent (Honest B), Agent (Honest A),
                     Nonce A' NA, msg.Real t2B, msg.Real t3B\<rbrace>
                \<in> subterms {me}"
  shows (* step SPS2 must have happened *)
        "\<exists> trec_start trec_complete.
           E = B \<and>
           trec_complete \<ge> trec_start + min_delay \<and>
           tsend \<ge> trec_complete \<and>
           (trec_start, Recv (Rx (Honest B) 0) \<lbrace>Agent (Honest A), Agent (Honest B), Nonce A' NA\<rbrace>) \<in> set tr \<and>
           me = Hash[Key (symKey (Honest B) (Honest A))]
                 \<lbrace>Agent (Honest B), Agent (Honest A), Nonce A' NA,
                  Real (trec_complete + coffset B),
                  Real (tsend + coffset B)\<rbrace>"
  using assms
proof (induct rule: SPS.induct)
  case Nil thus ?case by auto
next
  case (Con tr tc C mc D tab) thus ?case by auto
next
  case (Fake tr mintr I tsend) thus ?case by auto
next
  case (SPS1 tr t D ND C)
  have "(tsend, Send (Tx (Honest E) i) me Le) \<in> set tr" using prems
    apply -
    apply auto
    done
  thus ?case using prems by auto
next
  case (SPS3 tr t t1 D t2 t3 t4_complete t4_start C ND) thus ?case by auto
next
  case (SPS2 tr tsend trec_complete trec_start C D D' ND) thus ?case
    apply -
    apply auto
    done
qed

lemma SPS_inequalities:
  assumes sps: "tr \<in> SPS"
  assumes AnotB: "A \<noteq> B"
  assumes claim: "(ta,Claim (Honest A) \<lbrace>Agent (Honest B), Real off\<rbrace>) \<in> set tr"
  shows "\<exists> t1 t2 t3 t4. 
             min_delay \<le> t2 - t1 \<and>
             t2 \<le> t3 \<and>
             min_delay \<le> t4 - t3 \<and>
             off = offset (t1 + coffset A) (t2 + coffset B) (t3 + coffset B) (t4 + coffset A) \<and>
             delay t1 t2 t3 t4 \<le> bound_delay"
proof -
  (* step SPS3 must have been executed *)
  obtain t1 t2B t3B t4_start t4 NA t where
    A: "(delay (t1 + coffset A) t2B t3B (t4 + coffset A)) \<le> bound_delay" and
    B: "t4 \<ge> t4_start + min_delay" and
    C: "t \<ge> t4" and
    C2: "off =  offset (t1 + coffset A) t2B t3B (t4 + coffset A)" and
    D: "(t4_start,
         Recv (Rx (Honest A) 0)
              Hash[Key (symKey (Honest B) (Honest A))]
                  \<lbrace>Agent (Honest B), Agent (Honest A), Nonce (Honest A) NA, Real t2B, Real t3B\<rbrace>)
        \<in> set tr" and
    E: "(t1, Send (Tr (Honest A)) \<lbrace>Agent (Honest A),Agent (Honest B),Nonce (Honest A) NA\<rbrace> [])
        \<in> set tr" using sps claim
    apply -
    apply (drule claim_implies)
    by auto
  let ?mmac = "\<lbrace>Agent (Honest B), Agent (Honest A), Nonce (Honest A) NA, Real t2B, Real t3B\<rbrace>"
  let ?mac  = "Hash[Key (symKey (Honest B) (Honest A))] ?mmac"
  (* Either A or B must have sent something containing the MAC *)
  obtain t3 me i E Le where
    F: "(t3, Send (Tx (Honest E) i) me Le) \<in> set tr" and
    G: "(E = A \<or> E = B)" and
    H: "Hash \<lbrace>Key (symKey (Honest B) (Honest A)), ?mmac\<rbrace> \<in> subterms {me}" and
    I: "Hash \<lbrace>Key (symKey (Honest B) (Honest A)), ?mmac\<rbrace>
        \<notin> used (beforeEvent (t3, Send (Tx (Honest E) i) me Le) tr)"
    using sps D
    apply -
    apply (drule send_before_recv [simplified abstr_equal [THEN sym]])
    apply assumption
    apply (elim exE)
    apply (insert sps)
    apply (drule_tac mc="?mac" in mac_originates [simplified abstr_equal [THEN sym]])
    apply auto
    done
  (* This must have been before the MAC has been received *)
  hence J: "t3 \<le> t4_start" using D sps
    apply -
    apply (erule used_beforeEvent_later)
    apply force
    apply force
    apply force
    apply force
    done
  (* The first send of the MAC must come from step SPS2 *)
  obtain  t2_start t2 where
    K: "E = B" and
    L: "t2 \<ge> t2_start + min_delay" and
    M: "t3 \<ge> t2" and
    N: "(t2_start, Recv (Rx (Honest B) 0) \<lbrace>Agent (Honest A), Agent (Honest B), Nonce (Honest A) NA\<rbrace>) \<in> set tr" and
    O: "me = Hash[Key (symKey (Honest B) (Honest A))]
                 \<lbrace>Agent (Honest B), Agent (Honest A), Nonce (Honest A) NA,
                  Real (t2 + coffset B),
                  Real (t3 + coffset B)\<rbrace>" using sps F G H
    apply -
    apply (drule mac_implies)
    apply assumption
    apply assumption
    apply assumption
    apply auto
    done
  (* Freshness of nonce *)
  have P: "t1 \<le> t2_start" using prems
    apply -
    apply (frule_tac A="Honest A" and B="Honest B" and NA=NA and
                     mb="\<lbrace>Agent (Honest A),Agent (Honest B),Nonce (Honest A) NA\<rbrace>" and
                     ma = "\<lbrace>Agent (Honest A),Agent (Honest B),Nonce (Honest A) NA\<rbrace>"
           in fresh_nonce_earliest_recv [simplified abstr_equal [THEN sym]]) defer
    apply force
    apply force
    apply assumption
    apply force
    apply force
    apply (insert cdistl_noneg [where A="Honest A" and B="Honest B"], force)
    apply (drule nonce_send_honest)
    apply (unfold usedI_def)
    apply auto
    done
  (* The recv in SPS2 must have happened after the challenge has been sent in step SPS1 *)
  thus ?thesis using prems
    apply -
    apply (rule_tac x=t1 in exI)
    apply (rule_tac x=t2 in exI)
    apply (rule_tac x=t3 in exI)
    apply (rule_tac x=t4 in exI)
    apply auto
    apply (insert J, auto)
    done
qed

lemma SPS_secure:
  assumes sps: "tr \<in> SPS"
  assumes AnotB: "A \<noteq> B"
  assumes claim: "(ta,Claim (Honest A) \<lbrace>Agent (Honest B), Real off\<rbrace>) \<in> set tr"
  shows "abs (off - coffset B + coffset A) \<le> 2* bound_delay - 2* min_delay"
proof -
  obtain t1 t2 t3 t4 where
    A: "min_delay \<le> t2 - t1" and
    B: "t2 \<le> t3" and
    C: "min_delay \<le> t4 - t3" and
    D: "off = offset (t1 + coffset A) (t2 + coffset B) (t3 + coffset B) (t4 + coffset A)" and
    E: "delay t1 t2 t3 t4 \<le> bound_delay" using sps claim AnotB
    apply -
    apply (drule SPS_inequalities)
    apply auto
    done
  then have F: "off - coffset B + coffset A= offset t1 t2 t3 t4"
    apply (auto simp add: offset_def)
    done
  have G: "t2 -t1 \<le> 2 * bound_delay - min_delay" using A B C E
    by (auto simp add: delay_def)
  have H: "t4 -t3 \<le> 2 * bound_delay - min_delay" using A B C E
    by (auto simp add: delay_def)
  have I: "abs (offset t1 t2 t3 t4) \<le> 2 * bound_delay - 2 * min_delay" using A B G H
  proof cases
    assume "t2 - t1  \<ge> t4 - t3"
    hence "offset t1 t2 t3 t4 \<ge> 0" by (auto simp add: offset_def)
    thus ?thesis using A B C G H apply -
      by (auto simp add: abs_def offset_def)
  next
    assume "\<not>(t2 - t1 \<ge> t4 - t3)"
    hence  "t2 - t1 < t4 - t3" by auto
    hence "offset t1 t2 t3 t4 < 0" by (auto simp add: offset_def)
    hence "abs (offset t1 t2 t3 t4) = ((t4 - t3) - (t2 - t1)) / 2" apply (auto simp add: abs_def offset_def)
      apply (subgoal_tac "- ((t2 * 2 - t1 * 2 - (t4 * 2 - t3 * 2)) / 2) = - (t2 - t1 - (t4 - t3))")
      prefer 2
      apply (rule minus_eq)
      by auto
    thus ?thesis using A B C G H apply -
      by simp
  qed
  thus ?thesis using F G apply -
    by auto
qed

end
