(******************************************************************************

  ProtoVeriPhy

  Module: distance/MessageTheoryXor.thy

  Copyright (C) 2008-2012 Benedikt Schmidt <benedikt.schmidt@inf.ethz.ch>
  License: BSD (see COPYRIGHT)

******************************************************************************)

header {* Message Theory for XOR *}

theory MessageTheoryXor
imports MessageTheory Event 
        "~~/src/HOL/Library/List_lexord"
        "~~/src/HOL/Library/Multiset"
         NatTree
begin

section {* Message Algebra with XOR *}

text {* the term algebra for messages with xor *}

datatype
  fmsg =   AGENT  agent	     --{*Agent names*}
         | NUMBER int        --{*Ordinary integers *}
         | REAL   real       --{*Real Numbers, used for times, locations, ..*}
         | NONCE  agent nat
             --{*Unguessable nonces, tagged with agent to prevent collisions*}
         | KEY    key        --{*Crypto keys*}
	 | HASH   fmsg       --{*Hashing*}
	 | MPAIR  fmsg fmsg  --{*Compound messages*}
	 | CRYPT  key fmsg   --{*Encryption, public- or shared-key*}
         | XOR    fmsg fmsg  (infixr "\<oplus>" 65) --{*Exclusive-or of two messages*}
         | ZERO

subsection {* Linear Order on Messages via NatTree*}

datatype mleaf = TNat nat | TReal real | TInt int | TAgent agent

definition nil_tree[simp]: "nil = Leaf 0 (TNat 0)"

fun
  fmsg2tree :: "fmsg \<Rightarrow> mleaf tree"
 where
  "fmsg2tree (AGENT a)   = Leaf 1  (TAgent a)" |
  "fmsg2tree (NUMBER i)  = Leaf 2  (TInt i)" |
  "fmsg2tree (REAL r)    = Leaf 3  (TReal r)" | 
  "fmsg2tree (NONCE a n) = Node 4  (Leaf 41 (TAgent a)) (Node 42 (Leaf 42 (TNat n)) nil)" |
  "fmsg2tree (KEY k)     = Leaf 5  (TNat k)" |
  "fmsg2tree (HASH h)    = Node 6  (fmsg2tree h) nil" |
  "fmsg2tree (MPAIR a b) = Node 7  (fmsg2tree a) (Node 71 (fmsg2tree b) nil)" |
  "fmsg2tree (CRYPT k m) = Node 8  (Leaf 81 (TNat k)) (Node 81 (fmsg2tree m) nil)" |
  "fmsg2tree (XOR a b)   = Node 9  (fmsg2tree a) (Node 91 (fmsg2tree b) nil)" |
  "fmsg2tree ZERO        = Leaf 10 (TNat 0)"


instantiation mleaf :: linorder
begin

fun
  less_mleaf :: "mleaf \<Rightarrow> mleaf \<Rightarrow> bool"
 where
  "(TNat n)   < (TNat m)   = (n < m)" |
  "(TNat _)   < _          = True" |
  "(TReal r)  < (TReal s)  = (s < r)" |
  "(TReal _)  < (TNat _)   = False" |
  "(TReal _)  < _          = True" |
  "(TInt i)   < (TInt j)   = (i < j)" |
  "(TInt _)   < (TNat _)   = False" |
  "(TInt _)   < (TReal _)  = False" |
  "(TInt _)   < _          = True" |
  "(TAgent a) < (TAgent b) = (a < b)" |
  "(TAgent a) < _          = False"

definition less_eq_mleaf: "(a::mleaf) \<le> b = ((a = b) \<or> (a < b))"

instance proof
  fix x y :: mleaf show "(x < y) = (x \<le> y \<and> \<not> y \<le> x)"
    apply (auto simp add: less_eq_mleaf)
    apply (case_tac x, auto)
    apply (case_tac x)
    apply (case_tac y, auto)+
    done
next
  fix x :: mleaf show "x \<le> x" by (auto simp add: less_eq_mleaf)
next
  fix x y z ::mleaf show "\<lbrakk>x \<le> y; y \<le> z\<rbrakk> \<Longrightarrow> x \<le> z"
    apply (auto simp add: less_eq_mleaf)
    apply (case_tac x)
      apply (case_tac y)
        apply (case_tac z, auto)
      apply (case_tac z, auto)
      apply (case_tac z, auto)
      apply (case_tac z, auto)
      apply (case_tac z, auto)
    apply (case_tac y, auto)
    apply (case_tac y, auto)
    apply (case_tac y, auto)
      apply (case_tac z, auto)
      apply (case_tac z, auto)
    apply (case_tac y, auto)
    apply (case_tac z, auto)
    done
next
  fix x y :: mleaf show "\<lbrakk>x \<le> y; y \<le> x\<rbrakk> \<Longrightarrow> x = y"
    apply (auto simp add: less_eq_mleaf)
    apply (case_tac x)
      apply (case_tac y, auto)
      apply (case_tac y, auto)
      apply (case_tac y, auto)
    apply (case_tac y, auto)
    done
next
  fix x y :: mleaf show "x \<le> y \<or> y \<le> x"
    apply (auto simp add: less_eq_mleaf)
    apply (case_tac x)
      apply (case_tac y, auto)+
    done
qed

end

lemma fmsg2tree_inj: "inj fmsg2tree"
  apply (unfold inj_on_def)
  apply (rule ballI)
  apply (rule_tac fmsg=x in fmsg.induct) 
  apply auto
  apply (case_tac y, auto)+
done

lemmas fmsg2tree_inj2 = fmsg2tree_inj[simplified inj_on_def, rule_format, simplified]

instantiation fmsg :: linorder
begin

definition less_fmsg: "(a :: fmsg) < b = (fmsg2tree a < fmsg2tree b)"

definition less_eq_fmsg: "(a :: fmsg) \<le> b = (fmsg2tree a \<le> fmsg2tree b)"

instance proof
  fix x y :: fmsg show "(x < y) = (x \<le> y \<and> \<not> y \<le> x)"
    by (auto simp add: less_fmsg less_eq_fmsg)
next
  fix x :: fmsg show "x \<le> x" by (auto simp add: less_eq_fmsg)
next
  fix x y z :: fmsg show "\<lbrakk>x \<le> y; y \<le> z\<rbrakk> \<Longrightarrow> x \<le> z"
    by (auto simp add: less_eq_fmsg)
next
  fix x y :: fmsg show "\<lbrakk>x \<le> y; y \<le> x\<rbrakk> \<Longrightarrow> x = y"
    apply (auto simp add: less_eq_fmsg)
    apply (auto intro: fmsg2tree_inj2)
    done
next
  fix x y :: fmsg show "x \<le> y \<or> y \<le> x"
    apply (auto simp add: less_eq_fmsg)
    done
qed

end

subsection {* Normalization Function and its Properties *}

(* define norm as sorting and reducing the xors
   NOTE: 1. It takes a lot of time to expand the cases.
         2. The termination checker hates argument swapping, so don't do it! *)

definition
  XORnz :: "fmsg \<Rightarrow> fmsg \<Rightarrow> fmsg" (infixr "\<odot>" 65)
 where
  "XORnz a b = (if b = ZERO then a else a \<oplus> b)"

(* better definition, but have to adapt proofs 
function (sequential)
  normxor :: "fmsg \<Rightarrow> fmsg => fmsg" (infixr "\<otimes>" 65)
 where
  "x \<otimes> ZERO = x" |
  "ZERO \<otimes> x = x" |
  "(a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2) = 
     (if a1 = b1 then a2 \<otimes> b2
      else (if a1 < b1 then a1 \<odot> (a2 \<otimes> (b1 \<oplus> b2))
            else (b1 \<odot> ((a1 \<oplus> a2) \<otimes> b2))))" |
  "a \<otimes> (b1 \<oplus> b2) =
     (if a = b1 then b2
      else (if a < b1 then a \<oplus> (b1 \<oplus> b2)
            else b1 \<odot> (a \<otimes> b2)))" |
   "(b1 \<oplus> b2) \<otimes> a = a \<otimes> (b1 \<oplus> b2)" |
   "a \<otimes> b = (if a = b then ZERO else (if a < b then a \<oplus> b else b \<oplus> a))"
by pat_completeness auto
termination by sizechange*)


fun
  normxor :: "fmsg \<Rightarrow> fmsg => fmsg" (infixr "\<otimes>" 65)
 where
  "x \<otimes> ZERO = x" |
  "ZERO \<otimes> x = x" |
  "(a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2) = 
     (if a1 = b1 then a2 \<otimes> b2
      else (if a1 < b1 then a1 \<odot> (a2 \<otimes> (b1 \<oplus> b2))
            else (b1 \<odot> ((a1 \<oplus> a2) \<otimes> b2))))" |
  "a \<otimes> (b1 \<oplus> b2) =
     (if a = b1 then b2
      else (if a < b1 then a \<oplus> (b1 \<oplus> b2)
            else b1 \<odot> (a \<otimes> b2)))" |
   "(b1 \<oplus> b2) \<otimes> a =
      (if a = b1 then b2
       else (if a < b1 then a \<oplus> (b1 \<oplus> b2)
             else b1 \<odot> (b2 \<otimes> a)))" |
   "a \<otimes> b = (if a = b then ZERO else (if a < b then a \<oplus> b else b \<oplus> a))"

fun
  norm :: "fmsg \<Rightarrow> fmsg"
 where
  "norm (AGENT a)   = AGENT a" |
  "norm ZERO        = ZERO" |
  "norm (NUMBER n)  = NUMBER n" |
  "norm (REAL r)    = REAL r" |
  "norm (NONCE a t) = NONCE a t" |
  "norm (KEY k)     = KEY k" |
  "norm (HASH h)      = HASH (norm h)" |
  "norm (MPAIR a b)   = MPAIR (norm a) (norm b)" |
  "norm (CRYPT k m)   = CRYPT k (norm m)" |
  "norm (a \<oplus> b)   = (norm a) \<otimes> (norm b)"

lemma normxor_com: "x \<otimes> y = y \<otimes> x"
  apply (induct x arbitrary: y)
  apply (rule_tac fmsg=y in fmsg.induct, auto)+
done

definition
  standard :: "fmsg \<Rightarrow> bool"
 where
  "standard x \<equiv> x \<notin> {XOR x y | x y. True} \<union> {ZERO}"

lemma standard_xorD[dest]: "standard (XOR a b) \<Longrightarrow> P"
  apply (auto simp add: standard_def)
done

lemma standard_zeroD[dest]: "standard ZERO \<Longrightarrow> P"
  apply (auto simp add: standard_def)
done

(* use fun for standard? *)
lemma standard_AGENT[simp]: "standard (AGENT a)" by (auto simp add: standard_def)
lemma standard_NUMBER[simp]: "standard (NUMBER a)" by (auto simp add: standard_def)
lemma standard_REAL[simp]: "standard (REAL a)" by (auto simp add: standard_def)
lemma standard_NONCE[simp]: "standard (NONCE a b)" by (auto simp add: standard_def)
lemma standard_KEY[simp]: "standard (KEY a)" by (auto simp add: standard_def)
lemma standard_HASH[simp]: "standard (HASH h)" by (auto simp add: standard_def)
lemma standard_MPAIR[simp]: "standard (MPAIR a b)" by (auto simp add: standard_def)
lemma standard_CRYPT[simp]: "standard (CRYPT k m)" by (auto simp add: standard_def)

lemma normxor_case_standard_fst:
  "standard a \<Longrightarrow>
   a \<otimes> (x \<oplus> y) =
     (if a = x then y
        else (if a < x then a \<oplus> (x \<oplus> y)
              else x \<odot> (a \<otimes> y)))"
  apply (case_tac a, auto)
done

lemma normxor_case_standard_snd:
  "standard a \<Longrightarrow>
   (x \<oplus> y) \<otimes> a =
     (if a = x then y
      else (if a < x then
              a \<oplus> (x \<oplus> y)
            else x \<odot> (y \<otimes> a)))"
  apply (case_tac a, auto)
done

lemma normxor_case_standard_both:
  "\<lbrakk> standard a; standard b \<rbrakk> \<Longrightarrow>
   a \<otimes> b = (if a = b then ZERO else (if a < b then a \<oplus> b else b \<oplus> a))"
apply (case_tac a)
  apply (case_tac b)
    apply (force,force,force,force,force,force,force,force,force,force)
  apply (case_tac b)
    apply (force,force,force,force,force,force,force,force,force,force)
  apply (case_tac b)
    apply (force,force,force,force,force,force,force,force,force,force)
  apply (case_tac b)
    apply (force,force,force,force,force,force,force,force,force,force)
  apply (case_tac b)
    apply (force,force,force,force,force,force,force,force,force,force)
  apply (case_tac b)
    apply (force,force,force,force,force,force,force,force,force,force)
  apply (case_tac b)
    apply (force,force,force,force,force,force,force,force,force,force)
  apply (case_tac b)
    apply (force,force,force,force,force,force,force,force,force,force)
  apply (case_tac b)
    apply (force,force,force,force,force,force,force,force,force,force)
  apply (case_tac b)
    apply (force,force,force,force,force,force,force,force,force,force)
done

lemma normxor_case_zero_fst[simp]: "normxor ZERO x = x"
  apply (case_tac x)
  apply auto
done

lemma normxor_case_zero_snd[simp]: "normxor x ZERO = x"
  apply (case_tac x)
  apply auto
done

lemmas normxor_standard = normxor_case_standard_fst normxor_case_standard_snd normxor_case_standard_both

definition
  first :: "fmsg \<Rightarrow> fmsg"
 where
  "first x = (if standard x then x else case x of XOR a b \<Rightarrow> a | _ \<Rightarrow> x)"
  (* the redundant check for standard saves us a case distinction in the common case *)


lemma first_xor_fst_standard[simp]: "standard a \<Longrightarrow> first (XOR a b) = a"
  apply (auto simp add: first_def)
done

lemma first_standard[simp]: "standard x \<Longrightarrow> first x = x" by (auto simp add: first_def)
lemma first_ZERO[simp]: "first ZERO = ZERO" by (auto simp add: first_def)
lemma first_HASH[simp]: "first (HASH x) = HASH x" by (auto simp add: first_def)
lemma first_AGENT[simp]: "first (AGENT x) = AGENT x" by (auto simp add: first_def)
lemma first_NUMBER[simp]: "first (NUMBER x) = NUMBER x" by (auto simp add: first_def)
lemma first_REAL[simp]: "first (REAL x) = REAL x" by (auto simp add: first_def)
lemma first_NONCE[simp]: "first (NONCE x y) = NONCE x y" by (auto simp add: first_def)
lemma first_CRYPT[simp]: "first (CRYPT x y) = CRYPT x y" by (auto simp add: first_def)
lemma first_MPAIR[simp]: "first (MPAIR x y) = MPAIR x y" by (auto simp add: first_def)
lemma first_KEY[simp]: "first (KEY x) = KEY x" by (auto simp add: first_def)

inductive
  normed :: "fmsg => bool"
 where
   Agent[intro]:  "normed (AGENT a)"
 | Number[intro]: "normed (NUMBER n)"
 | Real[intro]:   "normed (REAL r)"
 | Nonce[intro]:  "normed (NONCE a t)"
 | Key[intro]:    "normed (KEY k)"
 | Zero[intro]:   "normed ZERO"
 | Hash[intro]:   "normed h \<Longrightarrow> normed (HASH h)"
 | MPair[intro]:  "\<lbrakk> normed a; normed b \<rbrakk> \<Longrightarrow> normed (MPAIR a b)"
 | Crypt[intro]:  "normed m \<Longrightarrow> normed (CRYPT k m)"
 | Xor:           "\<lbrakk> normed a; standard a; normed b; a < first b; b \<noteq> ZERO\<rbrakk>
                  \<Longrightarrow> normed (XOR a b)"


text {* Inversion rules for normed *}
lemma normed_XOR_ZERO_fst[intro]: "\<not> (normed (XOR ZERO a))"
proof -
  {
    fix x :: fmsg
    have "normed x \<Longrightarrow> \<forall> a. x \<noteq> XOR ZERO a"
      apply (induct x rule: normed.induct)
      apply auto
      done
  }
  thus ?thesis by auto
qed

lemma normed_XOR_ZERO_snd[intro]: "\<not> (normed (XOR a ZERO))"
proof -
  {
    fix x :: fmsg
    have "normed x \<Longrightarrow> \<forall> a. x \<noteq> XOR a ZERO"
      apply (induct x rule: normed.induct)
      apply auto
      done
  }
  thus ?thesis by auto
qed

lemma normed_XOR_XOR_fst[intro]: "\<not> (normed (XOR (XOR a b) c))"
proof -
  {
    fix x :: fmsg
    have "normed x \<Longrightarrow> \<forall> a b c. x \<noteq> XOR (XOR a b) c"
      apply (induct x rule: normed.induct)
      apply auto
      done
  }
  thus ?thesis by auto
qed

lemma normed_XOR_same: "\<not> normed (XOR x x)"
proof -
  {
    fix x :: fmsg
    have "normed x \<Longrightarrow> \<forall> a. x \<noteq> XOR a a"
      apply (induct x rule: normed.induct)
      apply (auto simp add: first_def)
      done
  }
  thus ?thesis by auto
qed

lemma normed_XOR_sameD[dest]: "normed (XOR x x) \<Longrightarrow> P"
by (insert normed_XOR_same, auto)

lemma normed_XOR_XOR_fstD[dest]: "normed (XOR (XOR a b) c) \<Longrightarrow> P"
by (insert normed_XOR_XOR_fst, auto)

lemma normed_XOR_ZERO_fstD[dest]: "normed (XOR ZERO x) \<Longrightarrow> P"
by (insert normed_XOR_ZERO_fst, auto)

lemma normed_XOR_ZERO_sndD[dest]: "normed (XOR x ZERO) \<Longrightarrow> P"
by (insert normed_XOR_ZERO_snd, auto)

lemma order_fmsg_total: "x \<noteq> y \<Longrightarrow> \<not> ((x::fmsg) < y) \<Longrightarrow> y < x"
by auto

inductive_cases normed_XOR_nested: "normed (XOR a (XOR b c))"
inductive_cases normed_XOR: "normed (XOR a b)"
inductive_cases normed_HASH: "normed (HASH a)"
inductive_cases normed_MPAIR: "normed (MPAIR a b)"
inductive_cases normed_CRYPT: "normed (CRYPT k m)"

lemma normed_xor_snd: "normed (XOR a b) \<Longrightarrow> normed b"
  apply (erule normed_XOR)
  apply auto
done

lemma normed_xor_fst: "normed (XOR a b) \<Longrightarrow> normed a"
  apply (erule normed_XOR)
  apply auto
done

lemma normed_xor_smaller_standard: "\<lbrakk> normed (XOR a b);  standard b \<rbrakk> \<Longrightarrow> a < b"
  apply (erule normed_XOR)
  apply (auto simp add: first_def)
done

lemma normed_xor_smaller_nested: "\<lbrakk> normed (XOR a (XOR b c)) \<rbrakk> \<Longrightarrow> a < b"
  apply (erule normed_XOR_nested)
  apply (auto simp add: first_def split: split_if_asm)
done

lemma normed_xor_fst_standard: "normed (XOR x1 x2) \<Longrightarrow> standard x1"
  apply (erule normed_XOR)
  apply (auto simp add: first_def split: split_if_asm)
done


lemma normed_xor_snd_nozero: "normed (XOR x1 x2) \<Longrightarrow> x2 \<noteq> ZERO"
  apply (erule normed_XOR)
  apply (auto simp add: first_def split: split_if_asm)
done

lemma normed_xor_not_nested_diff:
  "\<lbrakk> x < y; standard x; standard y; normed x; normed y \<rbrakk> \<Longrightarrow> normed (XOR x y)"
  apply (rule normed.Xor)
  apply (auto simp add: first_def split: split_if_asm)
done

lemma normed_XOR_XOR_smaller_trans:
  "\<lbrakk> normed (XOR a (XOR b c)); standard c \<rbrakk> \<Longrightarrow> a < c"
  apply (erule normed_XOR_nested)
  apply auto
  apply (erule normed_XOR)
  apply (auto simp add: first_def split: split_if_asm)
done

lemma standard_xor_nested_normxor:
  assumes normeda:    "normed a"
  and     standarda:  "standard a"
  and     normedb:    "normed b"
  and     standardb:  "standard b"
  and     normedxor:  "normed (b1 \<oplus> b2)"
  and     normedaxor: "normed (a \<otimes> (b1 \<oplus> b2))"
  and     bless:      "b < b1"
  shows "normed (a \<otimes> (b \<oplus> (b1 \<oplus> b2)))" using prems
proof -
  show ?thesis proof cases
    assume "a = b"
    thus ?thesis using prems
      apply (case_tac a)
      by auto
  next
    assume neq: "a \<noteq> b"
    show ?thesis proof cases
      assume "a < b"
      thus ?thesis using prems
	apply (case_tac a)
	apply (auto intro!: normed.Xor split: split_if_asm simp add: first_def)
	done
    next
      assume "\<not> a < b"
      hence xle: "b < a" using neq by auto
      thus ?thesis using prems
	apply (auto simp add: normxor_standard XORnz_def split: split_if_asm)
	apply (case_tac "standard b2") prefer 3
	apply (case_tac "standard b2") prefer 5
	apply (case_tac "standard b2") prefer 7
	apply (case_tac "standard b2") 
	apply (auto intro!: normed.Xor intro: order_fmsg_total
	  split: split_if_asm
	  simp add: first_def
	  dest: normed_xor_smaller_standard
	  normed_xor_smaller_nested normed_xor_fst_standard)
	apply (case_tac b2, auto)
	apply (drule normed_xor_smaller_nested)
	apply force
	done
    qed
  qed
qed

lemma standard_xor_normxor:
  assumes normeda:    "normed a"
  and     standarda:  "standard a"
  and     normedx:    "normed x"
  and     normedy:    "normed y"
  and     standardx:  "standard x"
  and     standardy:  "standard y"
  and     normedxor:  "normed (a \<otimes> y)"
  and     normedaxor: "normed (a \<otimes> x)"
  and     aless:      "x < y"
  shows "normed (a \<otimes> (x \<oplus> y))" using prems
  apply (case_tac a, auto simp add: normxor_standard XORnz_def)
  apply (auto intro!: normed.Xor split: split_if_asm simp add: first_def)
done

lemma xor_normxor:
  assumes normeda:    "normed a"
  and     standarda:  "standard a"
  and     normedx:    "normed (x \<oplus> y)"
  and     normedxor:  "normed (a \<otimes> y)"
  and     normedaxor: "normed (a \<otimes> x)"
  and     aless:      "x < first y"
  and     ynotzero:   "y \<noteq> ZERO"
  shows "normed (a \<otimes> (x \<oplus> y))" using prems
  apply -
  apply (frule normed_xor_fst)
  apply (frule normed_xor_snd)
  apply (frule normed_xor_fst_standard)
  apply (case_tac "standard y")
  apply (rule standard_xor_normxor)
  apply force apply force apply force
  apply force apply force apply force
  apply force apply force
  apply (force simp add: first_def)
  apply (case_tac y)
  apply force apply force apply force apply force
  apply force apply force apply force apply force
  apply (simp only: ext)
  apply (rule standard_xor_nested_normxor)
by (auto simp add: first_def)

lemma normxor_normed_com: "normed (a \<otimes> b) \<Longrightarrow> normed (b \<otimes> a)"
by (auto simp add: normxor_com)

lemma standard_standard_normxor:
  assumes "normed a"
  and "normed b"
  and "standard a"
  and "standard b"
  shows "normed (a \<otimes> b)" using prems
  apply (case_tac a)
  apply (auto intro!: normed.Xor order_fmsg_total simp add: first_def normxor_standard split: split_if_asm)
done

lemma normed_xor_smaller[intro]: "\<lbrakk> normed (XOR a b) \<rbrakk> \<Longrightarrow> a < first b"
  apply (erule normed_XOR)
  apply (auto simp add: first_def)
done

lemma normxor_assoc:
  assumes st: "standard a"
  and     le_b: "a < first b"
  and     le_c: "a < first c"
  and     bnz:  "b \<noteq> ZERO"
  and     cnz:  "c \<noteq> ZERO"
  shows   "(a \<oplus> b) \<otimes> c = b \<otimes> (a \<oplus> c)" using prems
proof cases
  assume "b \<otimes> c = ZERO"
  have "(a \<oplus> b) \<otimes> c = a" using prems
    apply (case_tac "standard c")
    apply (auto simp add: first_def normxor_standard XORnz_def split: split_if_asm)
    apply (case_tac c, auto simp add: first_def normxor_standard XORnz_def)
    apply (case_tac c, auto simp add: first_def normxor_standard XORnz_def)
    done
  also have "a =b \<otimes> (a \<oplus> c)" using prems
    apply (case_tac "standard b")
    apply (auto simp add: first_def normxor_standard XORnz_def split: split_if_asm)
    apply (case_tac b, auto simp add: first_def normxor_standard XORnz_def)
    apply (case_tac b, auto simp add: first_def normxor_standard XORnz_def)
    done
  finally show ?thesis by auto
next
  assume "b \<otimes> c \<noteq> ZERO"
  have "(a \<oplus> b) \<otimes> c = a \<oplus> (b \<otimes> c)" using prems
    apply (case_tac "standard c")
    apply (auto simp add: first_def normxor_standard XORnz_def split: split_if_asm)
    apply (case_tac c, auto simp add: first_def normxor_standard XORnz_def)
    apply (case_tac c, auto simp add: first_def normxor_standard XORnz_def)
    done
  also have "... = b \<otimes> (a \<oplus> c)" using prems
    apply (case_tac "standard b")
    apply (auto simp add: normxor_standard XORnz_def first_def split: split_if_asm)
    apply (case_tac b, auto simp add: first_def normxor_standard XORnz_def)
    apply (case_tac c, auto simp add: first_def normxor_standard XORnz_def)
    apply (case_tac b, auto simp add: first_def normxor_standard XORnz_def)
    done
  finally show "(a \<oplus> b) \<otimes> c = b \<otimes> (a \<oplus> c)" by auto
qed

lemma normxor_first:
  assumes "normed x"
  and     "normed y"
  and     "normxor x y \<noteq> ZERO"
  shows   "first (x \<otimes> y) \<ge> min (first x) (first y)" using prems
proof (induct x arbitrary: y)
  case (Agent a)
  thus ?case using prems
    apply (induct y)
    apply (auto simp add: normxor_standard XORnz_def)
    done
next
  case (Hash h)
   show ?case using prems(4,6-)
    apply (induct y)
    apply (auto simp add: normxor_standard XORnz_def)
    apply (auto dest: normed_xor_smaller normed_xor_fst_standard)
    done
next
  case (MPair a b)
  show ?case using prems(4,6,8-)
    apply (induct y)
    apply (auto simp add: normxor_standard XORnz_def)
    apply (auto dest: normed_xor_smaller normed_xor_fst_standard)
    done
next
  case (Real r)
  thus ?case
    apply (induct y)
    apply (auto simp add: normxor_standard XORnz_def)
    done
next
  case (Crypt m k)
  show ?case using prems(4,6-)
    apply (induct y)
    apply (auto simp add: normxor_standard XORnz_def)
    apply (auto dest: normed_xor_smaller normed_xor_fst_standard)
    done
next
  case (Number n)
  thus ?case
    apply (induct y)
    apply (auto simp add: normxor_standard XORnz_def)
    done
next
  case (Nonce a n)
  thus ?case
    apply (induct y)
    apply (auto simp add: normxor_standard XORnz_def)
    done
next
  case (Key k)
  thus ?case
    apply (induct y)
    apply (auto simp add: normxor_standard XORnz_def)
    done
next
  case Zero
  thus ?case
    apply (induct y)
    apply (auto simp add: normxor_standard XORnz_def)
    done
next
  case(Xor a1 a2)
  show ?case using prems(4,6,7,9-)
    apply (induct y)
    apply (auto simp add: normxor_standard XORnz_def) defer
    apply (force dest: normed_xor_smaller normed_xor_fst_standard) defer
    apply (force dest: normed_xor_smaller normed_xor_fst_standard)
    apply (case_tac "standard y2")
      apply (auto split: split_if_asm simp add: normxor_standard XORnz_def)
    apply (case_tac y2)
    apply (auto split: split_if_asm simp add: normxor_standard XORnz_def) prefer 2
    apply (drule normed_xor_snd)
    apply force
    apply (erule normed_XOR)
    apply auto
    apply (frule normed_xor_snd)
    apply simp
    apply (drule prems(8)) back back back
    apply force
    apply (auto simp add: min_def split: split_if_asm)
    apply (frule normed_xor_smaller)
    apply force
    done
qed

lemma normed_normxor:
  assumes na: "normed a"
  and     nb: "normed b"
  shows   "normed (a \<otimes> b)"
 using na nb
proof (induct a arbitrary: b rule: normed.induct)
  case (Agent a)
  show ?case using `normed b`
  proof (induct b)
    case (Agent x) show ?case by (rule standard_standard_normxor, auto)
  next case (Number x) show ?case by (rule standard_standard_normxor, auto)
  next case (Real x) show ?case by (rule standard_standard_normxor, auto)
  next case (Key k) show ?case by (rule standard_standard_normxor, auto)
  next case (Nonce x y) show ?case by (rule standard_standard_normxor, auto)
  next case Zero show ?case by auto
  next case (Hash h)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (MPair x y)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Crypt k m)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next
    case (Xor x y) show ?case using prems apply - by (rule xor_normxor, auto intro: normed.Xor)
  qed
next
  case (Xor a b c)
  have normab: "normed (XOR a b)" using prems apply - apply (rule normed.Xor) by auto
  have "normed (normxor c (XOR a b))" using `normed c`
  proof (induct c)
    case (Agent x)
    show ?case using prems(1-3) prems(5-6) prems(8-) normab apply -
      apply (rule xor_normxor)
      by (auto intro: prems(7)[of "AGENT x", THEN normxor_normed_com] normed.Xor
	       simp add: first_def normxor_standard XORnz_def)
  next case (Number x)
    show ?case using prems(1-3) prems(5-6) prems(8-) normab apply -
      apply (rule xor_normxor)
      by (auto intro: prems(7)[of "NUMBER x", THEN normxor_normed_com] normed.Xor
	       simp add: first_def normxor_standard XORnz_def)
  next case (Real x)
    show ?case using prems(1-3) prems(5-6) prems(8-) normab apply -
      apply (rule xor_normxor)
      by (auto intro: prems(7)[of "REAL x", THEN normxor_normed_com] normed.Xor
	       simp add: first_def normxor_standard XORnz_def)
  next case (Key k)
    show ?case using prems(1-3) prems(5-6) prems(8-) normab apply -
      apply (rule xor_normxor)
      by (auto intro: prems(7)[of "KEY k", THEN normxor_normed_com] normed.Xor
               simp add: first_def normxor_standard XORnz_def)
  next case Zero show ?case using prems
      apply simp
      apply (rule normed.Xor)
      by auto
  next case (Nonce x y)
    show ?case using prems(1-3) prems(5-6) prems(8-) normab apply -
      apply (rule xor_normxor)
      by (auto intro: prems(7)[of "NONCE x y", THEN normxor_normed_com] normed.Xor
	       simp add: first_def normxor_standard XORnz_def)
  next case (Hash h)
    show ?case using prems(1-3) prems(5-6) prems(8-) normab apply -
      apply (rule xor_normxor)
      by (auto intro: prems(7)[of "HASH h", THEN normxor_normed_com] normed.Xor
	       simp add: first_def normxor_standard XORnz_def)
  next case (MPair x y)
    show ?case using prems(1-3) prems(5-6) prems(8-) normab apply -
      apply (rule xor_normxor)
      by (auto intro: prems(7)[of "MPAIR x y", THEN normxor_normed_com] normed.Xor
               simp add: first_def normxor_standard XORnz_def)
  next case (Crypt k m)
    show ?case using prems(1-3) prems(5-6) prems(8-) normab apply -
      apply (rule xor_normxor)
      by (auto intro: prems(7)[of "CRYPT m k", THEN normxor_normed_com] normed.Xor
               simp add: first_def normxor_standard XORnz_def)
  next
    case (Xor x y)
    have normedxy: "normed (x \<oplus> y)"
      using `x < first y` `standard x` `normed x` `normed y` `y \<noteq> ZERO`
      by (auto intro: normed.Xor normxor_standard XORnz_def)
    show ?case proof cases
      assume "a = x"
      thus ?case using `y\<noteq>ZERO` `normed y` `normed b`
	apply (auto intro: normed.Xor)
	by (erule prems(7)[THEN normxor_normed_com])
    next
      assume neq: "a \<noteq> x"
      show ?case proof cases
	assume le: "a < x"
	show ?case proof cases
	  assume nxzero: "b \<otimes> (x \<oplus> y) = ZERO"
	  hence "(a \<oplus> b) \<otimes> (x \<oplus> y) = a" using le
	    by (auto split: split_if_asm simp add: normxor_standard XORnz_def)
	  thus ?case using `normed a` by (auto simp only: normxor_com)
	next 
	  assume nxnotzero: "b \<otimes> (x \<oplus> y) \<noteq> ZERO"
	  hence eq1: "(a \<oplus> b) \<otimes> (x \<oplus> y) = a \<oplus> (b \<otimes> (x \<oplus> y))"
	    using prems by (auto simp add: normxor_standard XORnz_def)
	  have normedxor: "normed (b \<otimes> (x \<oplus> y))" using normedxy
	    by (rule prems(7))
	  have less_a: "a < first (b \<otimes> (x \<oplus> y))"
	    proof cases
	      assume "b = x"
	      thus ?thesis using prems(1-3) prems(5-6) prems(8-)
		apply (case_tac "standard b")
		by (auto simp add: normxor_standard XORnz_def)
	    next
	      assume neq: "b \<noteq> x"
 	      thus ?thesis
		proof cases
		  assume "b < x"
		  show ?thesis using prems(1-3) prems(5-6) prems(8-)
		    apply (case_tac "standard b")
		    apply (force simp add: first_def normxor_standard XORnz_def)
		    apply (case_tac b)
		    apply force apply force apply force
		    apply force apply force apply force
		    apply force apply force prefer 2
		    apply force
		    apply simp
		    apply (frule normed_xor_fst_standard)
		    apply (drule normed_xor_smaller)
		    apply (auto split: split_if_asm simp add: normxor_assoc normxor_standard XORnz_def)
		    apply (frule normed_xor_snd)
		    apply (frule normed_xor_smaller)
		    apply (drule_tac x=fmsg2 and y=y in normxor_first)
		    apply (auto simp add: min_def  normxor_standard XORnz_def split: split_if_asm)
		    done
		next
		  assume "\<not> ( b < x)"
		  hence "x < b" using neq by auto
		  show ?thesis using prems(1-3) prems(5-6) prems(8-)
		    apply (auto simp add:  normxor_standard XORnz_def)
		    apply (case_tac "standard b")
		    apply (force simp add: first_def  normxor_standard XORnz_def)
		    apply (case_tac b)
		    apply force apply force
		    apply force apply force
		    apply force apply force
		    apply force apply force
		    apply (auto split: split_if_asm dest: normed_xor_fst_standard
		           simp add:  normxor_standard XORnz_def)
		    apply (frule normed_xor_fst_standard)
		    apply simp
		    apply (frule normed_xor_fst_standard)
		    apply (frule normed_xor_snd)
		    apply (frule normed_xor_smaller)
		    apply (drule_tac x=fmsg2 and y=y in normxor_first)
		    apply (auto simp add: min_def  normxor_standard XORnz_def split: split_if_asm)
		    done
		qed
	      qed
	  hence "normed (a \<oplus> (b \<otimes> (x \<oplus> y)))"
	    using `normed a` `standard a` normedxy nxnotzero less_a
	    apply (case_tac "standard (b \<otimes> (x \<oplus> y))")
	      apply (rule normed.Xor) prefer 6
	    apply (case_tac "b \<otimes> (x \<oplus> y)", auto intro: prems)
	    apply (rule normed.Xor)
	    apply auto
	    apply (drule_tac b="x \<oplus> y" in prems(7))
	    apply force
	    done
	  thus ?case apply (auto simp only: eq1 normxor_com) done
	qed
      next
	assume "\<not> (a < x)"
	hence le: "x < a" using neq by auto
	show ?case proof cases
	  assume nxzero: "(a \<oplus> b) \<otimes> y = ZERO"
	  hence "(a \<oplus> b) \<otimes> (x \<oplus> y) = x" using le
	    by (auto split: split_if_asm simp add:  normxor_standard XORnz_def)
	  thus ?case using `normed x` by (auto simp only: normxor_com)
	next
	  assume nxnotzero: "(a \<oplus> b) \<otimes> y \<noteq> ZERO"
	  hence eq1: "(a \<oplus> b) \<otimes> (x \<oplus> y) = x \<oplus> ((a \<oplus> b) \<otimes> y)"
	    using prems by (auto simp add: normxor_standard XORnz_def)
	  have normedxor: "normed ((a \<oplus> b) \<otimes> y)"
	    using normedxy prems(1-3) prems(5-6) prems(8-)
	    apply -
	    apply (auto simp add:  normxor_standard XORnz_def)
	    apply (case_tac "standard y")
	      apply (auto simp add:  normxor_standard XORnz_def)
	      apply (rule normed.Xor)
	      apply (auto simp add:  normxor_standard XORnz_def)
	      apply (rule prems(7))
	      apply auto prefer 2
	    apply (case_tac y)
	    by (auto split: split_if_asm simp add: normxor_com normxor_standard XORnz_def)
	  hence "normed (x \<oplus> ((a \<oplus> b) \<otimes>  y))"
	    using `normed x` `standard x` normedxy nxnotzero prems(1-3) prems(5-6) prems(8-) apply -
	    apply (rule normed.Xor)
	    apply (auto split: split_if_asm simp add: normxor_standard XORnz_def)
	    apply (frule normed_xor_fst_standard)
	    apply (frule normed_xor_snd) back
	    apply (frule normed_xor_smaller) back
	    apply (drule_tac x="XOR a b" and y=y in normxor_first)
	    apply auto
	    apply (auto simp add: min_def split: split_if_asm)
	    done
	  thus ?case apply (auto simp only: eq1 normxor_com) done
	qed
      qed
    qed
  qed
  thus ?case apply - by (erule normxor_normed_com)
next
  case (Number int)
  show ?case using `normed b`
  proof (induct b)
    case (Agent x) show ?case by (rule standard_standard_normxor, auto)
  next case (Number x) show ?case by (rule standard_standard_normxor, auto)
  next case (Real x) show ?case by (rule standard_standard_normxor, auto)
  next case (Key k) show ?case by (rule standard_standard_normxor, auto)
  next case (Nonce x y) show ?case by (rule standard_standard_normxor, auto)
  next case Zero show ?case by auto
  next case (Hash h)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (MPair x y)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Crypt k m)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next
    case (Xor x y) show ?case using prems apply - by (rule xor_normxor, auto intro: normed.Xor)
  qed
next
  case (Hash m)
  show ?case using `normed b`
  proof (induct b)
    case (Agent x)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Number x)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Real x)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Key k)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Nonce x y)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case Zero show ?case using prems by auto
  next case (Hash h)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (MPair x y)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Crypt k m)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next
    case (Xor x y) show ?case using prems apply - by (rule xor_normxor, auto intro: normed.Xor)
  qed
next
  case (MPair x y)
  show ?case using `normed b`
  proof (induct b)
    case (Agent x)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Number x)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Real x)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Key k)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Nonce x y)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case Zero show ?case using prems by auto
  next case (Hash h)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (MPair x y)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Crypt k m)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next
    case (Xor x y) show ?case using prems apply - by (rule xor_normxor, auto intro: normed.Xor)
  qed
next
  case (Crypt k m)
  show ?case using `normed b`
  proof (induct b)
    case (Agent x)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Number x)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Real x)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Key k)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Nonce x y)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case Zero show ?case using prems by auto
  next case (Hash h)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (MPair x y)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Crypt k m)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next
    case (Xor x y) show ?case using prems apply - by (rule xor_normxor, auto intro: normed.Xor)
  qed (* XXX: this is getting boring: define a new induction principle
              that splits standard, Zero, Xor, and nested Xor *)
next
  case (Real r)
  show ?case using `normed b`
  proof (induct b)
    case (Agent x)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Number x)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Real x)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Key k)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Nonce x y)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case Zero show ?case using prems by auto
  next case (Hash h)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (MPair x y)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Crypt k m)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next
    case (Xor x y) show ?case using prems apply - by (rule xor_normxor, auto intro: normed.Xor)
  qed
next
  case (Nonce a t)
  show ?case using `normed b`
  proof (induct b)
    case (Agent x)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Number x)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Real x)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Key k)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Nonce x y)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case Zero show ?case using prems by auto
  next case (Hash h)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (MPair x y)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Crypt k m)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next
    case (Xor x y) show ?case using prems apply - by (rule xor_normxor, auto intro: normed.Xor)
  qed
next
  case (Key k)
  show ?case using `normed b`
  proof (induct b)
    case (Agent x)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Number x)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Real x)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Key k)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Nonce x y)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case Zero show ?case using prems by auto
  next case (Hash h)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (MPair x y)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next case (Crypt k m)
    show ?case using prems apply - by (rule standard_standard_normxor, auto)
  next
    case (Xor x y) show ?case using prems apply -
      by (rule xor_normxor, auto intro: normed.Xor)
  qed
next
  case Zero
  show ?case using `normed b`
  proof (induct b)
    case (Agent x)
    show ?case using prems apply - by auto
  next case (Number x) show ?case using prems apply - by auto
  next case (Real x) show ?case using prems apply - by auto
  next case (Key k) show ?case using prems apply - by auto
  next case (Nonce x y) show ?case using prems apply - by auto
  next case Zero show ?case using prems by auto
  next case (Hash h) show ?case using prems apply - by auto
  next case (MPair x y) show ?case using prems apply - by auto
  next case (Crypt k m) show ?case using prems apply - by auto
  next case (Xor x y) show ?case using prems apply -
      apply auto
      apply (case_tac y)
      apply (auto intro!: normed.Xor)
      done
  qed
qed

  
lemma normed_norm: "normed (norm x)"
proof (induct x)
  case (NUMBER i)
  show ?case by auto
next
  case (AGENT a)
  show ?case by auto
next
  case ZERO
  show ?case by auto
next
  case (REAL r)
  show ?case by auto
next
  case (NONCE a n)
  show ?case by auto
next
  case (KEY k)
  show ?case by auto
next
  case (HASH h)
  show ?case using prems by auto
next
  case (MPAIR a b)
  show ?case using prems by auto
next
  case (CRYPT k m)
  show ?case using prems by auto
next
  case (XOR a b)
  show ?case using prems
    apply (auto intro: normed_normxor)
    done
qed

lemma normxor_normed_id:
  assumes nx: "normed (XOR a b)"
  shows   "a \<otimes> b = a \<oplus> b" using prems
proof -
  have norma: "normed a" using nx by (rule normed_xor_fst)
  have normb: "normed b" using nx by (rule normed_xor_snd)
  show ?thesis using norma normb nx
  proof (induct a arbitrary: b)
    case (Agent a)
    thus ?case apply -
      apply (frule normed_xor_smaller)
      apply (case_tac "standard b", auto simp add: first_def normxor_standard XORnz_def)
      apply (case_tac b, auto)
      done
  next
    case (Real x)
    thus ?case apply -
      apply (frule normed_xor_smaller)
      apply (case_tac "standard b", auto simp add: first_def normxor_standard XORnz_def)
      apply (case_tac b, auto)
      done
  next
    case (Number i)
    thus ?case apply -
      apply (frule normed_xor_smaller)
      apply (case_tac "standard b", auto simp add: first_def normxor_standard XORnz_def)
      apply (case_tac b, auto)
      done
  next
    case (Key k)
    thus ?case apply -
      apply (frule normed_xor_smaller)
      apply (case_tac "standard b", auto simp add: first_def normxor_standard XORnz_def)
      apply (case_tac b, auto)
      done
  next
    case (Nonce a m)
    thus ?case apply -
      apply (frule normed_xor_smaller)
      apply (case_tac "standard b", auto simp add: first_def normxor_standard XORnz_def)
      apply (case_tac b, auto)
      done
  next
    case (Hash h)
    show ?case using prems(2) prems(4-5) apply -
      apply (frule normed_xor_smaller)
      apply (case_tac "standard b")
      apply (auto simp add: first_def normxor_standard XORnz_def)
      apply (case_tac b)
      apply (auto simp add: first_def normxor_standard XORnz_def)
      done
  next
    case (MPair a b c)
    show ?case using prems(2) prems(4) prems(6-7) apply -
      apply (frule normed_xor_smaller)
      apply (case_tac "standard c")
      apply (auto simp add: first_def normxor_standard XORnz_def)
      apply (case_tac c)
      apply (auto simp add: first_def normxor_standard XORnz_def)
      done
  next
    case (Crypt m k c)
    show ?case using prems(2) prems(4-5) apply -
      apply (frule normed_xor_smaller)
      apply (case_tac "standard c")
      apply (auto simp add: first_def normxor_standard XORnz_def)
      apply (case_tac c)
      apply (auto simp add: first_def normxor_standard XORnz_def)
      done
  next
    case Zero
    show ?case using prems by auto
  next
    case (Xor a b c)
    thus ?case by auto
  qed
qed

lemma norm_normed_id:
  assumes nx: "normed x"
  shows   "norm x = x"
  using nx
  apply (induct x)
  apply (auto simp add: normxor_standard XORnz_def)
  apply (rule normxor_normed_id)
  apply (rule normed.Xor)
  apply auto
done

subsection {* Equivalence Relation $=_E$ on Messages *}

inductive
  xor_eq :: "fmsg => fmsg => bool" ("_ \<approx> _" [60,60])
 where
  Xor_assoc[intro]:  "(XOR X (XOR Y Z)) \<approx> (XOR (XOR X Y) Z)" |
  Xor_com[intro]:    "XOR X Y \<approx> XOR Y X" |
  Xor_Zero[intro]:   "XOR X ZERO \<approx> X" |
  Xor_cancel[intro]: "X \<approx> Y ==> XOR X Y \<approx> ZERO" |

  MPair_cong: "\<lbrakk> X \<approx> A ; Y \<approx> B \<rbrakk> \<Longrightarrow> MPAIR X Y \<approx> MPAIR A B" |
  Hash_cong:   "X \<approx> Y ==> HASH X \<approx> HASH Y" |
  Crypt_cong:  "M \<approx> N ==> CRYPT K M \<approx> CRYPT K N" |
  Xor_cong:   "\<lbrakk> X \<approx> A ; Y \<approx> B \<rbrakk> \<Longrightarrow> XOR X Y \<approx> XOR A B" |

  refl[intro]:        "X \<approx> X" |
  symm:         "X \<approx> Y ==> Y \<approx> X" |
  trans:       "[| X \<approx> Y; Y \<approx> Z |] ==> X \<approx> Z"

lemmas Xor_assoc_trans = xor_eq.Xor_assoc [THEN xor_eq.trans]
lemmas Xor_assoc_trans2 = xor_eq.Xor_assoc [THEN symm, THEN xor_eq.trans]
lemmas Xor_com_trans = xor_eq.Xor_com [THEN xor_eq.trans]
lemmas Xor_cong_trans = xor_eq.Xor_cong [THEN xor_eq.trans]

subsection {* Simplification Rules for normxor *}

lemma normxor_cancel[simp]: "x \<otimes> x = ZERO"
  apply (induct x)
  apply auto
done

lemma normxor_simp1[simp]:
  "\<lbrakk> normed a; normed b; standard a; a < first b; b \<noteq> ZERO \<rbrakk>
   \<Longrightarrow> a \<otimes> b = XOR a b"
  apply (induct b, auto simp add: normxor_standard XORnz_def)
  apply (frule normed_xor_snd)
  apply (frule normed_xor_fst_standard)
  apply simp
  apply (frule normed_xor_snd)
  apply (frule normed_xor_fst_standard)
  apply simp
done

lemma case_zero[simp]: "f \<noteq> ZERO \<Longrightarrow> (case f of ZERO \<Rightarrow> fzero | _ \<Rightarrow> fnonzero) = fnonzero"
  apply (case_tac f, auto)
done

lemma Xor_zero_fst[intro]: "ZERO \<oplus> x \<approx> x"
  apply (rule Xor_com_trans, auto)
done

lemma normxor_simp2[simp]:
  "\<lbrakk>normed a; normed b; standard a; a < first b; b \<noteq> ZERO\<rbrakk>
  \<Longrightarrow> b \<otimes> a = a \<oplus> b"
by (simp add: normxor_com)

lemma normxor_XORnz[simp]:
  "\<lbrakk> standard a; a < first b\<rbrakk> \<Longrightarrow> a \<otimes> b = a \<odot> b"
  apply (case_tac "standard b")
  apply (auto simp add: normxor_standard)
  apply (force simp add: XORnz_def)
  apply (case_tac b)
  apply force+ defer
  apply (force simp add: XORnz_def)
  apply (auto simp add: normxor_standard XORnz_def first_def)
done

lemma normxor_XORnz2[simp]:
  "\<lbrakk> standard a; standard c; c < a \<rbrakk> \<Longrightarrow> (a \<odot> b) \<otimes> c = c \<odot> (a \<odot> b)"
  apply (case_tac "standard b")
  apply (auto simp add: normxor_standard XORnz_def)
done


lemma normxor_simp3[simp]:
  "\<lbrakk> c1 < first b2; b2 \<otimes> c2 = ZERO; standard c1; b2 \<noteq> ZERO \<rbrakk>
   \<Longrightarrow> b2 \<otimes> c1 \<oplus> c2 = c1"
  apply (case_tac "standard b2")
    apply (force simp add: normxor_standard XORnz_def)
  apply (case_tac b2, auto)
  apply (auto simp add: first_def simp add: normxor_standard XORnz_def)
done

lemma normxor_simp4[simp]:
  "\<lbrakk> a < first c \<or> c = ZERO; standard a; b \<noteq> ZERO \<rbrakk>
  \<Longrightarrow> c \<otimes> (a \<oplus> b) = a \<odot> (c \<otimes> b)"
  apply (case_tac "standard c")
  apply (auto simp add: normxor_standard)
  apply (case_tac c)
  apply (auto simp add: normxor_standard XORnz_def first_def)
done

lemma normxor_simp5[simp]:
" \<lbrakk> standard a \<rbrakk> \<Longrightarrow>
  (a \<oplus> b) \<otimes> (a \<odot> c) = b \<otimes> c"
  apply (auto simp add: normxor_standard XORnz_def)
done

lemma normxor_simp6[simp]:
  "\<lbrakk> b < first a \<or> a = ZERO; standard b \<rbrakk>
  \<Longrightarrow> a \<otimes> b = b \<odot> a"
  apply (case_tac "standard a")
  apply (auto simp add: normxor_standard XORnz_def)
  apply (case_tac a)
  apply (auto simp add: normxor_standard XORnz_def first_def)
done

lemma normxor_simp7[simp]:
" \<lbrakk> standard a; standard c; c < a \<rbrakk> \<Longrightarrow>
  (a \<oplus> b) \<otimes> (c \<odot> d) = c \<odot> ((a \<oplus> b) \<otimes> d)"
  apply (auto simp add: normxor_standard XORnz_def)
done

lemma normxor_simp8[simp]:
  "\<lbrakk> standard a; a < first c \<or> c = ZERO \<rbrakk>
  \<Longrightarrow> c \<otimes> (a \<odot> b) = a \<odot> (c \<otimes> b)"
  apply (case_tac "standard c")
  apply (auto simp add: normxor_standard)
  apply (case_tac c)
  apply (auto simp add: normxor_standard XORnz_def first_def)
done

lemma normxor_simp9[simp]:
" \<lbrakk> standard a; standard c; a < c \<rbrakk> \<Longrightarrow>
  (a \<oplus> b) \<otimes> (c \<odot> d) = a \<odot> (b \<otimes> (c \<odot> d))"
  apply (auto simp add: normxor_standard XORnz_def)
done

lemma normxor_simp10[simp]:
" \<lbrakk> standard a; standard c; c < a \<rbrakk> \<Longrightarrow>
  (c \<odot> d) \<otimes> (a \<oplus> b) = c \<odot> (d \<otimes> (a \<oplus> b))"
  apply (auto simp add: normxor_standard XORnz_def)
done

lemma normxor_simp11[simp]:
" \<lbrakk> standard a \<rbrakk> \<Longrightarrow>
  (a \<oplus> b) \<otimes> (a \<oplus> c) = b \<otimes> c"
  apply (auto simp add: normxor_standard XORnz_def)
done

lemma normxor_simp12[simp]:
" \<lbrakk> standard a; standard c; a < c \<rbrakk> \<Longrightarrow>
  (a \<oplus> b) \<otimes> (c \<odot> d) = a \<odot> (b \<otimes> (c \<odot> d))"
  apply (auto simp add: normxor_standard XORnz_def)
done

lemma normxor_simp13[simp]:
 "\<lbrakk> standard a \<rbrakk> \<Longrightarrow> (a \<odot> b) \<otimes> a = b"
  apply (auto simp add: normxor_standard XORnz_def)
done

lemma normxor_simp14[simp]:
 "\<lbrakk> standard a; standard c; c < a \<rbrakk> \<Longrightarrow> (a \<odot> b) \<otimes> c = c \<odot> (a \<odot> b)"
  apply (auto simp add: normxor_standard XORnz_def)
done

lemma XORnz_left: "b = c \<Longrightarrow> a \<odot> b = a \<odot> c"
  apply (auto simp add: XORnz_def)
done

lemma XORnz_nonzero[simp]: "a \<odot> (b \<oplus> c) = a \<oplus> (b \<oplus> c)"
  apply (auto simp add: XORnz_def)
done

lemma XORnz_nonzero2[simp]: "b \<noteq> ZERO \<Longrightarrow> a \<odot> (b \<odot> c) = a \<oplus> (b \<odot> c)"
  apply (auto simp add: XORnz_def)
done

lemma XORnz_nonzero3[simp]: "b \<noteq> ZERO \<Longrightarrow> a \<odot> b = a \<oplus> b"
  apply (auto simp add: XORnz_def)
done

lemma XORnz_zero[simp,intro]:
  "a \<noteq> ZERO \<Longrightarrow> a \<odot> c \<noteq> ZERO"
  apply (auto simp add: XORnz_def)
done

subsection {* Reduced Message represent Equivalence Classes *}
text {* new induction principle *}

lemma normed_induct2  [consumes 1, case_names Zero Standard Xor]:
 "\<lbrakk>normed x; P ZERO;
   !! x. \<lbrakk>normed x; standard x\<rbrakk> \<Longrightarrow> P (x);
   !! a b. \<lbrakk>normed a; P a; standard a; normed b; P b; a < first b; b \<noteq> ZERO\<rbrakk> \<Longrightarrow> P (XOR a b)\<rbrakk>
  \<Longrightarrow> P x"
proof (induct x rule: normed.induct)
  case (Agent d)
  show ?case using prems by (auto intro: prems(2))
next
  case (Real d)
  show ?case using prems by (auto intro: prems(2))
next
  case (Number d)
  show ?case using prems by (auto intro: prems(2))
next
  case (Key d)
  show ?case using prems by (auto intro: prems(2))
next
  case (Nonce n k)
  show ?case using prems by (auto intro: prems(2))
next
  case (Hash h)
  show ?case using prems by (auto intro: prems(4))
next
  case (Crypt m k)
  show ?case using prems by (auto intro: prems(4))
next
  case (MPair a b)
  show ?case using prems by (auto intro: prems(6))
next
  case (Xor a b)
  show ?case using prems by auto
qed

(* this rule loops in most cases, so use with care *)
lemma normed_XOR2:
  "\<lbrakk>normed (a \<oplus> b);
    \<lbrakk>normed a; standard a; normed b; a < first b; b \<noteq> ZERO; normed (a \<oplus> b)\<rbrakk> \<Longrightarrow> P\<rbrakk>
  \<Longrightarrow> P"
  apply auto
  apply (erule normed_XOR)
  apply auto
done

lemma normxor_simp8_standard[simp]:
  "\<lbrakk> standard a; standard c; a < c \<rbrakk>
  \<Longrightarrow> c \<otimes> (a \<odot> b) = a \<odot> (c \<otimes> b)"
  apply (auto simp add: normxor_standard XORnz_def)
done

lemma normxor_simp5_com[simp]:
" \<lbrakk> standard a \<rbrakk> \<Longrightarrow>
  (a \<odot> c) \<otimes> (a \<oplus> b) = c \<otimes> b"
  apply (auto simp add: normxor_standard XORnz_def normxor_com)
done

lemma normxor_simp13_com[simp]:
 "\<lbrakk> standard a \<rbrakk> \<Longrightarrow> a \<otimes> (a \<odot> b) = b"
  apply (auto simp add: normxor_standard XORnz_def)
done

lemma normxor_simp14_com[simp]:
 "\<lbrakk> standard a; standard c; c < a \<rbrakk> \<Longrightarrow> c \<otimes> (a \<odot> b) = c \<odot> (a \<odot> b)"
  apply (auto simp add: normxor_standard XORnz_def)
done

lemma normxor_simp12_com[simp]:
" \<lbrakk> standard a; standard c; a < c \<rbrakk> \<Longrightarrow>
   (c \<odot> d) \<otimes> (a \<oplus> b) = a \<odot> ((c \<odot> d) \<otimes> b)"
  apply (auto simp add: normxor_standard XORnz_def)
done

lemma normxor_assoc2_s_s_x:
  assumes "normed a" and "standard a"
  and     "normed b" and "standard b"
  and     "normed (c1 \<oplus> c2)"
  and     "(a \<otimes> b) \<otimes> c1 = a \<otimes> (b \<otimes> c1)"
  and     "(a \<otimes> b) \<otimes> c2 = a \<otimes> (b \<otimes> c2)"
  shows "(a \<otimes> b) \<otimes> (c1 \<oplus> c2) = a \<otimes> (b \<otimes> (c1 \<oplus> c2))"
proof cases
  assume "a = b"
  hence R: "(a \<otimes> b) \<otimes> (c1 \<oplus> c2) = c1 \<oplus> c2" using prems by auto
  show ?thesis proof cases
    assume "b=c1"
    thus ?thesis using prems
      by (auto simp add: normxor_normed_id  normxor_standard)
  next
    assume "b\<noteq>c1"
    show ?thesis proof cases
      assume "b<c1"
      hence "a \<otimes> (b \<otimes> (c1 \<oplus> c2)) = a \<otimes> (b \<oplus> (c1 \<oplus> c2))" using prems
	by (auto simp add: normxor_normed_id  normxor_standard)
      also have "... = c1 \<oplus> c2" using prems
	by (auto simp add: normxor_normed_id  normxor_standard)
      finally show ?thesis using R by auto
    next
      assume "\<not> b<c1"
      hence "a \<otimes> (b \<otimes> (c1 \<oplus> c2)) = b \<otimes> (c1 \<odot> (b \<otimes> c2))" using prems
	by (auto simp add: normxor_normed_id  normxor_standard)
      also have "... =  c1 \<odot> c2" using prems apply -
	apply (erule normed_XOR2)
	by simp
      also have "... = c1 \<oplus> c2" using prems apply -
	apply (erule normed_XOR2) by auto
      finally show ?thesis using R by auto
    qed
  qed
next
  assume anb: "a \<noteq> b"
  thus ?thesis
  proof cases
    assume "a < b"
    show ?thesis proof cases
      assume "b=c1"
      thus ?thesis using prems apply -
	by (erule normed_XOR2, auto simp add: normxor_standard split: split_if_asm)
    next
      assume "b\<noteq>c1"
      show ?thesis proof cases
	assume "b < c1"
	thus ?thesis using prems apply -
	  by (erule normed_XOR2, auto simp add: normxor_standard split: split_if_asm)
      next
	assume "\<not> b < c1"
	hence le: "c1 < b" using prems by auto
	show ?thesis proof cases
	  assume "a=c1"
	  thus ?thesis using prems le apply -
	    by (auto simp add: normxor_standard split: split_if_asm)
	next
	  assume "a\<noteq>c1"
	  show ?thesis proof cases
	    assume "a < c1"
	    thus ?thesis using prems le apply -
	      apply (auto simp add: normxor_standard split: split_if_asm)
	      apply (erule normed_XOR2, auto simp add: normxor_com)
	      done
	  next
	    assume "\<not> a < c1"
	    hence "c1 < a" using prems by auto
	    thus ?thesis using prems le apply -
	      apply (auto simp add: normxor_standard split: split_if_asm)
	      apply (erule normed_XOR2, auto simp add: normxor_com)
	      done
	  qed
	qed
      qed
    qed
  next
    assume "\<not> (a < b)"
    hence "b < a" using anb by auto
    thus ?thesis using prems
      apply (auto simp add: normxor_standard)
      apply (erule normed_XOR2, auto simp add:  normxor_standard)
      apply (erule normed_XOR2, auto simp add:  normxor_standard)
      apply (erule normed_XOR2, auto simp add:  normxor_standard)
      done
  qed
qed

lemma normxor_assoc2_x_s_s:
  assumes "normed a" and "standard a"
  and     "normed b" and "standard b"
  and     "normed (c1 \<oplus> c2)"
  and     "(c1 \<otimes> b) \<otimes> a = c1 \<otimes> (b \<otimes> a)"
  and     "(c2 \<otimes> b) \<otimes> a = c2 \<otimes> (b \<otimes> a)"
  shows "((c1 \<oplus> c2) \<otimes> b) \<otimes> a = (c1 \<oplus> c2) \<otimes> (b \<otimes> a)"
proof -
  have "((c1 \<oplus> c2) \<otimes> b) \<otimes> a = a \<otimes> ((c1 \<oplus> c2) \<otimes> b)" by (auto simp add: normxor_com)
  also have "... =  a \<otimes> (b \<otimes> (c1 \<oplus> c2))" by (auto simp add: normxor_com)
  also have "... = (a \<otimes> b) \<otimes> (c1 \<oplus> c2)" using prems apply -
   apply (rule normxor_assoc2_s_s_x[THEN sym])
   apply force apply force apply force apply force apply force
   by (auto simp only: normxor_com)
  also have "... = (c1 \<oplus> c2) \<otimes> (b  \<otimes> a)" by (auto simp add: normxor_com)
  finally show "((c1 \<oplus> c2) \<otimes> b) \<otimes> a = (c1 \<oplus> c2) \<otimes> (b \<otimes> a)" by auto
qed

lemma normxor_assoc2_s_x_s:
  assumes "normed a" and "standard a"
  and     "normed (b1 \<oplus> b2)"
  and     "normed c" and "standard c"
  and     "(a \<otimes> b1) \<otimes> c = a \<otimes> (b1 \<otimes> c)"
  and     "(a \<otimes> b2) \<otimes> c = a \<otimes> (b2 \<otimes> c)"
  shows "(a \<otimes> (b1 \<oplus> b2)) \<otimes> c = a \<otimes> ((b1 \<oplus> b2) \<otimes> c)"
proof cases
  assume "a = b1"
  have A: "(a \<otimes> (b1 \<oplus> b2)) \<otimes> c = b2 \<otimes> c" using prems by (auto simp add: normxor_standard)
  thus ?thesis
  proof cases
    assume "b1=c"
    thus ?thesis using prems A apply -
      by (auto simp add: normxor_normed_id normxor_standard normxor_com split: split_if_asm)
  next
    assume b1nc: "b1\<noteq>c"
    show ?thesis using prems
      apply (case_tac "b1 < c")
        apply (force simp add: normxor_standard XORnz_def)
      apply (auto intro: normxor_simp2 elim: normed_XOR simp add: normxor_standard XORnz_def)
      done
  qed
next
  assume anb: "a \<noteq> b1"
  thus ?thesis
  proof cases
    assume "a < b1"
    thus ?thesis using prems
      apply auto
      apply (auto intro: normxor_simp2 elim: normed_XOR simp add: normxor_standard XORnz_def)
      done
  next
    assume nab1: "\<not> (a < b1)"
    have sb1: "standard b1" using `normed (b1 \<oplus> b2)` apply (rule normed_XOR) .
    hence b1lea: "b1 < a" using anb nab1 by auto
    thus ?thesis
    proof cases
      assume "b1=c"
      thus ?thesis using prems by (auto simp add: normxor_standard XORnz_def)
    next
      assume b1nc: "b1\<noteq>c"
      show ?thesis proof cases
	assume "b1 < c"
	thus ?thesis using prems
	  apply (auto simp add: normxor_standard)
	  apply (auto simp add: normxor_com intro: normxor_XORnz2 normxor_XORnz elim: normed_XOR)
	  done	
      next
	assume "\<not> b1 < c"
	hence "c < b1" using b1nc by auto
	thus ?thesis using prems
	  apply (auto simp add: normxor_standard)
	  apply (rule normxor_XORnz2)
	  apply (auto elim: normed_XOR)
	  done
      qed
    qed
  qed
qed

lemma normxor_simp4_com[simp]:
  "\<lbrakk> a < first c \<or> c = ZERO; standard a; b \<noteq> ZERO \<rbrakk>
  \<Longrightarrow> (a \<oplus> b) \<otimes> c = a \<odot> (b \<otimes> c)"
  apply (case_tac "standard c")
  apply (auto simp add: normxor_standard)
  apply (case_tac c)
  apply (auto simp add: normxor_standard XORnz_def first_def)
done

lemma normxor_assoc2_x_x_x:
  assumes  a1_assoc: "!!B C. \<lbrakk> normed B; normed C \<rbrakk> \<Longrightarrow> (a1 \<otimes> B) \<otimes> C = a1 \<otimes> B \<otimes> C"
  and      a2_assoc: "!!B C. \<lbrakk> normed B; normed C \<rbrakk> \<Longrightarrow> (a2 \<otimes> B) \<otimes> C = a2 \<otimes> B \<otimes> C"
  and      b1_assoc: "!!C. normed C \<Longrightarrow> ((a1 \<oplus> a2) \<otimes> b1) \<otimes> C = (a1 \<oplus> a2) \<otimes> (b1 \<otimes> C)"
  and      b2_assoc: "!!C. normed C \<Longrightarrow> ((a1 \<oplus> a2) \<otimes> b2) \<otimes> C = (a1 \<oplus> a2) \<otimes> (b2 \<otimes> C)"
  and      c1_assoc: "  ((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> c1
                       = (a1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> c1)"
  and      c2_assoc: "  ((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> c2
                       = (a1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> c2)"
  and     "normed (a1 \<oplus> a2)"
  and     "normed (b1 \<oplus> b2)"
  and     "normed (c1 \<oplus> c2)"
  shows "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2) = 
         (a1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> (c1 \<oplus> c2))"
proof -
  have sa1: "standard a1" using `normed (a1 \<oplus> a2)` by (rule normed_XOR)
  have sb1: "standard b1" using `normed (b1 \<oplus> b2)` by (rule normed_XOR)
  have sc1: "standard c1" using `normed (c1 \<oplus> c2)` by (rule normed_XOR)
  show ?thesis proof cases
    assume "a1=b1"
    show ?thesis proof cases
      assume "b1=c1"
      hence "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2) = (a2 \<otimes> b2) \<otimes> (c1 \<oplus> c2)" using prems by auto
      also have "... = c1 \<odot> ((a2 \<otimes> b2) \<otimes> c2)" using prems(7-) apply -
	apply (erule normed_XOR2)
	apply (auto simp add: normxor_standard normxor_com)
	apply (subst normxor_simp4_com)
	apply auto
	apply (drule_tac x=a2 and y=b2 in normxor_first)
	apply (auto elim: normed_XOR simp add: min_def split: split_if_asm)
	done
      finally have R: "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2) = c1 \<odot> ((a2 \<otimes> b2) \<otimes> c2)"
	by auto
      have "(a1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> (c1 \<oplus> c2)) = (a1 \<oplus> a2) \<otimes> (b2 \<otimes> c2)" using prems
	by auto
      also have "... = c1 \<odot> (a2 \<otimes>  (b2 \<otimes> c2))" using prems(7-) apply -
	apply (erule normed_XOR2) apply (erule normed_XOR2) apply (erule normed_XOR2)
	apply (auto simp add: normxor_standard normxor_com)
	apply (subst normxor_simp4_com)
	apply auto
	apply (drule_tac x=b2 and y=c2 in normxor_first)
	apply (auto elim: normed_XOR simp add: min_def split: split_if_asm)
	done
      also have "... = c1 \<odot> ((a2 \<otimes> b2) \<otimes> c2)" using prems(7-) apply -
	apply (drule normed_xor_snd)+
	by (simp only: a2_assoc)
      finally show ?thesis using R by simp
    next
      assume "b1 \<noteq> c1"
      show ?thesis proof cases
	assume "b1 < c1"
	show ?thesis using prems(7-) apply -
	  apply (frule normed_xor_fst_standard) back
	  apply (auto simp add: normxor_standard a2_assoc elim: normed_XOR)
	  done
      next
	assume "\<not> b1 < c1"
	hence c1leb1: "c1 < b1" using `b1 \<noteq> c1` by auto
	hence "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2) = a2 \<otimes> b2 \<otimes> c1 \<oplus> c2"
	  using prems by (auto dest: normed_xor_snd)
	also have "... = a2 \<otimes> (c1 \<odot> (b2 \<otimes> c2))" using prems
	  apply (subst normxor_simp4)
	  by (auto elim: normed_XOR)
	also have "... = c1 \<odot> (a2 \<otimes> (b2 \<otimes> c2))" using prems `a1=b1`
	  apply (subst normxor_simp8)
	  apply (auto elim: normed_XOR)
	  done
	finally have L1: "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2) =
	                  c1 \<odot> (a2 \<otimes> (b2 \<otimes> c2))" by auto
	have "(a1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> (c1 \<oplus> c2)) =
	      (b1 \<oplus> a2) \<otimes> c1 \<odot> (b1 \<oplus> b2) \<otimes> c2" using prems
	  by (auto simp add: normxor_standard)
	also have "... = c1 \<odot> (b1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> c2)" using prems(7-)
	  apply (subst normxor_simp8)
	  by (auto simp add: normxor_com elim: normed_XOR)
	finally have R1: "(a1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> (c1 \<oplus> c2)) =
	                  c1 \<odot> (b1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> c2)" by auto
	have "(a2 \<otimes> (b2 \<otimes> c2)) = (b1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> c2)"
	proof cases
	  assume "b1 < first c2"
	  have "(b1 \<oplus> b2) \<otimes> c2 = c2 \<otimes> (b1 \<oplus> b2)" by (simp add: normxor_com)
	  also have "... = b1 \<odot> (c2 \<otimes> b2)" using prems
	    apply (subst normxor_simp4)
	    by (auto simp add: normxor_com elim: normed_XOR)
	  finally have A: "(b1 \<oplus> b2) \<otimes> c2 = b1 \<odot> (c2 \<otimes> b2)" by auto
	  have "(b1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> c2) = (b1 \<oplus> a2) \<otimes> (b1 \<odot> (c2 \<otimes> b2))"
	    using A by auto
	  also have "... = a2 \<otimes> (c2 \<otimes> b2)" using prems(7-)
	    apply (subst normxor_simp5)
	    by (auto elim: normed_XOR)
	  finally show ?thesis by (auto simp add: normxor_com)
	next
	  assume nb1lec2: "\<not> b1 < first c2"
	  show ?thesis proof cases
	    assume "b1 = first c2"
	    have "(b1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> c2) = ((b1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> c2" using prems
	      by (simp only: c2_assoc)
	    thus ?thesis using prems apply simp apply (drule normed_xor_snd[where b=b2])
	      apply (drule normed_xor_snd[where b=c2])
	      apply simp
	      done
	  next
	    assume "b1 \<noteq> first c2"
	    hence " first c2 < b1" using nb1lec2 by auto
	    thus ?thesis using prems
	      apply (case_tac "standard c2")
	      by (auto dest!: normed_xor_snd)
	  qed
	qed
	thus ?thesis using L1 R1 by simp
      qed
    qed
  next
    assume a1nb1: "a1\<noteq>b1"
    show ?thesis proof cases
      assume "a1 < b1"
      show ?thesis proof cases
	assume "b1 = c1"
	have R1: "(a1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> (c1 \<oplus> c2)) =
	            (a1 \<oplus> a2) \<otimes> b2 \<otimes> c2" using prems(7-)
	  by (auto simp add: normxor_standard elim: normed_XOR)
	have "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2) =
   	      (a1 \<odot> (a2 \<otimes> (c1 \<oplus> b2))) \<otimes> (c1 \<oplus> c2)" using prems(7-)
	  by (auto simp add: normxor_standard elim: normed_XOR)
	also have "... = a1 \<odot> ((a2 \<otimes> (c1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2))" using prems(7-)
	  apply (subst normxor_simp10)
	  by (auto elim: normed_XOR)
	also have "... = a1 \<odot> (a2 \<otimes> ((c1 \<oplus> b2) \<otimes> (c1 \<oplus> c2)))"
	  using `normed (b1 \<oplus> b2)` `normed (c1 \<oplus> c2)` `b1=c1`
	  by (auto simp add: a2_assoc)
	also have "... = a1 \<odot> (a2 \<otimes> (b2 \<otimes> c2))" using prems(7-)
	  apply (subst normxor_simp11)
	  by (auto elim: normed_XOR)
	also have "... = (b2 \<otimes> c2) \<otimes> (a1 \<oplus> a2)" using prems(7-)
	  apply (subst normxor_simp4) prefer 3
	  apply (force elim: normed_XOR) prefer 2
	  apply (force elim: normed_XOR) prefer 2
	  apply (force simp add: normxor_com)
	  apply auto
	  apply (frule normed_xor_snd) back
	  apply (drule_tac x=b2 and y=c2 in normxor_first)
	  apply (force elim: normed_XOR)
	  apply (force elim: normed_XOR)
	  apply (subgoal_tac "c1 < first b2 \<and> c1 < first c2")
	  apply (auto simp add: min_def split: split_if_asm)
	  done
	also have "... = (a1 \<oplus> a2) \<otimes> b2 \<otimes> c2"
	  by (auto simp add: normxor_com)
	finally show ?thesis using R1 by simp
      next
	assume "b1 \<noteq> c1"
	show ?thesis proof cases
	  assume "b1 < c1"
	  have "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2) =
   	        (a1 \<odot> a2 \<otimes> b1 \<oplus> b2) \<otimes> c1 \<oplus> c2" using prems(7-)
	    by (auto simp add: normxor_standard elim: normed_XOR)
	  also have "... =  (c1 \<oplus> c2) \<otimes> (a1 \<odot> a2 \<otimes> b1 \<oplus> b2)" using prems(7-)
	    by (auto simp add: normxor_com)
	  also have "... = a1 \<odot> ((c1 \<oplus> c2) \<otimes> (a2 \<otimes> (b1 \<oplus> b2)))" using prems(7-)
	    apply (subst normxor_simp8)
	    by (auto elim: normed_XOR)
	  also have "... = a1 \<odot> ((a2 \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2))" using prems(7-)
	    by (auto simp add: normxor_com)
	  also have "... = a1 \<odot> (a2 \<otimes> ((b1 \<oplus> b2) \<otimes> (c1 \<oplus> c2)))"
	    using `normed (b1 \<oplus> b2)` `normed (c1 \<oplus> c2)`
	    by (auto simp add: a2_assoc)
	  also have "... = a1 \<odot> ((a2 \<otimes> (b1 \<odot> b2 \<otimes> (c1 \<oplus> c2))))" using prems(7-)
	    by (auto simp add: normxor_standard elim: normed_XOR)
	  finally have L1: "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2) =
	                     a1 \<odot> ((a2 \<otimes> (b1 \<odot> b2 \<otimes> (c1 \<oplus> c2))))" by auto

          have "(a1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> (c1 \<oplus> c2)) =
                 (a1 \<oplus> a2) \<otimes> b1 \<odot> b2 \<otimes> c1 \<oplus> c2" using prems(7-)
	    by (auto simp add: normxor_standard elim: normed_XOR)
	  also have "... = a1 \<odot> (a2 \<otimes> (b1 \<odot> b2 \<otimes> (c1 \<oplus> c2)))" using prems(7-)
	    apply (subst normxor_simp9)
	    by (auto elim: normed_XOR)
	  finally show ?thesis using L1 by auto
	next
	  assume "\<not> b1 < c1"
	  hence c1leb1: "c1 < b1" using prems(7-) by auto
	  show ?thesis proof cases
	    assume "a1=c1"
	    have "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2) =
     	          (c1 \<odot> (a2 \<otimes> (b1 \<oplus> b2))) \<otimes> (c1 \<oplus> c2)" using prems(7-) c1leb1
	      by (auto simp add: normxor_standard elim: normed_XOR)
 	    also have "... = (c1 \<oplus> c2) \<otimes> (c1 \<odot> (a2 \<otimes> (b1 \<oplus> b2)))"
	      by (auto simp add: normxor_com)
	    also have "... = c2 \<otimes> (a2 \<otimes> (b1 \<oplus> b2))" using prems(7-) c1leb1
	      apply (subst normxor_simp5)
	      by (auto elim: normed_XOR)
	    finally have R: "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2) =
	                     c2 \<otimes> (a2 \<otimes> (b1 \<oplus> b2))" by auto

	    have "(a1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> (c1 \<oplus> c2)) =
	          (c1 \<oplus> a2) \<otimes> (c1 \<odot> ((b1 \<oplus> b2) \<otimes> c2))" using prems(7-) c1leb1
	      by (auto simp add: normxor_standard elim: normed_XOR)
	    also have "... = a2 \<otimes> ((b1 \<oplus> b2) \<otimes> c2)" using prems(7-) c1leb1
	      apply (subst normxor_simp5)
	      by (auto elim: normed_XOR)
	    also have "... = (a2 \<otimes> (b1 \<oplus> b2)) \<otimes> c2"
	      using `normed (b1 \<oplus> b2)` `normed (c1 \<oplus> c2)` apply -
	      apply (drule normed_xor_snd[where b=c2])
	      by (simp only: a2_assoc)
	    also have "... = c2 \<otimes>  (a2 \<otimes> (b1 \<oplus> b2))" by (simp only: normxor_com)
	    finally show ?thesis using R by simp
	  next
	    assume "a1\<noteq>c1"
	    show ?thesis proof cases
	      assume "a1 < c1"
	      have "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2) =
    	            (a1 \<odot> (a2 \<otimes> (b1 \<oplus> b2))) \<otimes> (c1 \<oplus> c2)" using prems(7-) c1leb1
		by (auto simp add: normxor_standard elim: normed_XOR)
	      also have "... = (c1 \<oplus> c2) \<otimes> (a1 \<odot> (a2 \<otimes> (b1 \<oplus> b2)))"
		by (simp only: normxor_com)
	      also have "... = a1 \<odot> ((c1 \<oplus> c2) \<otimes> (a2 \<otimes> (b1 \<oplus> b2)))"
		using prems(7-) c1leb1
		apply (subst normxor_simp8)
		by (auto elim: normed_XOR)
	      also have "... = a1 \<odot> ((a2 \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2))"
		by (simp only: normxor_com)
	      also have "... = a1 \<odot> (a2 \<otimes> ((b1 \<oplus> b2) \<otimes> (c1 \<oplus> c2)))"
		using `normed (b1 \<oplus> b2)` `normed (c1 \<oplus> c2)`
		by (simp only: a2_assoc)
	      also have "... = a1 \<odot> (a2 \<otimes> (c1 \<odot> ((b1 \<oplus> b2) \<otimes> c2)))"
		using prems(7-) c1leb1
		by (auto simp add: normxor_standard)
	      finally have R: "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2) =
		               a1 \<odot> (a2 \<otimes> (c1 \<odot> ((b1 \<oplus> b2) \<otimes> c2)))" by auto
	      have "(a1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> (c1 \<oplus> c2)) =
	            (a1 \<oplus> a2) \<otimes> (c1 \<odot> ((b1 \<oplus> b2) \<otimes> c2))" using prems(7-) c1leb1
		by (auto simp add: normxor_standard elim: normed_XOR)
	      also have "... = a1 \<odot> (a2 \<otimes> (c1 \<odot> ((b1 \<oplus> b2) \<otimes> c2)))"
		using prems(7-) c1leb1
		apply (subst normxor_simp9)
		by (auto elim: normed_XOR)
	      finally show ?thesis using R by simp
	    next
	      assume "\<not> a1 < c1"
	      hence c1lea1: "c1 < a1" using prems(7-) by auto
	      have "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2) =
  		    (a1 \<odot> (a2 \<otimes> (b1 \<oplus> b2))) \<otimes> (c1 \<oplus> c2)"
		using prems(7-) c1lea1 c1leb1
		by (auto simp add: normxor_standard elim: normed_XOR)
	      also have "... = (c1 \<oplus> c2) \<otimes> (a1 \<odot> (a2 \<otimes> (b1 \<oplus> b2)))"
		by (simp only: normxor_com)
	      also have "... = c1 \<odot> (c2 \<otimes> (a1 \<odot> (a2 \<otimes> (b1 \<oplus> b2))))"
		using prems(7-) c1lea1 c1leb1
		apply (subst normxor_simp12)
		by (auto elim: normed_XOR)
	      finally have R: "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2) =
		               c1 \<odot> (c2 \<otimes> (a1 \<odot> (a2 \<otimes> (b1 \<oplus> b2))))" by simp
	      have "(a1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> (c1 \<oplus> c2)) =
		    (a1 \<oplus> a2) \<otimes> (c1 \<odot> ((b1 \<oplus> b2) \<otimes> c2))"
		using prems(7-) c1lea1 c1leb1
		by (auto simp add: normxor_standard elim: normed_XOR)
	      also have "... = c1 \<odot> ((a1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> c2))"
		using prems(7-) c1lea1 c1leb1
		apply (subst normxor_simp7)
		by (auto elim: normed_XOR)
	      also have "... = c1 \<odot> (((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> c2)"
		by (simp only: c2_assoc)
	      also have "... = c1 \<odot> ((a1 \<odot> (a2 \<otimes> (b1 \<oplus> b2))) \<otimes> c2)"
		using prems(7-) c1lea1 c1leb1
		by (auto simp add: normxor_standard)
	      also have "... = c1 \<odot> (c2 \<otimes> (a1 \<odot> (a2 \<otimes> (b1 \<oplus> b2))))"
		by (simp only: normxor_com)
	      finally show ?thesis using R by simp
	    qed
	  qed
	qed
      qed
    next
      assume "\<not> a1 < b1"
      hence "b1 < a1" using a1nb1 by auto
      show ?thesis proof cases
	assume "b1=c1"
	have R: "(a1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> (c1 \<oplus> c2)) =
	         (a1 \<oplus> a2) \<otimes> (b2 \<otimes> c2)" using prems(7-)
	  by (auto simp add: normxor_standard elim: normed_XOR)
	
	have "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2) =
   	      (c1 \<odot> (a1 \<oplus> a2) \<otimes> b2) \<otimes> (c1 \<oplus> c2)" using prems(7-)
	  by (auto simp add: normxor_standard elim: normed_XOR)
	also have "... = (c1 \<oplus> c2) \<otimes> (c1 \<odot> (a1 \<oplus> a2) \<otimes> b2)"
	  by (auto simp add: normxor_com)
	also have "... = c2 \<otimes> ((a1 \<oplus> a2) \<otimes> b2)" using prems(7-)
	  apply (subst normxor_simp5)
	  by (auto elim: normed_XOR)
	also have "... = (a1 \<oplus> a2) \<otimes> (b2 \<otimes> c2)"
	  using `normed (c1 \<oplus> c2)` apply -
	  apply (drule normed_xor_snd[where b=c2])
	  by (auto simp add: normxor_com b2_assoc[THEN sym])
	finally show ?thesis using R by auto
      next
	assume "b1 \<noteq> c1"
	show ?thesis proof cases
	  assume "b1 < c1"
	  have "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2)=
            (b1 \<odot> (a1 \<oplus> a2) \<otimes> b2) \<otimes> (c1 \<oplus> c2)" using prems(7-)
	    by (auto simp add: normxor_standard elim: normed_XOR)
	  also have "... = (c1 \<oplus> c2) \<otimes> (b1 \<odot> (a1 \<oplus> a2) \<otimes> b2)"
	    by (auto simp add: normxor_com)
	  also have "... = b1 \<odot> ((c1 \<oplus> c2) \<otimes> ((a1 \<oplus> a2) \<otimes> b2))" using prems(7-)
	    apply (subst normxor_simp8)
	    by (auto simp add: first_def elim: normed_XOR)
	  also have "... = b1 \<odot> ((a1 \<oplus> a2) \<otimes> (b2 \<otimes> (c1 \<oplus> c2)))" using `normed (c1 \<oplus> c2)` apply -
	    by (auto simp add: normxor_com b2_assoc[THEN sym])
	  finally have R: "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2) =
	                    b1 \<odot> ((a1 \<oplus> a2) \<otimes> (b2 \<otimes> (c1 \<oplus> c2)))" by auto
	  have "(a1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> (c1 \<oplus> c2)) =
               (a1 \<oplus> a2) \<otimes> (b1 \<odot> (b2 \<otimes> (c1 \<oplus> c2)))" using prems(7-)
	    by (auto simp add: normxor_standard elim: normed_XOR)
	  also have "... = b1 \<odot> ((a1 \<oplus> a2) \<otimes> (b2 \<otimes> (c1 \<oplus> c2)))" using prems(7-)
	    apply (subst normxor_simp8)
	    by (auto simp add: first_def elim: normed_XOR)
	  
	  finally show ?thesis using R by simp
	next
	  assume "\<not> b1 < c1"
	  hence c1leb1: "c1 < b1" using prems(7-) by auto
	  hence "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2) =
                  (b1 \<odot> (a1 \<oplus> a2) \<otimes> b2) \<otimes> c1 \<oplus> c2" using prems(7-)
	    by (auto simp add: normxor_standard elim: normed_XOR)
	  also have "... = c1 \<odot> ((b1 \<odot> (a1 \<oplus> a2) \<otimes> b2) \<otimes> c2)"
	    using prems(7-) c1leb1
	    apply (subst normxor_simp4) defer
	    apply (force elim: normed_XOR)
	    apply (force elim: normed_XOR)
	    apply force
	    apply (force elim: normed_XOR simp add: first_def XORnz_def)
	    done
	  finally have R: "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2) =
	                   c1 \<odot> ((b1 \<odot> (a1 \<oplus> a2) \<otimes> b2) \<otimes> c2)" by simp
	  have "(a1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> (c1 \<oplus> c2)) =
 	        (a1 \<oplus> a2) \<otimes> (c1 \<odot> ((b1 \<oplus> b2) \<otimes> c2))" using prems(7) c1leb1
	    by (auto simp add: normxor_standard elim: normed_XOR)
	  also have "... = c1 \<odot> ((a1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> c2))"
	    using prems(7-) c1leb1
	    apply (subst normxor_simp7)
	    by (auto elim: normed_XOR)
	  also have "... = c1 \<odot> (((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> c2)" using prems(7-)
	    by (simp only: c2_assoc)
	  also have "... = c1 \<odot> ((b1 \<odot> ((a1 \<oplus> a2) \<otimes> b2)) \<otimes> c2)"
	    using prems(7-) c1leb1
	    apply (auto simp add: normxor_standard)
	    done
	  finally show ?thesis using R by simp
	qed
      qed
    qed
  qed
qed


lemma normxor_assoc2_s_x_x:
  assumes  b1_assoc: "!!C. normed C \<Longrightarrow> (a \<otimes> b1) \<otimes> C = a \<otimes> (b1 \<otimes> C)"
  and      b2_assoc: "!!C. normed C \<Longrightarrow> (a \<otimes> b2) \<otimes> C = a \<otimes> (b2 \<otimes> C)"
  and      c1_assoc: "(a \<otimes> (b1 \<oplus> b2)) \<otimes> c1 = a \<otimes> ((b1 \<oplus> b2) \<otimes> c1)"
  and      c2_assoc: "(a \<otimes> (b1 \<oplus> b2)) \<otimes> c2 = a \<otimes> ((b1 \<oplus> b2) \<otimes> c2)"
  and     "normed a" and "standard a"
  and     "normed (b1 \<oplus> b2)"
  and     "normed (c1 \<oplus> c2)"
  shows "(a \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2) = a \<otimes> ((b1 \<oplus> b2) \<otimes> (c1 \<oplus> c2))"
proof -
  have sb1: "standard b1" using `normed (b1 \<oplus> b2)` by (rule normed_XOR)
  have sc1: "standard c1" using `normed (c1 \<oplus> c2)` by (rule normed_XOR)
  show ?thesis proof cases
    assume "a = b1"
    show ?thesis proof cases
      assume "b1 = c1"
      hence "(a \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2) = b2 \<otimes> (c1 \<oplus> c2)" using prems(5-)
	by (auto simp add: normxor_standard)
      also have "... = c1 \<odot> (b2 \<otimes> c2)" using prems(5-)
	apply (subst normxor_simp4)
	by auto
      finally have R: "(a \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2) = c1 \<odot> (b2 \<otimes> c2)" by auto
      have "a \<otimes> ((b1 \<oplus> b2) \<otimes> (c1 \<oplus> c2)) = c1 \<otimes> b2 \<otimes> c2" using prems(5-)
	by (auto simp add: normxor_standard)
      also have "... = c1 \<odot> (b2 \<otimes> c2)" using prems(5-) apply -
	apply (simp only: normxor_com[where x=c1])
	apply (subst normxor_simp6) back
	apply (auto elim: normed_XOR)
	apply (frule normed_xor_snd)
	apply (drule_tac x=b2 and y=c2 in normxor_first)
	apply (auto elim: normed_XOR simp add: `b1=c1` min_def split: split_if_asm)
	done
      finally show ?thesis using R by auto
    next
      assume "b1\<noteq>c1"
      show ?thesis proof cases
	assume "b1 < c1"
	thus ?thesis using prems(5-) by (auto simp add: normxor_standard)
      next
	assume "\<not> b1 < c1"
	hence le: "c1 < b1" using prems(5-) by auto
	have "(a \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2) = b2 \<otimes> c1 \<oplus> c2" using prems(5-)
	  by (auto simp add: normxor_standard)
	also have "... = c1 \<odot> (b2 \<otimes> c2)" using prems(5-)
	  apply (subst normxor_simp4)
	  by (auto elim: normed_XOR)
	finally have R: "(a \<otimes> (b1 \<oplus> b2)) \<otimes> (c1 \<oplus> c2) = c1 \<odot> (b2 \<otimes> c2)" by auto  
	have "a \<otimes> ((b1 \<oplus> b2) \<otimes> (c1 \<oplus> c2)) = b1 \<otimes> c1 \<odot> (b1 \<oplus> b2) \<otimes> c2" using prems(5-)
	  by (auto simp add: normxor_standard)
	also have "... = c1 \<odot> (b1 \<otimes> ((b1 \<oplus> b2) \<otimes> c2))" using prems(5-)
	  apply (subst normxor_simp8)
	  by (auto elim: normed_XOR)
	also have "... = c1 \<odot> ((b1 \<otimes> (b1 \<oplus> b2)) \<otimes> c2)"
	  by (simp only: c2_assoc[simplified `a=b1`])
	also have "... = c1 \<odot> (b2 \<otimes> c2)" using prems(5-) apply -
	  apply (frule normed_xor_fst_standard)
	  by (auto simp add: normxor_standard)
	finally show ?thesis using R by auto
      qed
    qed
  next
    assume "a\<noteq>b1"
    show ?thesis proof cases
      assume "a < b1"
      show ?thesis proof cases
	assume "b1=c1"
	thus ?thesis using prems(5-)
	  apply (auto simp add: normxor_standard)
	  apply (simp only: normxor_com[where x=a])
	  apply (subst normxor_simp6) back back
	  apply (auto elim: normed_XOR simp add: normxor_com)
	  apply (frule normed_xor_snd)
	  apply (frule_tac x=b2 and y=c2 in normxor_first)
	  apply (force elim: normed_XOR)
	  apply force
	  apply (rule xt1(8))
	  apply assumption
	  apply (auto elim: normed_XOR simp add: first_def)
	  done
      next
	assume "b1\<noteq>c1"
	show ?thesis proof cases
	  assume "b1 < c1"
	  thus ?thesis using prems(5-)
	    apply (auto simp add: normxor_standard)
	    apply (simp only: normxor_com[where x=a])
	    apply (subst normxor_simp6) back back
	    apply (auto elim: normed_XOR simp add: normxor_com)
	    apply (auto simp add: first_def XORnz_def)
	    apply (case_tac b1, auto)
	    done
	next
	  assume "\<not> b1 < c1"
	  hence "c1 < b1" using prems(5-) by auto
	  show ?thesis proof cases
	    assume "a=c1"
	    thus ?thesis using prems(5-) by (auto simp add: normxor_standard)
	  next
	    assume "a\<noteq>c1"
	    show ?thesis proof cases
	      assume "a < c1"
	      thus ?thesis using prems(5-)
		apply (auto simp add: normxor_standard)
		apply (simp only: normxor_com[where x=a])
		apply (subst normxor_simp6) back back
		apply (auto elim: normed_XOR simp add: normxor_com)
		apply (auto simp add: first_def XORnz_def)
		apply (case_tac c1, auto)
		done
	    next
	      assume "\<not> a < c1"
	      hence "c1 < a" using prems(5-) by auto
	      thus ?thesis using prems(5-)
		apply (auto simp add: normxor_standard)
		apply (subst normxor_simp8)
		apply (auto elim: normed_XOR simp add: normxor_com)
		apply (simp add: normxor_com[where x=c2])
		apply (simp only: c2_assoc[THEN sym])
		apply (auto simp add: normxor_standard)
		done
	    qed
	  qed
	qed
      qed
    next
      assume "\<not> a < b1"
      hence "b1 < a" using prems(5-) by auto
      show ?thesis proof cases
	assume "b1=c1"
	thus ?thesis using prems(5-)
	  apply (auto simp add: normxor_standard normxor_com)
	  apply (subst normxor_simp5)
	  apply (auto elim: normed_XOR)
	  apply (simp add: normxor_com[where x=c2])
	  apply (drule normed_xor_snd) back
	  apply (simp add: b2_assoc[THEN sym])
	  done
      next
	assume "b1\<noteq>c1"
	show ?thesis proof cases
	  assume "b1 < c1"
	  thus ?thesis using prems(5-)
	    apply (auto simp add: normxor_standard normxor_com)
	    apply (simp add: normxor_com[where x="c1 \<oplus> c2"])
	    apply (subst normxor_simp10)
	    apply (auto elim: normed_XOR)
	    apply (subst normxor_simp8)
	    apply (auto elim: normed_XOR)
	    apply (simp add: b2_assoc)
	    done
	next
	  assume "\<not> b1 < c1"
	  hence "c1 < b1" using prems(5-) by auto
	  thus ?thesis using prems(5-)
	    apply (auto simp add: normxor_standard normxor_com)
	    apply (subst normxor_simp9)
	    apply (auto elim: normed_XOR)
	    apply (subst normxor_simp8) back
	    apply (auto elim: normed_XOR)
	    apply (simp add: normxor_com[where y="b1 \<oplus> b2"])
	    apply (simp add: c2_assoc[THEN sym])
	    apply (auto simp add: normxor_standard normxor_com)
	    done
	qed
      qed
    qed
  qed
qed

lemma normxor_assoc2_x_s_x:
  assumes  a1_assoc: "!!B C. \<lbrakk> normed B; normed C \<rbrakk> \<Longrightarrow> (a1 \<otimes> B) \<otimes> C = a1 \<otimes> (B \<otimes> C)"
  and      a2_assoc: "!!B C. \<lbrakk> normed B; normed C \<rbrakk> \<Longrightarrow> (a2 \<otimes> B) \<otimes> C = a2 \<otimes> (B \<otimes> C)"
  and      c1_assoc: "((a1 \<oplus> a2) \<otimes> b) \<otimes> c1 = (a1 \<oplus> a2) \<otimes> (b \<otimes> c1)"
  and      c2_assoc: "((a1 \<oplus> a2) \<otimes> b) \<otimes> c2 = (a1 \<oplus> a2) \<otimes> (b \<otimes> c2)"
  and     "normed (a1 \<oplus> a2)"
  and     "normed b" and "standard b"
  and     "normed (c1 \<oplus> c2)"
  shows "((a1 \<oplus> a2) \<otimes> b) \<otimes> (c1 \<oplus> c2) = (a1 \<oplus> a2) \<otimes> (b \<otimes> (c1 \<oplus> c2))"
proof -
  have sa1: "standard a1" using `normed (a1 \<oplus> a2)` by (rule normed_XOR)
  have sc1: "standard c1" using `normed (c1 \<oplus> c2)` by (rule normed_XOR)
  show ?thesis proof cases
    assume "a1 = b"
    show ?thesis proof cases
      assume "b = c1"
      thus ?thesis using prems(5-) apply -
	apply (erule normed_XOR) apply (erule normed_XOR)
	apply (auto simp add: normxor_standard XORnz_def)
	done
    next
      assume "b\<noteq>c1"
      show ?thesis proof cases
	assume "b < c1"
	thus ?thesis using prems(5-) by (auto simp add: normxor_standard)
      next
	assume "\<not> b < c1"
	hence le: "c1 < b" using prems(5-) by auto
	thus ?thesis using prems(5-) apply -
	  apply (erule normed_XOR) apply (erule normed_XOR) 
	  apply (auto simp add: normxor_standard)
	  apply (simp only: c2_assoc[simplified `a1=b`, THEN sym])
	  by (auto simp add: normxor_standard)
      qed
    qed
  next
    assume "a1\<noteq>b"
    show ?thesis proof cases
      assume "a1 < b"
      show ?thesis proof cases
	assume "b=c1"
	thus ?thesis using prems(5-) apply -
	  apply (erule normed_XOR2) apply (erule normed_XOR2)
	  apply (auto simp add: normxor_standard)
	  apply (simp only: a2_assoc[simplified `b=c1`])
	  apply (auto simp add: normxor_standard)
	  done
      next
	assume "b\<noteq>c1"
	show ?thesis proof cases
	  assume "b < c1"
	  thus ?thesis using prems(5-) apply -
	    apply (erule normed_XOR2) apply (erule normed_XOR2)
	    apply (auto simp add: normxor_standard)
	    apply (auto simp add: a2_assoc normxor_standard)
	    done
	next
	  assume "\<not> b < c1"
	  hence "c1 < b" using prems(5-) by auto
	  show ?thesis proof cases
	    assume "a1=c1"
	    thus ?thesis using prems(5-)
	      apply (auto simp add: normxor_standard)
	      apply (subst normxor_simp5)
	      apply (auto elim: normed_XOR)
	      apply (simp only: normxor_com)
	      apply (subst normxor_simp5)
	      apply (auto elim: normed_XOR simp add: normxor_com[where x=c2])
	      apply (drule normed_xor_snd[where b=c2])
	      apply (auto simp add: a2_assoc)
	      done
	  next
	    assume "a1\<noteq>c1"
	    show ?thesis proof cases
	      assume "a1 < c1"
	      thus ?thesis using prems(5-)
		apply (auto simp add: normxor_standard)
		apply (subst normxor_simp9)
		apply (auto elim: normed_XOR)
		apply (simp only: normxor_com[where y="c1 \<oplus> c2"])
		apply (subst normxor_simp7) 
		apply (auto elim: normed_XOR)
		apply (simp only: normxor_com[where x="c1 \<oplus> c2"])
		apply (auto simp add: a2_assoc)
		by (auto simp add: normxor_standard)
	    next
	      assume "\<not> a1 < c1"
	      hence "c1 < a1" using prems(5-) by auto
	      thus ?thesis using prems(5-)
		apply (auto simp add: normxor_standard)
		apply (subst normxor_simp7)
		apply (auto elim: normed_XOR)
		apply (simp only: c2_assoc[THEN sym])
		apply (auto simp add: normxor_standard)
		apply (simp add: normxor_com[where y="c1 \<oplus> c2"])
		apply (subst normxor_simp9)
		apply (auto elim: normed_XOR simp add: normxor_com)
		done
	    qed
	  qed
	qed
      qed
    next
      assume "\<not> a1 < b"
      hence "b < a1" using prems(5-) by auto
      show ?thesis proof cases
	assume "b=c1"
	thus ?thesis using prems(5-)
	  apply (auto simp add: normxor_standard normxor_com)
	  done
      next
	assume "b\<noteq>c1"
	show ?thesis proof cases
	  assume "b < c1"
	  thus ?thesis using prems(5-)
	    apply (auto simp add: normxor_standard normxor_com)
	    done
	next
	  assume "\<not> b < c1"
	  hence "c1 < b" using prems(5-) by auto
	  thus ?thesis using prems(5-)
	    apply (auto simp add: normxor_standard normxor_com)
	    apply (subst normxor_simp7)
	    apply (auto elim: normed_XOR)
	    apply (simp add: c2_assoc[THEN sym])
	    apply (auto simp add: normxor_standard normxor_com)
	    done
	qed
      qed
    qed
  qed
qed


lemma normxor_assoc2_x_x_s:
  assumes  a1_assoc: "!!B C. \<lbrakk> normed B; normed C \<rbrakk> \<Longrightarrow> (a1 \<otimes> B) \<otimes> C = a1 \<otimes> B \<otimes> C"
  and      a2_assoc: "!!B C. \<lbrakk> normed B; normed C \<rbrakk> \<Longrightarrow> (a2 \<otimes> B) \<otimes> C = a2 \<otimes> B \<otimes> C"
  and      b1_assoc: "!!C. normed C \<Longrightarrow> ((a1 \<oplus> a2) \<otimes> b1) \<otimes> C = (a1 \<oplus> a2) \<otimes> (b1 \<otimes> C)"
  and      b2_assoc: "!!C. normed C \<Longrightarrow> ((a1 \<oplus> a2) \<otimes> b2) \<otimes> C = (a1 \<oplus> a2) \<otimes> (b2 \<otimes> C)"
  and     "normed (a1 \<oplus> a2)"
  and     "normed (b1 \<oplus> b2)"
  and     "normed c" and "standard c"
  shows "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> c = (a1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> c)"
proof -
  have sa1: "standard a1" using `normed (a1 \<oplus> a2)` by (rule normed_XOR)
  have sb1: "standard b1" using `normed (b1 \<oplus> b2)` by (rule normed_XOR)
  show ?thesis proof cases
    assume "a1=b1"
    hence A: "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> c = (a2 \<otimes> b2) \<otimes> c" using prems(5-) by auto
    show ?thesis proof cases
      assume "b1 = c"
      thus ?thesis using prems(5-)
	apply (auto simp add: normxor_standard)
	apply (subst normxor_simp6) back
	apply (auto elim: normed_XOR intro: normed_normxor)
	apply (frule normed_xor_snd)
	apply (drule_tac x=a2 and y=b2 in normxor_first)
	apply (auto elim: normed_XOR simp add: min_def split: split_if_asm)
	apply (simp only: normxor_com[where y=b2])
	apply (subst normxor_simp4)
	apply (auto elim: normed_XOR intro: normed_normxor)
	done
    next
      assume "b1 \<noteq> c"
      show ?thesis proof cases
	assume "b1 < c"
	thus ?thesis using prems(5-)
	  apply (auto simp add: normxor_standard)
	  apply (subst normxor_simp5)
	  apply (auto elim: normed_XOR)
	  apply (drule normed_xor_snd) back
	  apply (simp add: a2_assoc)
	  done
      next
	assume "\<not> b1 < c"
	hence cleb1: "c < b1" using prems(5-) by auto
	thus ?thesis using prems(5-)
	  apply (auto simp add: normxor_standard)
	  apply (subst normxor_simp6) back
	  apply (auto elim: normed_XOR intro: normed_normxor)
	  apply (frule normed_xor_snd)
	  apply (drule_tac x=a2 and y=b2 in normxor_first)
	  apply (auto elim: normed_XOR simp add: min_def split: split_if_asm)
	  done
      qed
    qed
  next
    assume "a1\<noteq>b1"
    show ?thesis proof cases
      assume "a1 < b1"
      show ?thesis proof cases
	assume "b1=c"
	have A: "(a1 \<odot> a2 \<otimes> c \<oplus> b2) \<otimes> c =
	         a1 \<odot> ((a2 \<otimes> (c \<oplus> b2)) \<otimes> c)" using prems(5-)
	  apply (simp only: normxor_com[where y=c])
	  apply (subst normxor_simp8)
	  by (auto elim: normed_XOR)
	show ?thesis using prems(5-)
	  apply (auto simp add: normxor_standard)
	  apply (simp only: A a2_assoc)
	  apply (auto simp add: normxor_standard)
	  apply (simp only: normxor_com[where y=b2])
	  apply (subst normxor_simp4)
	  by (auto elim: normed_XOR)
      next
	assume "b1\<noteq>c"
	show ?thesis proof cases
	  assume "b1 < c"
	  have "(a1 \<odot> a2 \<otimes> b1 \<oplus> b2) \<otimes> c =
	        a1 \<odot> ((a2 \<otimes> (b1 \<oplus> b2)) \<otimes> c)" using prems(5-)
	    apply (simp only: normxor_com[where y=c])
	    apply (subst normxor_simp8)
	    by (auto elim: normed_XOR)
	  also have "... = a1 \<odot> (a2 \<otimes> ((b1 \<oplus> b2) \<otimes> c))" using `normed c` `normed (b1 \<oplus> b2)`
	    by (simp only: a2_assoc)
	  also have "... = a1 \<odot> (a2 \<otimes> (b1 \<odot> (b2 \<otimes> c)))" using prems(5-)
	    by (auto simp add: normxor_standard)
	  finally have A: "(a1 \<odot> a2 \<otimes> b1 \<oplus> b2) \<otimes> c =  a1 \<odot> (a2 \<otimes> (b1 \<odot> (b2 \<otimes> c)))"
	    by auto
	  thus ?thesis using prems(5-)
	    apply (auto simp add: normxor_standard)
	    apply (subst normxor_simp12)
	    apply (auto elim: normed_XOR)
	    done
	next
	  assume "\<not> b1 < c"
	  hence "c < b1" using prems(5-) by auto
	  show ?thesis proof cases
	    assume "a1=c"
	    have "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> c = (c \<odot> a2 \<otimes> b1 \<oplus> b2) \<otimes> c"
	      using prems(5-) by (auto simp add: normxor_standard)
	    also have "... =  a2 \<otimes> b1 \<oplus> b2" using prems(5-)
	      by (auto elim: normed_XOR)
	    finally have R: "((a1 \<oplus>  a2) \<otimes> b1 \<oplus> b2) \<otimes> c =  a2 \<otimes> b1 \<oplus> b2" by auto
	    have "(a1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> c) = a2 \<otimes> b1 \<oplus> b2" using prems(5-)
	      by (auto simp add: normxor_standard)
	    thus ?thesis using R by simp
	  next
	    assume "a1\<noteq>c"
	    show ?thesis proof cases
	      assume "a1 < c"
	      have "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> c = (a1 \<odot> a2 \<otimes> b1 \<oplus> b2) \<otimes> c"
		using prems(5-) by (auto simp add: normxor_standard)
	      also have "... = c \<otimes>  (a1 \<odot> (a2 \<otimes> (b1 \<oplus> b2)))" by (simp only: normxor_com)
	      also have "... = a1 \<odot> (c \<otimes> (a2 \<otimes> (b1 \<oplus> b2)))" using prems(5-)
		apply (subst normxor_simp8)
		by (auto elim: normed_XOR)
	      also have "... = a1 \<odot> ((a2 \<otimes> (b1 \<oplus> b2)) \<otimes> c)" by (simp only: normxor_com)
	      also have "... = a1 \<odot> (a2 \<otimes> ((b1 \<oplus> b2) \<otimes> c))" using `normed (b1 \<oplus> b2)` `normed c`
		by (simp only: a2_assoc)
	      also have "... = a1 \<odot> (a2 \<otimes> (c \<oplus> (b1 \<oplus> b2)))" using prems(5-)
		by (auto simp add: normxor_standard)
	      finally have R: "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> c = a1 \<odot> (a2 \<otimes> (c \<oplus> (b1 \<oplus> b2)))" by simp
	      have "(a1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> c) = a1 \<odot> (a2 \<otimes> (c \<oplus> (b1 \<oplus> b2)))"
		using prems(5-) by (auto simp add: normxor_standard)
	      thus ?thesis using R by simp
	    next
	      assume "\<not> a1 < c"
	      hence clea1: "c < a1" using prems(5-) by auto
              have "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> c =  (a1 \<odot> a2 \<otimes> b1 \<oplus> b2) \<otimes> c" using prems(5-)
		by (auto simp add: normxor_standard)
	      also have "... = c \<odot> (a1 \<odot> a2 \<otimes> b1 \<oplus> b2)" using prems(5-)
		apply (subst normxor_simp14)
		by (auto elim: normed_XOR)
	      finally have R: "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> c = c \<odot> (a1 \<odot> a2 \<otimes> b1 \<oplus> b2)" by simp
	      have "(a1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> c) = c \<odot> a1 \<odot> a2 \<otimes> b1 \<oplus> b2" using prems(5-)
		by (auto simp add: normxor_standard)
	      thus ?thesis using R by simp
	    qed
	  qed
	qed
      qed
    next
      assume "\<not> a1 < b1"
      hence b1lea1: "b1 < a1" using prems(5-) by auto
      show ?thesis proof cases
	assume "b1=c"
	thus ?thesis using prems(5-)
	  by (auto simp add: normxor_standard)
      next
	assume "b1\<noteq>c"
 	show ?thesis proof cases
	  assume "b1 < c"
	  show ?thesis proof cases
	    assume "a1=c"
	    have "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> c = (b1 \<odot> ((c \<oplus> a2) \<otimes> b2)) \<otimes> c" using prems(5-)
	      by (auto simp add: normxor_standard elim: normed_XOR)
	    also have "... = c \<otimes> (b1 \<odot> ((c \<oplus> a2) \<otimes> b2))"  by (simp only: normxor_com)
	    also have "... = b1 \<odot> (c \<otimes> ((c \<oplus> a2) \<otimes> b2))" using prems(5-)
	      apply (subst normxor_simp8)
	      by (auto elim: normed_XOR)
	    also have "... = b1 \<odot> (((c \<oplus> a2) \<otimes> b2) \<otimes> c)" by (simp only: normxor_com)
	    also have "... = b1 \<odot> ((c \<oplus> a2) \<otimes> (b2 \<otimes> c))" using `normed c` `normed (a1 \<oplus> a2)`
	      by (simp only: b2_assoc `a1=c`[THEN sym])
	    finally have R: "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> c = b1 \<odot> ((c \<oplus> a2) \<otimes> (b2 \<otimes> c))" by simp
	    have "(a1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> c) = (c \<oplus> a2) \<otimes> (b1 \<odot> (b2 \<otimes> c))" using prems(5-)
	      by (auto simp add: normxor_standard elim: normed_XOR)
	    also have "... = b1 \<odot> ((c \<oplus> a2) \<otimes> (b2 \<otimes> c))" using prems(5-)
	      apply (subst normxor_simp7)
	      by (auto elim: normed_XOR)
	    finally show ?thesis using R by simp
	  next
	    assume "a1\<noteq>c"
	    show ?thesis proof cases
	      assume "a1 < c"
	      show ?thesis proof cases
 		assume "a1=b1"
		thus ?thesis using prems(5-)
		  by (auto simp add: normxor_standard elim: normed_XOR)
	      next
		assume "a1\<noteq>b1"
		show ?thesis proof cases
		  assume "a1<b1"
		  thus ?thesis using prems(5-)
		    by (auto simp add: normxor_standard elim: normed_XOR)
		next
		  assume "\<not> a1 < b1"
		  hence "b1 < a1" using prems(5-) by auto
		  thus ?thesis using prems(5-)
		    apply (auto simp add: normxor_standard elim: normed_XOR)
		    apply (subst normxor_simp7)
		    apply (auto simp add: normxor_standard elim: normed_XOR)
		    apply (simp only: normxor_com[where y=c])
		    apply (subst normxor_simp8)
		    apply (auto simp add: normxor_standard elim: normed_XOR)
		    apply (simp only: normxor_com[where x=c])
		    apply (simp only: b2_assoc)
		    done
		qed
	      qed
	    next
	      assume "\<not> a1 < c"
	      hence "c < a1" using prems(5-) by auto
	      have "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> c = (b1 \<odot> (a1 \<oplus> a2) \<otimes> b2) \<otimes> c" using prems(5-)
		by (auto simp add: normxor_standard elim: normed_XOR)
	      also have "... = c \<otimes> (b1 \<odot> (a1 \<oplus> a2) \<otimes> b2)" by (simp only: normxor_com)
	      also have "... = b1 \<odot> (c \<otimes> ((a1 \<oplus> a2) \<otimes> b2))" using prems(5-)
		apply (subst normxor_simp8)
		by (auto elim: normed_XOR)
	      finally have R: "((a1 \<oplus> a2) \<otimes> (b1 \<oplus> b2)) \<otimes> c = b1 \<odot> (c \<otimes> ((a1 \<oplus> a2) \<otimes> b2))" by simp
	      have "(a1 \<oplus> a2) \<otimes> ((b1 \<oplus> b2) \<otimes> c) = (a1 \<oplus> a2) \<otimes> (b1 \<odot> (b2 \<otimes> c))" using prems(5-)
		by (auto simp add: normxor_standard elim: normed_XOR)
	      also have "... = b1 \<odot> ((a1 \<oplus> a2) \<otimes> (b2 \<otimes> c))" using prems(5-)
		apply (subst normxor_simp7)
		by (auto elim: normed_XOR)
	      also have "... = b1 \<odot> (((a1 \<oplus> a2) \<otimes> b2) \<otimes> c)" using `normed (a1 \<oplus> a2)` `normed c`
		by (simp only: b2_assoc)
	      also have "... = b1 \<odot> (c \<otimes> ((a1 \<oplus> a2) \<otimes> b2))" by (simp only: normxor_com)
	      finally show ?thesis using R by simp
	    qed
	  qed
	next
	  assume "\<not> b1 < c"
	  hence cleb1: "c < b1" using prems(5-) by auto
 	  thus ?thesis using prems(5-)
	    by (auto simp add: normxor_standard elim: normed_XOR)
	qed
      qed
    qed
  qed
qed

lemma normxor_assoc2:
  assumes normedx: "normed X"
  and     normedy: "normed Y"
  and     normedz:  "normed Z"
  shows "(X \<otimes> Y) \<otimes> Z = X \<otimes> (Y \<otimes> Z)" using prems
proof (induct X arbitrary: Y Z rule: normed_induct2)
  case Zero thus ?case by auto
next
  case (Standard a)
  have "\<forall> Z. normed Z \<longrightarrow> (a \<otimes> Y) \<otimes> Z = a \<otimes> Y \<otimes> Z"
  proof (rule_tac P="%Y. \<forall> Z. normed Z \<longrightarrow> (a \<otimes> Y) \<otimes> Z = a \<otimes> (Y \<otimes> Z)" in normed_induct2)
    show "normed Y" using prems by auto
  next
    {
      fix Z :: fmsg
      assume "normed Z"
      have "(a \<otimes> ZERO) \<otimes> Z = a \<otimes> ZERO \<otimes> Z" by auto
    } thus "\<forall>Z. normed Z \<longrightarrow> (a \<otimes> ZERO) \<otimes> Z = a \<otimes> ZERO \<otimes> Z" by auto
  next
    fix x :: fmsg
    assume "normed x" and "standard x"
    { 
      fix x a Z :: fmsg
      assume "normed Z" and "normed x" and "standard x" and "normed a" and "standard a"
      have "(a \<otimes> x) \<otimes> Z = a \<otimes> x \<otimes> Z"
      proof (rule_tac P="%Z. (a \<otimes> x) \<otimes> Z = a \<otimes> (x \<otimes> Z)" in normed_induct2)
	show "normed Z" using prems by auto
      next
	show "(a \<otimes> x) \<otimes> ZERO = a \<otimes> x \<otimes> ZERO" using prems by auto
      next
	fix z :: fmsg
	assume "normed z" and "standard z"
	show "(a \<otimes> x) \<otimes> z = a \<otimes> x \<otimes> z" using prems by (auto  simp add: normxor_standard XORnz_def)
      next
	fix ca cb :: fmsg
	assume "normed ca" and "(a \<otimes> x) \<otimes> ca = a \<otimes> x \<otimes> ca" and "standard ca" and
	       "normed cb" and "(a \<otimes> x) \<otimes> cb = a \<otimes> x \<otimes> cb" and
	       "ca < first cb" and "cb \<noteq> ZERO"
	thus "(a \<otimes> x) \<otimes> ca \<oplus> cb = a \<otimes> x \<otimes> ca \<oplus> cb" using prems apply -
	  apply (rule normxor_assoc2_s_s_x)
	  apply force apply force apply force apply force defer
	  apply force apply force
	  apply (rule normed.Xor) apply force
	  apply force apply force apply force apply force
	  done
      qed 
    } thus "\<forall>Z. normed Z \<longrightarrow> (a \<otimes> x) \<otimes> Z = a \<otimes> x \<otimes> Z" using prems by auto
  next
    fix ba bb :: fmsg
    assume "normed ba" and "\<forall> Z. normed Z \<longrightarrow> (a \<otimes> ba) \<otimes> Z = a \<otimes> ba \<otimes> Z" and "standard ba" and
           "normed bb" and "\<forall>Z. normed Z \<longrightarrow> (a \<otimes> bb) \<otimes> Z = a \<otimes> bb \<otimes> Z" and "ba < first bb" and
           "bb \<noteq> ZERO"
    show "\<forall>Z. normed Z \<longrightarrow> (a \<otimes> (ba \<oplus> bb)) \<otimes> Z = a \<otimes> (ba \<oplus> bb) \<otimes> Z"
    proof (auto)
      fix Z :: fmsg
      assume "normed Z"
      have ba_assoc: "!!Z. normed Z \<Longrightarrow> (a \<otimes> ba) \<otimes> Z = a \<otimes> ba \<otimes> Z" using prems by auto
      have bb_assoc: "!!Z. normed Z \<Longrightarrow> (a \<otimes> bb) \<otimes> Z = a \<otimes> bb \<otimes> Z" using prems by auto
      show "(a \<otimes> (ba \<oplus> bb)) \<otimes> Z = a \<otimes> (ba \<oplus> bb) \<otimes> Z"
      proof (rule_tac P="%Z. (a \<otimes> (ba \<oplus> bb)) \<otimes> Z = a \<otimes> ((ba \<oplus> bb) \<otimes> Z)" in normed_induct2)
	show "normed Z" using prems by auto
      next
	show "(a \<otimes> ba \<oplus> bb) \<otimes> ZERO = a \<otimes> (ba \<oplus> bb) \<otimes> ZERO" by auto
      next
	fix c :: fmsg
	assume "normed c" and "standard c"
	thus "(a \<otimes> ba \<oplus> bb) \<otimes> c = a \<otimes> (ba \<oplus> bb) \<otimes> c" using prems ba_assoc bb_assoc apply -
	  apply (rule normxor_assoc2_s_x_s)
	  apply force apply force defer apply force apply force
	  apply (erule ba_assoc[where Z=c])
	  apply (erule bb_assoc[where Z=c])
	  apply (rule normed.Xor)
	  apply force apply force apply force apply force apply force
	  done
      next
	fix ca cb :: fmsg
	assume "normed ca" and "(a \<otimes> ba \<oplus> bb) \<otimes> ca = a \<otimes> (ba \<oplus> bb) \<otimes> ca" and "standard ca" and
	       "normed cb" and "(a \<otimes> ba \<oplus> bb) \<otimes> cb = a \<otimes> (ba \<oplus> bb) \<otimes> cb" and "ca < first cb" and
	       "cb \<noteq> ZERO"
	show "(a \<otimes> ba \<oplus> bb) \<otimes> (ca \<oplus> cb) = a \<otimes> (ba \<oplus> bb) \<otimes> (ca \<oplus> cb)" using prems ba_assoc bb_assoc
	  apply -
	  apply (rule normxor_assoc2_s_x_x) defer defer
	  apply force apply force apply force apply force
	  apply (rule normed.Xor) apply force apply force apply force apply force apply force
	  apply (rule normed.Xor) apply force apply force apply force apply force apply force
	  apply (erule ba_assoc)
	  apply (erule bb_assoc)
	  done
      qed
    qed
  qed
  thus ?case using prems by auto
next
  case (Xor aa ab)
  have "\<forall> Z. normed Z \<longrightarrow> ((aa \<oplus> ab) \<otimes> Y) \<otimes> Z = (aa \<oplus> ab) \<otimes> Y \<otimes> Z"
  proof (rule_tac P="%Y. \<forall> Z. normed Z \<longrightarrow> ((aa \<oplus> ab) \<otimes> Y) \<otimes> Z = (aa \<oplus> ab) \<otimes> (Y \<otimes> Z)" in normed_induct2)
    show "normed Y" using prems by auto
  next
    {
      fix Z :: fmsg
      assume "normed Z"
      have "((aa \<oplus> ab) \<otimes> ZERO) \<otimes> Z = (aa \<oplus> ab) \<otimes> ZERO \<otimes> Z" by auto
    } thus "\<forall>Z. normed Z \<longrightarrow> ((aa \<oplus> ab) \<otimes> ZERO) \<otimes> Z = (aa \<oplus> ab) \<otimes> ZERO \<otimes> Z" by auto
  next
    fix b :: fmsg
    assume "normed b" and "standard b"
    {
      fix Z :: fmsg
      assume "normed Z"
      have "((aa \<oplus> ab) \<otimes> b) \<otimes> Z = (aa \<oplus> ab) \<otimes> b \<otimes> Z"  using prems
      proof (rule_tac P="%Z. ((aa \<oplus> ab) \<otimes> b) \<otimes> Z = (aa \<oplus> ab) \<otimes> (b \<otimes> Z)" in normed_induct2)
	show "normed Z" using prems by auto
      next
	show "((aa \<oplus> ab) \<otimes> b) \<otimes> ZERO = (aa \<oplus> ab) \<otimes> b \<otimes> ZERO" using prems by auto
      next
	fix c :: fmsg
	assume "normed c" and "standard c"
	show "((aa \<oplus> ab) \<otimes> b) \<otimes> c = (aa \<oplus> ab) \<otimes> b \<otimes> c" using prems apply -
	  apply (rule normxor_assoc2_x_s_s)
	  apply force apply force apply force apply force defer
	  apply force apply force
	  apply (rule normed.Xor)
	  apply force apply force
	  apply force apply force apply force
	  done
      next
	fix ca cb :: fmsg
	assume  "normed ca" and "((aa \<oplus> ab) \<otimes> b) \<otimes> ca = (aa \<oplus> ab) \<otimes> b \<otimes> ca" and "standard ca"
	  and "normed cb" and "((aa \<oplus> ab) \<otimes> b) \<otimes> cb = (aa \<oplus> ab) \<otimes> b \<otimes> cb"
	  and "ca < first cb" and "cb \<noteq> ZERO"
	show "((aa \<oplus> ab) \<otimes> b) \<otimes> (ca \<oplus> cb) = (aa \<oplus> ab) \<otimes> b \<otimes> (ca \<oplus> cb)" using prems apply -
	  apply (rule normxor_assoc2_x_s_x)
	  apply force apply force apply force apply force defer
	  apply force apply force
	  apply (rule normed.Xor)
	  apply force apply force apply force apply force apply force
	  apply (rule normed.Xor)
	  apply force apply force apply force apply force apply force
	  done
      qed
    }
    thus "\<forall>Z. normed Z \<longrightarrow> ((aa \<oplus> ab) \<otimes> b) \<otimes> Z = (aa \<oplus> ab) \<otimes> b \<otimes> Z" by auto
  next
    fix ba bb :: fmsg
    assume "normed ba" and "\<forall>Z. normed Z \<longrightarrow> ((aa \<oplus> ab) \<otimes> ba) \<otimes> Z = (aa \<oplus> ab) \<otimes> ba \<otimes> Z" and
           "standard ba" and "normed bb" and
           "\<forall>Z. normed Z \<longrightarrow> ((aa \<oplus> ab) \<otimes> bb) \<otimes> Z = (aa \<oplus> ab) \<otimes> bb \<otimes> Z" and
           "ba < first bb" and  "bb \<noteq> ZERO"
    show "\<forall>Z. normed Z \<longrightarrow> ((aa \<oplus> ab) \<otimes> ba \<oplus> bb) \<otimes> Z = (aa \<oplus> ab) \<otimes> (ba \<oplus> bb) \<otimes> Z"
    proof (safe)
      fix Z :: fmsg
      assume "normed Z"
      have ba_assoc: "!!Z. normed Z \<Longrightarrow> ((aa \<oplus> ab) \<otimes> ba) \<otimes> Z = (aa \<oplus> ab) \<otimes> ba \<otimes> Z" using prems by auto
      have bb_assoc: "!!Z. normed Z \<Longrightarrow> ((aa \<oplus> ab) \<otimes> bb) \<otimes> Z = (aa \<oplus> ab) \<otimes> bb \<otimes> Z" using prems by auto
      show "((aa \<oplus> ab) \<otimes> ba \<oplus> bb) \<otimes> Z = (aa \<oplus> ab) \<otimes> (ba \<oplus> bb) \<otimes> Z"
      proof (rule_tac P="%Z. ((aa \<oplus> ab) \<otimes> (ba \<oplus> bb)) \<otimes> Z = (aa \<oplus> ab) \<otimes> ((ba \<oplus> bb) \<otimes> Z)" in normed_induct2)
	show "normed Z" using prems by auto
      next
	show "((aa \<oplus> ab) \<otimes> ba \<oplus> bb) \<otimes> ZERO = (aa \<oplus> ab) \<otimes> (ba \<oplus> bb) \<otimes> ZERO" by auto
      next
	fix c :: fmsg
	assume "normed c" and "standard c"
	thus "((aa \<oplus> ab) \<otimes> ba \<oplus> bb) \<otimes> c = (aa \<oplus> ab) \<otimes> (ba \<oplus> bb) \<otimes> c" using prems ba_assoc bb_assoc apply -
	  apply (rule normxor_assoc2_x_x_s)
	  defer defer defer defer
	  apply (rule normed.Xor) apply force apply force apply force
	  apply force apply force
	  apply (rule normed.Xor) apply force apply force apply force
	  apply force apply force apply force apply force
	  apply (erule prems(5)) apply force
	  apply (erule prems(8)) apply force
	  apply (erule ba_assoc)
	  apply (erule bb_assoc)
	  done
      next
	fix ca cb :: fmsg
	assume "normed ca" and "((aa \<oplus> ab) \<otimes> ba \<oplus> bb) \<otimes> ca = (aa \<oplus> ab) \<otimes> (ba \<oplus> bb) \<otimes> ca"
	        and "standard ca" and
	       "normed cb" and "((aa \<oplus> ab) \<otimes> ba \<oplus> bb) \<otimes> cb = (aa \<oplus> ab) \<otimes> (ba \<oplus> bb) \<otimes> cb"
	       and "ca < first cb" and "cb \<noteq> ZERO"
	show "((aa \<oplus> ab) \<otimes> ba \<oplus> bb) \<otimes> (ca \<oplus> cb) = (aa \<oplus> ab) \<otimes> (ba \<oplus> bb) \<otimes> (ca \<oplus> cb)"
	  using prems ba_assoc bb_assoc
	  apply -
	  apply (rule normxor_assoc2_x_x_x)
	  apply (erule prems(5)) apply force
	  apply (erule prems(8)) apply force
	  apply (erule ba_assoc)
	  apply (erule bb_assoc)
	  apply force
	  apply force
	  apply (rule normed.Xor) apply force apply force apply force apply force apply force
	  apply (rule normed.Xor) apply force apply force apply force apply force apply force
	  apply (rule normed.Xor) apply force apply force apply force apply force apply force
	  done
      qed
    qed
  qed
  thus ?case using prems by auto
qed

lemma equiv_imp_norm: "x \<approx> y ==> norm x = norm y"
  apply (erule xor_eq.induct)
  apply (auto)
  apply (rule normxor_assoc2[THEN sym])
  apply (auto simp add: normxor_com)
  apply (auto intro: normed_norm)
done

lemma normxor_equiv:
  "\<lbrakk> normed a; normed b \<rbrakk>
   \<Longrightarrow> XOR a b \<approx> normxor a b"
proof (induct a arbitrary: b rule: normed_induct2)
  case (Standard x)
  show ?case using prems apply -
   apply (rule normed_induct2[where P="%b. XOR x b  \<approx> normxor x b"])
   apply force
   apply force
   apply (auto simp add: normxor_standard XORnz_def)
   apply (rule_tac A1=x and B1="b \<oplus> a" in Xor_cong_trans)
   apply force
   apply force
   apply (rule_tac Xor_assoc_trans)
   apply (rule_tac A1=ZERO and B1=a in Xor_cong_trans)
   apply force apply force
   apply force
   apply (rule Xor_assoc_trans)
   apply (rule_tac A1=ZERO and B1=b in Xor_cong_trans)
   apply force apply force apply force
   apply (rule xor_eq.symm)
   apply (rule_tac A1=a and B1="x \<oplus> b" in Xor_cong_trans)
   apply force
   apply (rule xor_eq.symm) apply force
   apply (rule Xor_com_trans)
   apply (rule xor_eq.symm) 
   apply (rule_tac A1=x and B1="b \<oplus> a" in Xor_cong_trans)
   apply force apply force
   apply (rule_tac Xor_assoc_trans) apply force
   done
next
  case Zero
  show ?case using prems by auto
next
  case (Xor x y)
  show ?case using `normed b`
  proof (induct b rule: normed_induct2[where P="%b. XOR (XOR x y) b  \<approx> normxor (XOR x y) b"])
    case Zero
    show ?case using prems by auto
  next
    case (Standard z)
    show ?case using prems(1,3,4,6-) thm prems apply -
      apply (auto simp add: normxor_standard XORnz_def)
      apply (subgoal_tac "y \<oplus>  z \<approx> y \<otimes> z") prefer 2
      apply (erule prems(5))
      apply simp
      apply (rule Xor_assoc_trans2)
      apply (rule_tac A1="x" and B1=ZERO in Xor_cong_trans)
      apply force apply force apply force
      apply (rule_tac A1="y \<oplus> x" and B1=x in Xor_cong_trans)
      apply force apply force
      apply (rule Xor_assoc_trans2)
      apply (rule_tac A1="y" and B1=ZERO in Xor_cong_trans)
      apply force apply force apply force
      apply (rule Xor_assoc_trans2)
      apply (subgoal_tac "y \<oplus> z \<approx> y \<otimes> z") prefer 2
      apply (auto intro: prems)
      apply (rule xor_eq.symm)
      apply (rule_tac A1="x" and B1="y \<oplus> z" in Xor_cong_trans)
      apply force apply (rule xor_eq.symm) apply force
      apply force
      done
  next
    case (Xor u v)
    show ?case using prems(1,3,4,6-)
      apply (auto simp add: normxor_standard XORnz_def split: split_if_asm)

      apply (rule_tac A1="x \<oplus> y" and B1="v \<oplus> x" in Xor_cong_trans)
      apply force apply force
      apply (rule Xor_assoc_trans)
      apply (rule_tac A1="x" and B1="x" in Xor_cong_trans)
      apply force apply force apply force

      apply (rule Xor_assoc_trans)
      apply (rule_tac A1="y" and B1="v" in Xor_cong_trans)
      apply (rule_tac A1="y \<oplus> x" and B1="x" in Xor_cong_trans)
      apply force apply force
      apply (rule Xor_assoc_trans2)
      apply (rule_tac A1="y" and B1="ZERO" in Xor_cong_trans)
      apply force apply force
      apply force apply force
      apply (erule prems(5)) defer
      
      apply (rule xor_eq.symm)
      apply (rule_tac A1="u" and B1="(x \<oplus> y) \<oplus> v" in Xor_cong_trans)
      apply force apply (rule xor_eq.symm, force)
      apply (rule Xor_assoc_trans)
      apply (rule_tac A1="(x \<oplus> y) \<oplus> u" and B1="v" in Xor_cong_trans) prefer 3
      apply (rule Xor_assoc_trans2) apply force prefer 2 apply force
      apply (rule Xor_com_trans) apply force defer

      apply (rule xor_eq.symm)
      apply (rule_tac A1="u" and B1="(x \<oplus> y) \<oplus> v" in Xor_cong_trans)
      apply force apply (rule xor_eq.symm, force)
      apply (rule Xor_assoc_trans)
      apply (rule_tac A1="(x \<oplus> y) \<oplus> u" and B1="v" in Xor_cong_trans) prefer 3
      apply (rule Xor_assoc_trans2) apply force prefer 2 apply force
      apply (rule Xor_com_trans) apply force

      apply (rule Xor_assoc_trans2)
      apply (subgoal_tac  "y \<oplus> u \<oplus> v \<approx> y \<otimes> u \<oplus> v") prefer 2
      apply (rule prems(5))
      apply (rule normed.Xor) apply force apply force apply force apply force
      apply force
      apply (rule_tac A1="x" and B1="ZERO" in Xor_cong_trans)
      apply force apply force apply force defer

      apply (rule Xor_assoc_trans2)
      apply (subgoal_tac  "y \<oplus> u \<oplus> v \<approx> y \<otimes> u \<oplus> v") prefer 2
      apply (rule prems(5))
      apply (rule normed.Xor) apply force apply force apply force apply force
      apply force
      apply (rule_tac A1="x" and B1="ZERO" in Xor_cong_trans)
      apply force apply force apply force defer

      apply (rule Xor_assoc_trans2)
      apply (subgoal_tac  "y \<oplus> u \<oplus> v \<approx> y \<otimes> u \<oplus> v") prefer 2
      apply (rule prems(5))
      apply (rule normed.Xor) apply force apply force apply force apply force
      apply force
      apply (rule_tac A1="x" and B1="ZERO" in Xor_cong_trans)
      apply force apply force apply force defer

      apply (rule Xor_assoc_trans2)
      apply (subgoal_tac  "y \<oplus> u \<oplus> v \<approx> y \<otimes> u \<oplus> v") prefer 2
      apply (rule prems(5))
      apply (rule normed.Xor) apply force apply force apply force apply force
      apply force
      apply (rule_tac A1="x" and B1="ZERO" in Xor_cong_trans)
      apply force apply force apply force
      
      apply (rule Xor_assoc_trans2)
      apply (subgoal_tac  "y \<oplus> u \<oplus> v \<approx> y \<otimes> u \<oplus> v") prefer 2
      apply (rule prems(5))
      apply (rule normed.Xor) apply force apply force apply force apply force apply force
      apply (rule_tac A1="x" in Xor_cong_trans)
      apply force apply force apply force

      apply (rule_tac A1="x \<oplus> y" and B1="v \<oplus> u" in Xor_cong_trans)
      apply auto
      apply (rule Xor_assoc_trans)
      apply (rule_tac A1="ZERO" and B1="u" in Xor_cong_trans)
      apply auto

      apply (rule_tac A1="x \<oplus> y" and B1="v \<oplus> u" in Xor_cong_trans)
      apply auto
      apply (rule Xor_assoc_trans)
      apply (rule_tac A1="ZERO" and B1="u" in Xor_cong_trans)
      apply auto

      apply (rule Xor_assoc_trans2)
      apply (subgoal_tac  "y \<oplus> u \<oplus> v \<approx> y \<otimes> u \<oplus> v") prefer 2
      apply (rule prems(5))
      apply (rule normed.Xor) apply force apply force apply force apply force
      apply force
      apply (rule_tac A1="x" in Xor_cong_trans)
      apply force apply force apply force
      
      apply (rule Xor_assoc_trans2)
      apply (subgoal_tac  "y \<oplus> u \<oplus> v \<approx> y \<otimes> u \<oplus> v") prefer 2
      apply (rule prems(5))
      apply (rule normed.Xor) apply force apply force apply force apply force
      apply force
      apply (rule_tac A1="x" in Xor_cong_trans)
      apply force apply force apply force

      apply (rule Xor_assoc_trans2)
      apply (subgoal_tac  "y \<oplus> u \<oplus> v \<approx> y \<otimes> u \<oplus> v") prefer 2
      apply (rule prems(5))
      apply (rule normed.Xor) apply force apply force apply force apply force
      apply force
      apply (rule_tac A1="x" in Xor_cong_trans)
      apply force apply force apply force

      done
  qed
qed

lemma norm_equiv: "x \<approx> norm x"
  apply (induct x)
  apply (auto intro: xor_eq.refl)
  apply (erule xor_eq.Hash_cong)
  apply (erule xor_eq.MPair_cong)
  apply force
  apply (erule xor_eq.Crypt_cong)
  apply (rule_tac A1="norm x1" and B1="norm x2" in Xor_cong_trans)
  apply auto
  apply (rule normxor_equiv)
  apply (rule normed_norm)+
done

lemma norm_imp_equiv: "norm x = norm y ==> x \<approx> y"
  apply (rule xor_eq.trans)
  apply (rule  norm_equiv)
  apply (rule xor_eq.symm)
  apply simp
  apply (rule norm_equiv)
done

lemma equiv_norm: "(x \<approx> y) = (norm x = norm y)"
  apply auto
  apply (auto intro: norm_imp_equiv equiv_imp_norm)
done

end