(*
  ProtoVeriPhy

  Module: distance/MessageTheoryXor2.thy

  Copyright (C) 2008-2012 Benedikt Schmidt <benedikt.schmidt@inf.ethz.ch>
  License: BSD (see COPYRIGHT)
*)

theory MessageTheoryXor2 imports MessageTheoryXor begin

subsection {* parts, subterms, and quotient type *}

typedef msg = "{m | m. normed m}"
  apply (rule_tac x="NUMBER 1" in exI)
  apply force
done

definition
  Agent :: "agent \<Rightarrow> msg"
 where
  "Agent a = Abs_msg (AGENT a)"

definition
  Number :: "int \<Rightarrow> msg"
 where
  "Number i = Abs_msg (NUMBER i)"

definition
  Real :: "real \<Rightarrow> msg"
 where
  "Real i = Abs_msg (REAL i)"

definition
  Key :: "key \<Rightarrow> msg"
 where
  "Key i = Abs_msg (KEY i)"

definition
  Hash :: "msg \<Rightarrow> msg"
 where
  "Hash m = Abs_msg (HASH (Rep_msg m))"

definition
  MPair :: "msg \<Rightarrow> msg \<Rightarrow> msg"
 where
  "MPair a b = Abs_msg (MPAIR (Rep_msg a) (Rep_msg b))"

definition
  Crypt :: "key \<Rightarrow> msg \<Rightarrow> msg"
 where
  "Crypt k m = Abs_msg (CRYPT k (Rep_msg m))"

definition
  Xor :: "msg \<Rightarrow> msg \<Rightarrow> msg"
 where
  "Xor a b = Abs_msg (norm ((Rep_msg a) \<oplus> (Rep_msg b)))"

definition
  Zero :: "msg"
 where
  "Zero = Abs_msg ZERO"

definition
  Nonce :: "agent \<Rightarrow> nat \<Rightarrow> msg"
 where
  "Nonce a n = Abs_msg (NONCE a n)"

interpretation MESSAGE_THEORY_DATA Key Crypt Nonce MPair Hash Number
  apply (unfold_locales)
done

lemma normed_Rep_msg[simp,intro]: "normed (Rep_msg m)"
  apply (subgoal_tac "Rep_msg m \<in> msg") prefer 2
  apply (rule Rep_msg)
  apply (auto simp add: msg_def)
done

lemma Abs_msg_normed[simp]: "normed m \<Longrightarrow> Rep_msg (Abs_msg m) = m"
  apply (rule Abs_msg_inverse)
  apply (auto simp add: msg_def)
done

inductive_set
  fparts :: "fmsg set => fmsg set"
  for H ::  "fmsg set "
  where
    Inj [intro]: "X \<in> H                  \<Longrightarrow> X \<in> fparts H"
  | Fst:         "MPAIR X Y   \<in> fparts H \<Longrightarrow> X \<in> fparts H"
  | Snd:         "MPAIR X Y   \<in> fparts H \<Longrightarrow> Y \<in> fparts H"
  | Ctext:       "CRYPT k M   \<in> fparts H \<Longrightarrow> M \<in> fparts H"
  | Xor1:        "X \<oplus> Y       \<in> fparts H \<Longrightarrow> X \<in> fparts H"
  | Xor2:        "X \<oplus> Y       \<in> fparts H \<Longrightarrow> Y \<in> fparts H"

lemma normed_fparts:
  "\<lbrakk> Y \<in> fparts {X}; normed X \<rbrakk> \<Longrightarrow> normed Y"
  apply (erule fparts.induct)
  apply auto
  apply (erule normed_MPAIR)
  apply (auto elim: normed_MPAIR normed_HASH normed_XOR normed_CRYPT)
done

lemma fparts_inj:
  "X \<in> H \<Longrightarrow> X \<in> fparts H"
  apply (erule fparts.Inj)
done

lemma fparts_singleton:
  "X \<in> fparts H \<Longrightarrow> \<exists>Y\<in>H. X \<in> fparts {Y}"
  apply (erule fparts.induct)
  apply (auto elim: fparts.Fst fparts.Snd fparts.Xor1 fparts.Xor2
                    fparts.Ctext)
done

lemma fparts_mono:
  "G \<subseteq> H \<Longrightarrow> fparts G \<subseteq> fparts H"
  apply auto
  apply (erule fparts.induct)
  apply (auto elim: fparts.Fst fparts.Snd fparts.Xor1 fparts.Xor2
                    fparts.Ctext)
done

lemma fparts_idem:
  "fparts (fparts H) = fparts H"
  apply auto
  apply (erule fparts.induct)
  apply (auto elim: fparts.Fst fparts.Snd fparts.Xor1 fparts.Xor2
                    Hash fparts.Ctext)
done

interpretation fparts: MESSAGE_THEORY_SUBTERM_NOTION fparts
  apply (unfold_locales)
  apply (erule fparts_inj)
  apply (erule fparts_singleton)
  apply (erule fparts_mono)
  apply (rule fparts_idem)
done

subsubsection {* rewrite rules for pulling out atomic messages *}

lemma fparts_insert_AGENT [simp]:
     "fparts (insert (AGENT agt) H) = insert (AGENT agt) (fparts H)"
  apply (rule fparts.insert_eq_I) 
  apply (erule fparts.induct, auto dest: fparts.Fst fparts.Snd fparts.Ctext
                                         fparts.Xor1 fparts.Xor2)
done

lemma fparts_insert_NONCE [simp]:
     "fparts (insert (NONCE B N) H) = insert (NONCE B N) (fparts H)"
  apply (rule fparts.insert_eq_I)
  apply (erule fparts.induct, auto dest: fparts.Fst fparts.Snd fparts.Ctext
                                         fparts.Xor1 fparts.Xor2)
done

lemma fparts_insert_NUMBER [simp]:
     "fparts (insert (NUMBER N) H) = insert (NUMBER N) (fparts H)"
  apply (rule fparts.insert_eq_I)
  apply (erule fparts.induct, auto dest: fparts.Fst fparts.Snd fparts.Ctext
                                         fparts.Xor1 fparts.Xor2)
done

lemma fparts_insert_Real [simp]:
     "fparts (insert (REAL N) H) = insert (REAL N) (fparts H)"
  apply (rule fparts.insert_eq_I)
  apply (erule fparts.induct, auto dest: fparts.Fst fparts.Snd fparts.Ctext
                                         fparts.Xor1 fparts.Xor2)
done

lemma fparts_insert_KEY [simp]:
     "fparts (insert (KEY K) H) = insert (KEY K) (fparts H)"
  apply (rule fparts.insert_eq_I)
  apply (erule fparts.induct, auto dest: fparts.Fst fparts.Snd fparts.Ctext
                                         fparts.Xor1 fparts.Xor2)
done

lemma fparts_insert_ZERO [simp]:
     "fparts (insert (ZERO) H) = insert ZERO (fparts H)"
  apply (rule fparts.insert_eq_I)
  apply (erule fparts.induct, auto dest: fparts.Fst fparts.Snd fparts.Ctext
                                         Hash fparts.Xor1 fparts.Xor2)
done

lemma fparts_insert_HASH [simp]:
     "fparts (insert (HASH X) H) = insert (HASH X) (fparts H)"
  apply (rule equalityI)
  apply (rule subsetI)
  apply (erule fparts.induct, auto dest: fparts.Fst fparts.Snd fparts.Ctext
                                         fparts.Xor1 fparts.Xor2)
done

lemma fparts_insert_CRYPT [simp]:
     "fparts (insert (CRYPT K X) H) = insert (CRYPT K X) (fparts (insert X H))"
  apply (rule equalityI)
  apply (rule subsetI)
  apply (erule fparts.induct, auto dest: fparts.Fst fparts.Snd fparts.Ctext
                                         fparts.Xor1 fparts.Xor2)
  apply (blast intro: fparts.Ctext)
done

lemma fparts_insert_MPAIR [simp]:
     "fparts (insert (MPAIR X Y) H) =  
      insert (MPAIR X Y) (fparts (insert X (insert Y H)))"
  apply (rule equalityI)
  apply (rule subsetI)
  apply (erule fparts.induct, auto dest: fparts.Fst fparts.Snd fparts.Ctext
                                         fparts.Xor1 fparts.Xor2)
  apply (blast intro: fparts.Fst fparts.Snd)+
done

lemma fparts_insert_XOR [simp]:
     "fparts (insert (X \<oplus> Y) H) =  
      insert (X \<oplus> Y) (fparts (insert X (insert Y H)))"
  apply (rule equalityI)
  apply (rule subsetI)
  apply (erule fparts.induct, auto dest: fparts.Fst fparts.Snd fparts.Ctext
                                         fparts.Xor1 fparts.Xor2)
  apply (blast intro: fparts.Xor1 fparts.Xor2)+
done

subsubsection {* fsubterms *}

inductive_set
  fsubterms :: "fmsg set => fmsg set"
  for H ::  "fmsg set "
  where
    Inj [intro]: "X \<in> H                  \<Longrightarrow> X \<in> fsubterms H"
  | Fst:         "MPAIR X Y   \<in> fsubterms H \<Longrightarrow> X \<in> fsubterms H"
  | Snd:         "MPAIR X Y   \<in> fsubterms H \<Longrightarrow> Y \<in> fsubterms H"
  | Ctext:       "CRYPT k M   \<in> fsubterms H \<Longrightarrow> M \<in> fsubterms H"
  | Hash:        "HASH M      \<in> fsubterms H \<Longrightarrow> M \<in> fsubterms H"
  | Xor1:        "X \<oplus> Y       \<in> fsubterms H \<Longrightarrow> X \<in> fsubterms H"
  | Xor2:        "X \<oplus> Y       \<in> fsubterms H \<Longrightarrow> Y \<in> fsubterms H"

lemma normed_fsubterms:
  "\<lbrakk> Y \<in> fsubterms {X}; normed X \<rbrakk> \<Longrightarrow> normed Y"
  apply (erule fsubterms.induct)
  apply auto
  apply (erule normed_MPAIR)
  apply (auto elim: normed_MPAIR normed_HASH normed_XOR normed_CRYPT)
done

lemma fsubterms_inj:
  "X \<in> H \<Longrightarrow> X \<in> fsubterms H"
  apply (erule fsubterms.Inj)
done

lemma fsubterms_singleton:
  "X \<in> fsubterms H \<Longrightarrow> \<exists>Y\<in>H. X \<in> fsubterms {Y}"
  apply (erule fsubterms.induct)
  apply (auto elim: fsubterms.Fst fsubterms.Snd fsubterms.Xor1 fsubterms.Xor2
                    fsubterms.Ctext fsubterms.Hash)
done

lemma fsubterms_mono:
  "G \<subseteq> H \<Longrightarrow> fsubterms G \<subseteq> fsubterms H"
  apply auto
  apply (erule fsubterms.induct)
  apply (auto elim: fsubterms.Fst fsubterms.Snd fsubterms.Xor1 fsubterms.Xor2
                    fsubterms.Hash fsubterms.Ctext)
done

lemma fsubterms_idem:
  "fsubterms (fsubterms H) = fsubterms H"
  apply auto
  apply (erule fsubterms.induct)
  apply (auto elim: fsubterms.Fst fsubterms.Snd fsubterms.Xor1 fsubterms.Xor2
                    fsubterms.Hash fsubterms.Ctext)
done

interpretation fsubterms: MESSAGE_THEORY_SUBTERM_NOTION fsubterms
  apply (unfold_locales)
  apply (erule fsubterms_inj)
  apply (erule fsubterms_singleton)
  apply (erule fsubterms_mono)
  apply (rule fsubterms_idem)
done

subsubsection {* rewrite rules for pulling out atomic messages *}

lemma fsubterms_insert_AGENT [simp]:
     "fsubterms (insert (AGENT agt) H) = insert (AGENT agt) (fsubterms H)"
  apply (rule fsubterms.insert_eq_I) 
  apply (erule fsubterms.induct, auto dest: fsubterms.Fst fsubterms.Snd fsubterms.Ctext
                                            fsubterms.Hash fsubterms.Xor1 fsubterms.Xor2)
done

lemma fsubterms_insert_NONCE [simp]:
     "fsubterms (insert (NONCE B N) H) = insert (NONCE B N) (fsubterms H)"
  apply (rule fsubterms.insert_eq_I)
  apply (erule fsubterms.induct, auto dest: fsubterms.Fst fsubterms.Snd fsubterms.Ctext
                                         fsubterms.Hash fsubterms.Xor1 fsubterms.Xor2)
done

lemma fsubterms_insert_NUMBER [simp]:
     "fsubterms (insert (NUMBER N) H) = insert (NUMBER N) (fsubterms H)"
  apply (rule fsubterms.insert_eq_I)
  apply (erule fsubterms.induct, auto dest: fsubterms.Fst fsubterms.Snd fsubterms.Ctext
                                         fsubterms.Hash fsubterms.Xor1 fsubterms.Xor2)
done

lemma fsubterms_insert_Real [simp]:
     "fsubterms (insert (REAL N) H) = insert (REAL N) (fsubterms H)"
  apply (rule fsubterms.insert_eq_I)
  apply (erule fsubterms.induct, auto dest: fsubterms.Fst fsubterms.Snd fsubterms.Ctext
                                         fsubterms.Hash fsubterms.Xor1 fsubterms.Xor2)
done

lemma fsubterms_insert_KEY [simp]:
     "fsubterms (insert (KEY K) H) = insert (KEY K) (fsubterms H)"
  apply (rule fsubterms.insert_eq_I)
  apply (erule fsubterms.induct, auto dest: fsubterms.Fst fsubterms.Snd fsubterms.Ctext
                                         fsubterms.Hash fsubterms.Xor1 fsubterms.Xor2)
done

lemma fsubterms_insert_ZERO [simp]:
     "fsubterms (insert (ZERO) H) = insert ZERO (fsubterms H)"
  apply (rule fsubterms.insert_eq_I)
  apply (erule fsubterms.induct, auto dest: fsubterms.Fst fsubterms.Snd fsubterms.Ctext
                                         fsubterms.Hash fsubterms.Xor1 fsubterms.Xor2)
done

lemma fsubterms_insert_HASH [simp]:
     "fsubterms (insert (HASH X) H) = insert (HASH X) (fsubterms (insert X H))"
  apply (rule equalityI)
  apply (rule subsetI)
  apply (erule fsubterms.induct, auto dest: fsubterms.Fst fsubterms.Snd fsubterms.Ctext
                                         fsubterms.Hash fsubterms.Xor1 fsubterms.Xor2)
  apply (blast intro: fsubterms.Hash)
done

lemma fsubterms_insert_CRYPT [simp]:
     "fsubterms (insert (CRYPT K X) H) = insert (CRYPT K X) (fsubterms (insert X H))"
  apply (rule equalityI)
  apply (rule subsetI)
  apply (erule fsubterms.induct, auto dest: fsubterms.Fst fsubterms.Snd fsubterms.Ctext
                                         fsubterms.Hash fsubterms.Xor1 fsubterms.Xor2)
  apply (blast intro: fsubterms.Ctext)
done

lemma fsubterms_insert_MPAIR [simp]:
     "fsubterms (insert (MPAIR X Y) H) =  
      insert (MPAIR X Y) (fsubterms (insert X (insert Y H)))"
  apply (rule equalityI)
  apply (rule subsetI)
  apply (erule fsubterms.induct, auto dest: fsubterms.Fst fsubterms.Snd fsubterms.Ctext
                                         fsubterms.Hash fsubterms.Xor1 fsubterms.Xor2)
  apply (blast intro: fsubterms.Fst fsubterms.Snd)+
done

lemma fsubterms_insert_XOR [simp]:
     "fsubterms (insert (X \<oplus> Y) H) =  
      insert (X \<oplus> Y) (fsubterms (insert X (insert Y H)))"
  apply (rule equalityI)
  apply (rule subsetI)
  apply (erule fsubterms.induct, auto dest: fsubterms.Fst fsubterms.Snd fsubterms.Ctext
                                         fsubterms.Hash fsubterms.Xor1 fsubterms.Xor2)
  apply (blast intro: fsubterms.Xor1 fsubterms.Xor2)+
done

subsubsection {* parts *}

definition
  parts :: "msg set \<Rightarrow> msg set"
 where
  "parts H = { Abs_msg m | m . m \<in> fparts (Rep_msg`H)}"

lemma parts_inj1:
  "X \<in> H \<Longrightarrow> X \<in> parts H"
  apply (unfold parts_def)
  apply auto
  apply (rule_tac x="Rep_msg X" in exI)
  apply (auto simp add: Rep_msg_inverse)
done

lemma parts_singleton1:
  "X \<in> parts H \<Longrightarrow> \<exists>Y\<in>H. X \<in> parts {Y}"
  apply (unfold parts_def)
  apply auto
  apply (drule fparts_singleton)
by auto

lemma parts_mono1:
  "G \<subseteq> H \<Longrightarrow> parts G \<subseteq> parts H"
  apply (unfold parts_def)
  apply auto
  apply (rule_tac x=m in exI)
  apply (subgoal_tac "(Rep_msg`G) \<subseteq> (Rep_msg`H)") prefer 2
  apply force
  apply (drule fparts_mono)
  apply (rule conjI)
  apply force
  apply (erule rev_subsetD)
  apply force
done

lemma vimage_inside:
  "f`{g m | m. p m} = {f (g m) | m . p m}"
by auto

lemma parts_idem1:
  "parts (parts H) = parts H"
  apply (unfold parts_def)
  apply auto
  apply (rule_tac x=m in exI) prefer 2
  apply (rule_tac x=m in exI)
  apply (auto simp add: vimage_inside)
  apply (subgoal_tac
         "\<exists> nm \<in> {Rep_msg (Abs_msg m) |m. m \<in> fparts (Rep_msg ` H)}. m \<in> fparts {nm}")
  prefer 2
  apply (rule_tac x=m in bexI)
  apply auto
  apply (rule_tac x=m in exI)
  apply auto
  apply (subst Abs_msg_normed)
  apply auto
  apply (drule fparts_singleton, auto)
  apply (drule normed_fparts)
  apply auto
  apply (subgoal_tac
         "{Rep_msg (Abs_msg ma)} \<subseteq> {Rep_msg (Abs_msg m) |m. m \<in> fparts (Rep_msg ` H)}")
  apply (drule fparts_mono)
  apply (erule rev_subsetD) back
  apply force
  apply force

  apply (drule fparts_singleton)
  apply auto
  apply (subgoal_tac "ma \<in> msg")
  apply (simp add: Abs_msg_inverse) prefer 2
  apply (auto simp add: msg_def)
  apply (drule_tac X=ma in fparts_singleton)
  apply auto
  apply (erule normed_fparts)
  apply (auto simp add: Rep_msg)
  apply (subgoal_tac "m \<in> fparts (fparts (Rep_msg ` H))")
  apply (force simp add: fparts_idem)
  apply (subgoal_tac "{ma} \<subseteq> fparts (Rep_msg ` H)") prefer 2
  apply force
  apply (drule fparts_mono)
  apply (erule rev_subsetD)
  apply force
done

subsubsection {* simplification rules for parts *}

lemma parts_Number[simp]: "parts {Number i} = {Number i}"
  apply (auto simp add: parts_def) prefer 2
  apply (rule_tac x="NUMBER i" in exI)
  apply auto prefer 2
  apply (rule fparts.Inj)
  apply (auto simp add: Number_def)
  apply (subst Abs_msg_inverse)
  apply (auto simp add: msg_def)
  apply (subgoal_tac "normed (NUMBER i)")
  apply (simp only: Abs_msg_normed)
  apply force
  apply force
done

lemma parts_Real[simp]: "parts {Real i} = {Real i}"
  apply (auto simp add: parts_def) prefer 2
  apply (rule_tac x="REAL i" in exI)
  apply auto prefer 2
  apply (rule fparts.Inj)
  apply (auto simp add: Real_def)
  apply (subst Abs_msg_inverse)
  apply (auto simp add: msg_def)
  apply (subgoal_tac "normed (REAL i)")
  apply (simp only: Abs_msg_normed)
  apply force
  apply force
done

lemma parts_Nonce[simp]: "parts {Nonce a i} = {Nonce a i}"
  apply (auto simp add: parts_def) prefer 2
  apply (rule_tac x="NONCE a i" in exI)
  apply auto prefer 2
  apply (rule fparts.Inj)
  apply (auto simp add: Nonce_def)
  apply (subst Abs_msg_inverse)
  apply (auto simp add: msg_def)
  apply (subgoal_tac "normed (NONCE a i)")
  apply (simp only: Abs_msg_normed)
  apply force
  apply force
done

lemma parts_Key[simp]: "parts {Key k} = {Key k}"
  apply (auto simp add: parts_def) prefer 2
  apply (rule_tac x="KEY k" in exI)
  apply auto prefer 2
  apply (rule fparts.Inj)
  apply (auto simp add: Key_def)
  apply (subst Abs_msg_inverse)
  apply (auto simp add: msg_def)
  apply (subgoal_tac "normed (KEY k)")
  apply (simp only: Abs_msg_normed)
  apply force
  apply force
done

lemma parts_Agent[simp]: "parts {Agent a} = {Agent a}"
  apply (auto simp add: parts_def) prefer 2
  apply (rule_tac x="AGENT a" in exI)
  apply auto prefer 2
  apply (rule fparts.Inj)
  apply (auto simp add: Agent_def)
  apply (subst Abs_msg_inverse)
  apply (auto simp add: msg_def)
  apply (subgoal_tac "normed (AGENT a)")
  apply (simp only: Abs_msg_normed)
  apply force
  apply force
done

lemma parts_Hash[simp]: "parts {Hash h} = {Hash h}"
  apply (auto simp add: parts_def) prefer 2
  apply (rule_tac x="HASH (Rep_msg h)" in exI)
  apply auto prefer 2
  apply (rule fparts.Inj)
  apply (auto simp add: Hash_def)
  apply (subst Abs_msg_inverse)
  apply (auto simp add: msg_def)
  apply (subgoal_tac "normed (HASH (Rep_msg h))") prefer 2
  apply auto
done

lemma fparts_mono_elem:
  "\<lbrakk> X \<in> fparts H; H \<subseteq> G \<rbrakk> \<Longrightarrow> X \<in> fparts G"
  apply (drule fparts_mono)
by (erule rev_subsetD)

lemma parts_MPair[simp]: "parts {MPair a b} = {MPair a b} \<union> parts {a} \<union> parts {b}"
  apply (auto simp add: parts_def) prefer 2
  apply (rule_tac x="MPAIR (Rep_msg a) (Rep_msg b)" in exI)
  apply auto prefer 2
  apply (rule fparts.Inj)
  apply (auto simp add: MPair_def)
  apply (subst Abs_msg_inverse)
  apply (auto simp add: msg_def)
  apply (subgoal_tac "normed (MPAIR (Rep_msg a) (Rep_msg b))") prefer 2
  apply auto
  apply (drule fparts_singleton)
  apply auto
  apply (subgoal_tac "normed (MPAIR (Rep_msg a) (Rep_msg b))") prefer 2
  apply auto
  apply (rule_tac x=m in exI)
  apply (rule conjI)
  apply force
  apply (rule disjI2)
  apply (erule fparts_mono_elem)
  apply force

  apply (subgoal_tac "normed (MPAIR (Rep_msg a) (Rep_msg b))") prefer 2
  apply auto
  apply (rule_tac x=m in exI)
  apply (rule conjI)
  apply force
  apply (rule disjI2)
  apply (erule fparts_mono_elem)
  apply force
done

lemma parts_Crypt[simp]: "parts {Crypt k m} = {Crypt k m} \<union> parts {m}"
  apply (auto simp add: parts_def) prefer 2
  apply (rule_tac x="CRYPT k (Rep_msg m)" in exI)
  apply auto prefer 2
  apply (rule fparts.Inj)
  apply (auto simp add: Crypt_def)
  apply (subst Abs_msg_inverse)
  apply (auto simp add: msg_def)
  apply (subgoal_tac "normed (CRYPT k (Rep_msg m))") prefer 2
  apply auto
  apply (subgoal_tac "normed (CRYPT k (Rep_msg m))") prefer 2
  apply auto
done

interpretation parts: MESSAGE_THEORY_PARTS Crypt Nonce MPair Hash Number Key parts
  apply (unfold_locales)
  apply (erule parts_inj1)
  apply (erule parts_singleton1)
  apply (erule parts_mono1)
  apply (rule parts_idem1)
done

subsubsection {* subterms *}

definition
  subterms :: "msg set \<Rightarrow> msg set"
 where
  "subterms H = { Abs_msg m | m . m \<in> fsubterms (Rep_msg`H)}"

lemma subterms_inj1:
  "X \<in> H \<Longrightarrow> X \<in> subterms H"
  apply (unfold subterms_def)
  apply auto
  apply (rule_tac x="Rep_msg X" in exI)
  apply (auto simp add: Rep_msg_inverse)
done

lemma subterms_singleton1:
  "X \<in> subterms H \<Longrightarrow> \<exists>Y\<in>H. X \<in> subterms {Y}"
  apply (unfold subterms_def)
  apply auto
  apply (drule fsubterms_singleton)
by auto

lemma subterms_mono1:
  "G \<subseteq> H \<Longrightarrow> subterms G \<subseteq> subterms H"
  apply (unfold subterms_def)
  apply auto
  apply (rule_tac x=m in exI)
  apply (subgoal_tac "(Rep_msg`G) \<subseteq> (Rep_msg`H)") prefer 2
  apply force
  apply (drule fsubterms_mono)
  apply (rule conjI)
  apply force
  apply (erule rev_subsetD)
  apply force
done

lemma subterms_idem1:
  "subterms (subterms H) = subterms H"
  apply (unfold subterms_def)
  apply auto
  apply (rule_tac x=m in exI) prefer 2
  apply (rule_tac x=m in exI)
  apply (auto simp add: vimage_inside)
  apply (subgoal_tac
         "\<exists> nm \<in> {Rep_msg (Abs_msg m) |m. m \<in> fsubterms (Rep_msg ` H)}. m \<in> fsubterms {nm}")
  prefer 2
  apply (rule_tac x=m in bexI)
  apply auto
  apply (rule_tac x=m in exI)
  apply auto
  apply (subst Abs_msg_normed)
  apply auto
  apply (drule fsubterms_singleton, auto)
  apply (drule normed_fsubterms)
  apply auto
  apply (subgoal_tac
         "{Rep_msg (Abs_msg ma)} \<subseteq> {Rep_msg (Abs_msg m) |m. m \<in> fsubterms (Rep_msg ` H)}")
  apply (drule fsubterms_mono)
  apply (erule rev_subsetD) back
  apply force
  apply force

  apply (drule fsubterms_singleton)
  apply auto
  apply (subgoal_tac "ma \<in> msg")
  apply (simp add: Abs_msg_inverse) prefer 2
  apply (auto simp add: msg_def)
  apply (drule_tac X=ma in fsubterms_singleton)
  apply auto
  apply (erule normed_fsubterms)
  apply (auto simp add: Rep_msg)
  apply (subgoal_tac "m \<in> fsubterms (fsubterms (Rep_msg ` H))")
  apply (force simp add: fsubterms_idem)
  apply (subgoal_tac "{ma} \<subseteq> fsubterms (Rep_msg ` H)") prefer 2
  apply force
  apply (drule fsubterms_mono)
  apply (erule rev_subsetD)
  apply force
done

subsubsection {* simplification rules for subterms *}

lemma subterms_Number[simp]: "subterms {Number i} = {Number i}"
  apply (auto simp add: subterms_def) prefer 2
  apply (rule_tac x="NUMBER i" in exI)
  apply auto prefer 2
  apply (rule fsubterms.Inj)
  apply (auto simp add: Number_def)
  apply (subst Abs_msg_inverse)
  apply (auto simp add: msg_def)
  apply (subgoal_tac "normed (NUMBER i)")
  apply (simp only: Abs_msg_normed)
  apply force
  apply force
done

lemma subterms_Real[simp]: "subterms {Real i} = {Real i}"
  apply (auto simp add: subterms_def) prefer 2
  apply (rule_tac x="REAL i" in exI)
  apply auto prefer 2
  apply (rule fsubterms.Inj)
  apply (auto simp add: Real_def)
  apply (subst Abs_msg_inverse)
  apply (auto simp add: msg_def)
  apply (subgoal_tac "normed (REAL i)")
  apply (simp only: Abs_msg_normed)
  apply force
  apply force
done

lemma subterms_Nonce[simp]: "subterms {Nonce a i} = {Nonce a i}"
  apply (auto simp add: subterms_def) prefer 2
  apply (rule_tac x="NONCE a i" in exI)
  apply auto prefer 2
  apply (rule fsubterms.Inj)
  apply (auto simp add: Nonce_def)
  apply (subst Abs_msg_inverse)
  apply (auto simp add: msg_def)
  apply (subgoal_tac "normed (NONCE a i)")
  apply (simp only: Abs_msg_normed)
  apply force
  apply force
done

lemma subterms_Key[simp]: "subterms {Key k} = {Key k}"
  apply (auto simp add: subterms_def) prefer 2
  apply (rule_tac x="KEY k" in exI)
  apply auto prefer 2
  apply (rule fsubterms.Inj)
  apply (auto simp add: Key_def)
  apply (subst Abs_msg_inverse)
  apply (auto simp add: msg_def)
  apply (subgoal_tac "normed (KEY k)")
  apply (simp only: Abs_msg_normed)
  apply force
  apply force
done

lemma subterms_Agent[simp]: "subterms {Agent a} = {Agent a}"
  apply (auto simp add: subterms_def) prefer 2
  apply (rule_tac x="AGENT a" in exI)
  apply auto prefer 2
  apply (rule fsubterms.Inj)
  apply (auto simp add: Agent_def)
  apply (subst Abs_msg_inverse)
  apply (auto simp add: msg_def)
  apply (subgoal_tac "normed (AGENT a)")
  apply (simp only: Abs_msg_normed)
  apply force
  apply force
done

lemma subterms_Hash[simp]: "subterms {Hash h} = {Hash h} \<union> subterms {h}"
  apply (auto simp add: subterms_def) prefer 2
  apply (rule_tac x="HASH (Rep_msg h)" in exI)
  apply auto prefer 2
  apply (rule fsubterms.Inj)
  apply (auto simp add: Hash_def)
  apply (subst Abs_msg_inverse)
  apply (auto simp add: msg_def)
  apply (subgoal_tac "normed (HASH (Rep_msg h))") prefer 2
  apply auto
  apply (subgoal_tac "normed (HASH (Rep_msg h))") prefer 2
  apply auto
done

lemma fsubterms_mono_elem:
  "\<lbrakk> X \<in> fsubterms H; H \<subseteq> G \<rbrakk> \<Longrightarrow> X \<in> fsubterms G"
  apply (drule fsubterms_mono)
by (erule rev_subsetD)

lemma subterms_MPair[simp]: "subterms {MPair a b} = {MPair a b} \<union> subterms {a} \<union> subterms {b}"
  apply (auto simp add: subterms_def) prefer 2
  apply (rule_tac x="MPAIR (Rep_msg a) (Rep_msg b)" in exI)
  apply auto prefer 2
  apply (rule fsubterms.Inj)
  apply (auto simp add: MPair_def)
  apply (subst Abs_msg_inverse)
  apply (auto simp add: msg_def)
  apply (subgoal_tac "normed (MPAIR (Rep_msg a) (Rep_msg b))") prefer 2
  apply auto
  apply (drule fsubterms_singleton)
  apply auto
  apply (subgoal_tac "normed (MPAIR (Rep_msg a) (Rep_msg b))") prefer 2
  apply auto
  apply (rule_tac x=m in exI)
  apply (rule conjI)
  apply force
  apply (rule disjI2)
  apply (erule fsubterms_mono_elem)
  apply force

  apply (subgoal_tac "normed (MPAIR (Rep_msg a) (Rep_msg b))") prefer 2
  apply auto
  apply (rule_tac x=m in exI)
  apply (rule conjI)
  apply force
  apply (rule disjI2)
  apply (erule fsubterms_mono_elem)
  apply force
done

lemma subterms_Crypt[simp]: "subterms {Crypt k m} = {Crypt k m} \<union> subterms {m}"
  apply (auto simp add: subterms_def) prefer 2
  apply (rule_tac x="CRYPT k (Rep_msg m)" in exI)
  apply auto prefer 2
  apply (rule fsubterms.Inj)
  apply (auto simp add: Crypt_def)
  apply (subst Abs_msg_inverse)
  apply (auto simp add: msg_def)
  apply (subgoal_tac "normed (CRYPT k (Rep_msg m))") prefer 2
  apply auto
  apply (subgoal_tac "normed (CRYPT k (Rep_msg m))") prefer 2
  apply auto
done

lemma Abs_eq_normed[dest]: "\<lbrakk> Abs_msg a = Abs_msg b; normed a; normed b \<rbrakk> \<Longrightarrow> a = b \<and> normed b"
  apply (subgoal_tac "Rep_msg (Abs_msg a) = Rep_msg (Abs_msg b)") prefer 2
  apply force
  apply (thin_tac "Abs_msg a = Abs_msg b")
  apply (force simp only: Abs_msg_normed)
done

lemma fparts_fsubterms_Abs_msg:
  "\<lbrakk> m' \<in> fparts (Rep_msg ` H); Abs_msg m' = Abs_msg m; m \<in> fsubterms (Rep_msg ` H) \<rbrakk>
   \<Longrightarrow> m = m'"
  apply (drule fparts_singleton)
  apply (drule fsubterms_singleton)
  apply auto
  apply (drule normed_fsubterms)
  apply force
  apply (drule normed_fparts)
  apply auto
done

interpretation subterms: MESSAGE_THEORY_SUBTERM Crypt Nonce MPair Hash Number parts Key subterms
  apply (unfold_locales)
  apply (erule subterms_inj1)
  apply (erule subterms_singleton1)
  apply (erule subterms_mono1)
  apply (rule subterms_idem1)
  apply (unfold parts_def subterms_def)
  apply auto
  apply (erule fparts.induct)
  apply (auto intro: fsubterms.Inj fsubterms.Fst fsubterms.Snd fsubterms.Ctext fsubterms.Hash
              fsubterms.Xor1 fsubterms.Xor2
              dest: fparts_fsubterms_Abs_msg)
done

subsubsection {* results about parts and subterms *}

notation MPair  ("(2\<lbrace>_,/ _\<rbrace>)") 

notation MACM  ("(4Hash[_] /_)" [0, 1000])

inductive
  xor_red :: "fmsg => fmsg => bool" ("_ ~> _" [60,60])
 where
  Xor_assoc_1[intro]:  "(X \<oplus> (Y \<oplus> Z)) ~> ((X \<oplus> Y) \<oplus> Z)" |
  Xor_assoc_2[intro]:  "((X \<oplus> Y) \<oplus> Z) ~> (X \<oplus> (Y \<oplus> Z))" |
  Xor_com[intro]:    "X \<oplus> Y ~> Y \<oplus> X" |
  Xor_Zero[intro]:   "X \<oplus> ZERO ~> X" |
  Xor_cancel[intro]: "X ~> Y ==> X \<oplus> Y ~> ZERO" |

  MPair_cong: "\<lbrakk> X ~> A ; Y ~> B \<rbrakk> \<Longrightarrow> MPAIR X Y ~> MPAIR A B" |
  Hash_cong:   "X ~> Y ==> HASH X ~> HASH Y" |
  Crypt_cong:  "M ~> N ==> CRYPT K M ~> CRYPT K N" |
  Xor_cong:   "\<lbrakk> X ~> A ; Y ~> B \<rbrakk> \<Longrightarrow> X \<oplus> Y ~> A \<oplus> B" |

  refl[intro]: "X ~> X" |
  trans:       "[| X ~> Y; Y ~> Z |] ==> X ~> Z"

lemma xor_red_imp_xor_eq: "X ~> Y \<Longrightarrow> X \<approx> Y"
  apply (erule xor_red.induct)
  apply auto
  apply (rule xor_eq.symm)
  apply (rule xor_eq.Xor_assoc)  
  apply (auto intro: xor_eq.MPair_cong xor_eq.Hash_cong
                     xor_eq.Crypt_cong xor_eq.Xor_cong xor_eq.trans)
done

lemma set_reorder_XOR:
  "{X,Y \<oplus> Z} = {Y \<oplus> Z, X}"
by auto

lemma set_reorder_insert:
  "insert X (insert Y H) = insert Y (insert X H)"
by auto

lemma set_reorder_insert_ZERO:
  "insert X (insert ZERO H) = insert ZERO (insert X H)"
by auto

lemma fsubterms_reduce_NONCE[rule_format]:
  "\<lbrakk> A ~> B; NONCE C N \<in> fsubterms {B} \<rbrakk> \<Longrightarrow> NONCE C N \<in> fsubterms {A}"
  apply (induct A B rule: xor_red.induct)
  apply (auto simp add: set_reorder_XOR)

  apply (force simp add: set_reorder_insert)
  apply (force simp add: set_reorder_insert)
  apply (force simp add: set_reorder_insert)
  apply (force simp add: set_reorder_insert_ZERO)
  apply (drule fsubterms_singleton)
  apply (auto elim: fsubterms.insertI)
  apply (rule fsubterms_mono[THEN subsetD]) prefer 2
  apply assumption
  apply force
  apply (drule fsubterms_singleton)
  apply (auto elim: fsubterms.insertI)
  apply (rule fsubterms.mono[THEN subsetD]) prefer 2
  apply assumption
  apply force
done


lemma fsubterms_reduce_AGENT[rule_format]:
  "\<lbrakk> A ~> B; AGENT C \<in> fsubterms {B} \<rbrakk> \<Longrightarrow> AGENT C \<in> fsubterms {A}"
  apply (induct A B rule: xor_red.induct)
  apply (auto simp add: set_reorder_XOR)

  apply (force simp add: set_reorder_insert)
  apply (force simp add: set_reorder_insert)
  apply (force simp add: set_reorder_insert)
  apply (force simp add: set_reorder_insert_ZERO)
  apply (drule fsubterms_singleton)
  apply (auto elim: fsubterms.insertI)
  apply (rule fsubterms_mono[THEN subsetD]) prefer 2
  apply assumption
  apply force
  apply (drule fsubterms_singleton)
  apply (auto elim: fsubterms.insertI)
  apply (rule fsubterms.mono[THEN subsetD]) prefer 2
  apply assumption
  apply force
done


(* Have to prove X ~> norm X for this to be useful *)
lemma fsubterms_reduce_KEY[rule_format]:
  "\<lbrakk> A ~> B; KEY k \<in> fsubterms {B} \<rbrakk> \<Longrightarrow> KEY k \<in> fsubterms {A}"
  apply (induct A B rule: xor_red.induct)
  apply (auto simp add: set_reorder_XOR)

  apply (force simp add: set_reorder_insert)
  apply (force simp add: set_reorder_insert)
  apply (force simp add: set_reorder_insert)
  apply (force simp add: set_reorder_insert_ZERO)

  apply (drule fsubterms_singleton)
  apply (auto elim: fsubterms.insertI)
  apply (rule fsubterms_mono[THEN subsetD]) prefer 2
  apply assumption
  apply force

  apply (drule fsubterms_singleton)
  apply (auto elim: fsubterms.insertI)
  apply (rule fsubterms_mono[THEN subsetD]) prefer 2
  apply assumption
  apply force
done

(* Have to prove X ~> norm X for this to be useful *)
lemma fparts_reduce_KEY[rule_format]:
  "\<lbrakk> A ~> B; KEY k \<in> fparts {B} \<rbrakk> \<Longrightarrow> KEY k \<in> fparts {A}"
  apply (induct A B rule: xor_red.induct)
  apply (auto simp add: set_reorder_XOR)

  apply (force simp add: set_reorder_insert)
  apply (force simp add: set_reorder_insert)
  apply (force simp add: set_reorder_insert)
  apply (force simp add: set_reorder_insert_ZERO)

  apply (drule fparts_singleton)
  apply (auto elim: fparts.insertI)
  apply (rule fparts_mono[THEN subsetD]) prefer 2
  apply assumption
  apply force

  apply (drule fparts_singleton)
  apply (auto elim: fparts.insertI)
  apply (rule fparts_mono[THEN subsetD]) prefer 2
  apply assumption
  apply force
done

(* Have to prove X ~> norm X for this to be useful *)
lemma fparts_reduce_NONCE[rule_format]:
  "\<lbrakk> A ~> B; NONCE a na \<in> fparts {B} \<rbrakk> \<Longrightarrow> NONCE a na \<in> fparts {A}"
  apply (induct A B rule: xor_red.induct)
  apply (auto simp add: set_reorder_XOR)

  apply (force simp add: set_reorder_insert)
  apply (force simp add: set_reorder_insert)
  apply (force simp add: set_reorder_insert)
  apply (force simp add: set_reorder_insert_ZERO)

  apply (drule fparts_singleton)
  apply (auto elim: fparts.insertI)
  apply (rule fparts_mono[THEN subsetD]) prefer 2
  apply assumption
  apply force

  apply (drule fparts_singleton)
  apply (auto elim: fparts.insertI)
  apply (rule fparts_mono[THEN subsetD]) prefer 2
  apply assumption
  apply force
done

(* Have to prove X ~> norm X for this to be useful *)
lemma fparts_reduce_CRYPT[rule_format]:
  "\<lbrakk> A ~> B; CRYPT k msig \<in> fparts {B} \<rbrakk>
   \<Longrightarrow> \<exists> msig'. CRYPT k msig' \<in> fparts {A} \<and> msig' ~> msig"
  apply (induct A B arbitrary: msig rule: xor_red.induct)
  apply (rule_tac x=msig in exI)
  apply (simp add: set_reorder_XOR)
  apply (force simp add: set_reorder_insert)

  apply (rule_tac x=msig in exI)
  apply (simp add: set_reorder_XOR)
  apply (force simp add: set_reorder_insert)

  apply (rule_tac x=msig in exI)
  apply (simp add: set_reorder_XOR)
  apply (force simp add: set_reorder_insert)

  apply (rule_tac x=msig in exI)
  apply (simp add: set_reorder_insert_ZERO)
  apply force

  apply force

  prefer 2
  apply simp prefer 4
  apply force prefer 3

  apply simp
  apply (drule fparts_singleton)
  apply safe
  apply clarsimp
  apply (subgoal_tac "\<exists>msig'. CRYPT k msig' \<in> fparts {X} \<and> msig' ~> msig") prefer 2
  apply force
  apply clarify
  apply (rule_tac x=msig' in exI)
  apply (force intro: fparts_mono_elem)
  apply (subgoal_tac "\<exists>msig'. CRYPT k msig' \<in> fparts {Y} \<and> msig' ~> msig") prefer 2
  apply force
  apply clarify
  apply (rule_tac x=msig' in exI)
  apply (force intro: fparts_mono_elem)
  
  apply clarsimp
  apply (drule fparts_singleton)
  apply safe
  apply clarsimp
  apply (subgoal_tac "\<exists>msig'. CRYPT k msig' \<in> fparts {X} \<and> msig' ~> msig") prefer 2
  apply force
  apply clarify
  apply (rule_tac x=msig' in exI)
  apply (force intro: fparts_mono_elem)
  apply (subgoal_tac "\<exists>msig'. CRYPT k msig' \<in> fparts {Y} \<and> msig' ~> msig") prefer 2
  apply force
  apply clarify
  apply (rule_tac x=msig' in exI)
  apply (force intro: fparts_mono_elem)

  apply (case_tac "CRYPT k msig = CRYPT K N")
    apply (rule_tac x=M in exI)
    apply force
  apply auto

  apply (subgoal_tac "\<exists>msig'. CRYPT k msig' \<in> fparts {Y} \<and> msig' ~> msig") prefer 2
  apply force
  apply clarify
  apply (subgoal_tac "\<exists>msig''. CRYPT k msig'' \<in> fparts {X} \<and> msig'' ~> msig'")
  prefer 2
  apply force
  apply clarify
  apply (rule_tac x=msig'' in exI)
  apply (rule conjI)
  apply force
  apply (rule xor_red.trans)
  apply auto
done

lemma fsubterms_reduce_CRYPT[rule_format]:
  "\<lbrakk> A ~> B; CRYPT k msig \<in> fsubterms {B} \<rbrakk>
   \<Longrightarrow> \<exists> msig'. CRYPT k msig' \<in> fsubterms {A} \<and> msig' ~> msig"
  apply (induct A B arbitrary: msig rule: xor_red.induct)
  apply (rule_tac x=msig in exI)
  apply (simp add: set_reorder_XOR)
  apply (force simp add: set_reorder_insert)

  apply (rule_tac x=msig in exI)
  apply (simp add: set_reorder_XOR)
  apply (force simp add: set_reorder_insert)

  apply (rule_tac x=msig in exI)
  apply (simp add: set_reorder_XOR)
  apply (force simp add: set_reorder_insert)

  apply (rule_tac x=msig in exI)
  apply (simp add: set_reorder_insert_ZERO)
  apply force

  apply force

  prefer 2
  apply simp prefer 4
  apply force prefer 3

  apply simp
  apply (drule fsubterms_singleton)
  apply safe
  apply clarsimp
  apply (subgoal_tac "\<exists>msig'. CRYPT k msig' \<in> fsubterms {X} \<and> msig' ~> msig") prefer 2
  apply force
  apply clarify
  apply (rule_tac x=msig' in exI)
  apply (force intro: fsubterms_mono_elem)
  apply (subgoal_tac "\<exists>msig'. CRYPT k msig' \<in> fsubterms {Y} \<and> msig' ~> msig") prefer 2
  apply force
  apply clarify
  apply (rule_tac x=msig' in exI)
  apply (force intro: fsubterms_mono_elem)
  
  apply clarsimp
  apply (drule fsubterms_singleton)
  apply safe
  apply clarsimp
  apply (subgoal_tac "\<exists>msig'. CRYPT k msig' \<in> fsubterms {X} \<and> msig' ~> msig") prefer 2
  apply force
  apply clarify
  apply (rule_tac x=msig' in exI)
  apply (force intro: fsubterms_mono_elem)
  apply (subgoal_tac "\<exists>msig'. CRYPT k msig' \<in> fsubterms {Y} \<and> msig' ~> msig") prefer 2
  apply force
  apply clarify
  apply (rule_tac x=msig' in exI)
  apply (force intro: fsubterms_mono_elem)

  apply (case_tac "CRYPT k msig = CRYPT K N")
    apply (rule_tac x=M in exI)
    apply force
  apply auto

  apply (subgoal_tac "\<exists>msig'. CRYPT k msig' \<in> fsubterms {Y} \<and> msig' ~> msig") prefer 2
  apply force
  apply clarify
  apply (subgoal_tac "\<exists>msig''. CRYPT k msig'' \<in> fsubterms {X} \<and> msig'' ~> msig'")
  prefer 2
  apply force
  apply clarify
  apply (rule_tac x=msig'' in exI)
  apply (rule conjI)
  apply force
  apply (rule xor_red.trans)
  apply auto
done

lemma fsubterms_reduce_HASH[rule_format]:
  "\<lbrakk> A ~> B; HASH m \<in> fsubterms {B} \<rbrakk>
   \<Longrightarrow> \<exists> m'. HASH m' \<in> fsubterms {A} \<and> m' ~> m"
  apply (induct A B arbitrary: m rule: xor_red.induct)
  apply (rule_tac x=m in exI)
  apply (simp add: set_reorder_XOR)
  apply (force simp add: set_reorder_insert)

  apply (rule_tac x=m in exI)
  apply (simp add: set_reorder_XOR)
  apply (force simp add: set_reorder_insert)

  apply (rule_tac x=m in exI)
  apply (simp add: set_reorder_XOR)
  apply (force simp add: set_reorder_insert)

  apply (rule_tac x=m in exI)
  apply (simp add: set_reorder_insert_ZERO)
  apply force

  apply force prefer 3
  apply force prefer 4
  apply force

  apply (drule fsubterms_singleton)
  apply auto

  apply (drule fsubterms_singleton)
  apply auto
  apply (subgoal_tac "\<exists>m'. HASH m' \<in> fsubterms {X} \<and> m' ~> m") prefer 2
  apply force
  apply (elim exE)
  apply (rule_tac x=m' in exI)
  apply (force intro: intro: fsubterms_mono_elem)

  apply (subgoal_tac "\<exists>m'. HASH m' \<in> fsubterms {Y} \<and> m' ~> m") prefer 2
  apply force
  apply (elim exE)
  apply (rule_tac x=m' in exI)
  apply (force intro: intro: fsubterms_mono_elem)

  apply (drule fsubterms_singleton)
  apply auto
  apply (subgoal_tac "\<exists>m'. HASH m' \<in> fsubterms {X} \<and> m' ~> m") prefer 2
  apply force
  apply (elim exE)
  apply (rule_tac x=m' in exI)
  apply (force intro: intro: fsubterms_mono_elem)

  apply (subgoal_tac "\<exists>m'. HASH m' \<in> fsubterms {Y} \<and> m' ~> m") prefer 2
  apply force
  apply (elim exE)
  apply (rule_tac x=m' in exI)
  apply (force intro: intro: fsubterms_mono_elem)

  
  apply (subgoal_tac "\<exists>m'. HASH m' \<in> fsubterms {Y} \<and> m' ~> m") prefer 2
  apply force
  apply (elim exE)
  apply (subgoal_tac "\<exists>m''. HASH m'' \<in> fsubterms {X} \<and> m'' ~> m'") prefer 2
  apply force
  apply (elim exE)
  apply (rule_tac x=m'' in exI)
  apply (rule conjI)
  apply auto
  apply (erule xor_red.trans)
  apply force
done

lemma fsubterms_reduce_MPAIR[rule_format]:
  "\<lbrakk> M ~> N; MPAIR a b \<in> fsubterms {N} \<rbrakk>
   \<Longrightarrow> \<exists> a' b'. MPAIR a' b' \<in> fsubterms {M} \<and> a' ~> a \<and> b' ~> b"
  apply (induct M N arbitrary: a b rule: xor_red.induct)
  apply (rule_tac x=a in exI, rule_tac x=b in exI)
  apply (simp add: set_reorder_XOR)
  apply (force simp add: set_reorder_insert)

  apply (rule_tac x=a in exI, rule_tac x=b in exI)
  apply (simp add: set_reorder_XOR)
  apply (force simp add: set_reorder_insert)

  apply (rule_tac x=a in exI, rule_tac x=b in exI)
  apply (simp add: set_reorder_XOR)
  apply (force simp add: set_reorder_insert)

  apply (rule_tac x=a in exI, rule_tac x=b in exI)
  apply (simp add: set_reorder_insert_ZERO)
  apply force

  apply force prefer 3
  apply force prefer 4
  apply force defer

  apply (drule fsubterms_singleton)
  apply auto

  apply (drule fsubterms_singleton)
  apply auto
  apply (subgoal_tac "\<exists>a' b'. MPAIR a' b' \<in> fsubterms {X} \<and> a' ~> a \<and> b' ~> b") prefer 2
  apply force
  apply (elim exE)
  apply (rule_tac x=a' in exI, rule_tac x=b' in exI)
  apply (force intro: intro: fsubterms_mono_elem)
  apply (subgoal_tac "\<exists>a' b'. MPAIR a' b' \<in> fsubterms {Y} \<and> a' ~> a \<and> b' ~> b") prefer 2
  apply force
  apply (elim exE)
  apply (rule_tac x=a' in exI, rule_tac x=b' in exI)
  apply (force intro: intro: fsubterms_mono_elem)
  
  apply (subgoal_tac "\<exists>a' b'. MPAIR a' b' \<in> fsubterms {Y} \<and> a' ~> a \<and> b' ~> b") prefer 2
  apply force
  apply (elim exE)
  apply (subgoal_tac "\<exists>a'' b''. MPAIR a'' b'' \<in> fsubterms {X} \<and> a'' ~> a' \<and> b'' ~> b'") prefer 2
  apply force
  apply (elim exE)
  apply (rule_tac x=a'' in exI, rule_tac x=b'' in exI)
  apply (rule conjI)
  apply auto
  apply (erule xor_red.trans)
  apply force
  apply (erule xor_red.trans)
  apply force
  
  apply (drule fsubterms_singleton)
  apply auto
  apply (subgoal_tac "\<exists>a' b'. MPAIR a' b' \<in> fsubterms {X} \<and> a' ~> a \<and> b' ~> b") prefer 2
  apply force
  apply (elim exE)
  apply (rule_tac x=a' in exI, rule_tac x=b' in exI)
  apply (intro conjI)
  apply (rule disjI2)
  apply (force intro: intro: fsubterms_mono_elem)
  apply force
  apply force
  apply (subgoal_tac "\<exists>a' b'. MPAIR a' b' \<in> fsubterms {Y} \<and> a' ~> a \<and> b' ~> b") prefer 2
  apply force
  apply (elim exE)
  apply (rule_tac x=a' in exI, rule_tac x=b' in exI)
  apply (intro conjI)
  apply (rule disjI2)
  apply (force intro: intro: fsubterms_mono_elem)
  apply force
  apply force 
done

lemmas Red_com_trans = xor_red.trans[OF xor_red.Xor_com]
lemmas Red_Zero2_trans[intro] = xor_red.trans[OF xor_red.Xor_Zero]
lemmas Red_Zero1_trans[intro] = Red_Zero2_trans[THEN  Red_com_trans]
lemmas Red_assoc1_trans = xor_red.Xor_assoc_1 [THEN xor_red.trans]
lemmas Red_assoc2_trans = xor_red.Xor_assoc_2 [THEN xor_red.trans]
lemmas Red_cong_trans = xor_red.Xor_cong [THEN xor_red.trans]

lemma normxor_reduce:
  "\<lbrakk> normed a; normed b \<rbrakk> \<Longrightarrow> XOR a b ~> normxor a b"
proof (induct a arbitrary: b rule: normed_induct2)
  case Zero
  show ?case using prems by auto
next
  case (Standard x)
  show ?case using prems apply -
   apply (rule normed_induct2[where P="%b. x \<oplus> b  ~> x \<otimes> b"])
   apply force
   apply force
   apply (auto simp add: normxor_standard XORnz_def)
   apply (rule_tac A1=x and B1="b \<oplus> a" in Red_cong_trans)
   apply force
   apply force
   apply (rule_tac Red_assoc1_trans)
   apply (rule_tac A1=ZERO and B1=a in Red_cong_trans)
   apply force apply force
   apply force
   apply (rule Red_assoc1_trans)
   apply (rule_tac A1=ZERO and B1=b in Red_cong_trans)
   apply force apply force apply force
   apply (rule Red_assoc1_trans)
   apply (rule_tac A1="a \<oplus> x" and B1=b in Red_cong_trans)
   apply (rule Red_com_trans)
   apply force apply force
   apply (rule Red_assoc2_trans)
   apply (rule_tac A1="a" and B1="x \<otimes> b" in Red_cong_trans)
   apply force+
   done
next
  case (Xor x y)
  show ?case using `normed b`
  proof (induct b rule: normed_induct2[where P="%b.  (x \<oplus> y) \<oplus> b  ~> (x \<oplus> y) \<otimes> b"])
    case Zero
    show ?case using prems by auto
  next
    case (Standard z)
    show ?case using prems(1,3,4,6-) thm prems apply -
      apply (auto simp add: normxor_standard XORnz_def)
      apply (subgoal_tac "y \<oplus>  z ~> y \<otimes> z") prefer 2
      apply (erule prems(5))
      apply simp
      apply (rule Red_assoc2_trans)
      apply (rule_tac A1="x" and B1=ZERO in Red_cong_trans)
      apply force apply force apply force
      apply (rule_tac  A1="y \<oplus> x" and B1=x in Red_cong_trans)
      apply force
      apply force
      apply (rule Red_assoc2_trans)
      apply (rule_tac  A1="y" and B1=ZERO in Red_cong_trans)
      apply auto
      apply (subgoal_tac "y \<oplus> z ~> y \<otimes> z") prefer 2
      apply (auto intro: prems)
      apply (rule Red_assoc2_trans)
      apply (rule_tac A1="x" and B1="y \<otimes> z" in Red_cong_trans)
      apply force  apply force apply force
      done
  next
    case (Xor u v)
    show ?case using prems(1,3,4,6-)
      apply (auto simp add: normxor_standard XORnz_def split: split_if_asm)

      apply (rule_tac A1="x \<oplus> y" and B1="v \<oplus> x" in Red_cong_trans)
      apply force apply force
      apply (rule Red_assoc1_trans)
      apply (rule_tac A1="x" and B1="x" in Red_cong_trans)
      apply force apply force apply force

      apply (rule Red_assoc1_trans)
      apply (rule_tac A1="y" and B1="v" in Red_cong_trans)
      apply force apply force
      apply (erule prems(5))

      apply (rule_tac A1="x \<oplus> y" and B1="v \<oplus> u" in Red_cong_trans)
      apply force apply force
      apply (rule Red_assoc1_trans)
      apply (rule_tac A1="ZERO" and B1="u" in Red_cong_trans)
      apply force apply force
      apply force
      
      apply (rule_tac A1="x \<oplus> y" and B1="v \<oplus> u" in Red_cong_trans)
      apply force apply force
      apply (rule Red_assoc1_trans)
      apply (rule_tac A1="(x \<oplus> y) \<otimes> v" and B1="u" in Red_cong_trans)
      apply force apply force
      apply (rule Red_com_trans)
      apply force

      apply (rule_tac A1="x \<oplus> y" and B1="v \<oplus> u" in Red_cong_trans)
      apply force apply force
      apply (rule Red_assoc1_trans)
      apply (rule_tac A1="ZERO" and B1="u" in Red_cong_trans)
      apply force apply force apply force

      apply (rule_tac A1="x \<oplus> y" and B1="v \<oplus> u" in Red_cong_trans)
      apply force apply force
      apply (rule Red_assoc1_trans)
      apply (rule_tac A1="(x \<oplus> y) \<otimes> v" and B1="u" in Red_cong_trans)
      apply force apply force
      apply (rule Red_com_trans)
      apply force

      apply (rule Red_assoc2_trans)
      apply (subgoal_tac "y \<oplus> u \<oplus> v ~> y \<otimes> u \<oplus> v") prefer 2
      apply (rule prems(5))
      apply (erule normed.Xor)
      apply force apply force apply force apply force
      apply (rule_tac A1="x" in Red_cong_trans)
      apply force apply assumption
      apply force

      apply (rule Red_assoc2_trans)
      apply (subgoal_tac "y \<oplus> u \<oplus> v ~> y \<otimes> u \<oplus> v") prefer 2
      apply (rule prems(5))
      apply (erule normed.Xor)
      apply force apply force apply force apply force
       apply (rule_tac A1="x" in Red_cong_trans)
      apply force apply assumption
      apply force

      apply (rule Red_assoc2_trans)
      apply (subgoal_tac "y \<oplus> u \<oplus> v ~> y \<otimes> u \<oplus> v") prefer 2
      apply (rule prems(5))
      apply (erule normed.Xor)
      apply force apply force apply force apply force
      apply (rule_tac A1="x" in Red_cong_trans)
      apply force apply assumption
      apply force

      apply (rule Red_assoc2_trans)
      apply (subgoal_tac "y \<oplus> u \<oplus> v ~> y \<otimes> u \<oplus> v") prefer 2
      apply (rule prems(5))
      apply (erule normed.Xor)
      apply force apply force apply force apply force
       apply (rule_tac A1="x" in Red_cong_trans)
      apply force apply assumption
      apply force

      apply (rule Red_assoc2_trans)
      apply (subgoal_tac "y \<oplus> u \<oplus> v ~> y \<otimes> u \<oplus> v") prefer 2
      apply (rule prems(5))
      apply (erule normed.Xor)
      apply force apply force apply force apply force
      apply (rule_tac A1="x" in Red_cong_trans)
      apply force apply assumption
      apply force

      apply (rule Red_assoc2_trans)
      apply (subgoal_tac "y \<oplus> u \<oplus> v ~> y \<otimes> u \<oplus> v") prefer 2
      apply (rule prems(5))
      apply (erule normed.Xor)
      apply force apply force apply force apply force
      apply (rule_tac A1="x" in Red_cong_trans)
      apply force apply assumption
      apply force
  
      apply (rule Red_assoc2_trans)
      apply (subgoal_tac "y \<oplus> u \<oplus> v ~> y \<otimes> u \<oplus> v") prefer 2
      apply (rule prems(5))
      apply (erule normed.Xor)
      apply force apply force apply force apply force
      apply (rule_tac A1="x" in Red_cong_trans)
      apply force apply assumption
      apply force
      
      apply (rule Red_assoc2_trans)
      apply (subgoal_tac "y \<oplus> u \<oplus> v ~> y \<otimes> u \<oplus> v") prefer 2
      apply (rule prems(5))
      apply (erule normed.Xor)
      apply force apply force apply force apply force
      apply (rule_tac A1="x" in Red_cong_trans)
      apply force apply assumption
      apply force
      done
  qed
qed

lemma norm_reduce: "x ~> norm x"
  apply (induct x)
  apply (auto intro: xor_red.refl)
  apply (erule xor_red.Hash_cong)
  apply (erule xor_red.MPair_cong)
  apply force
  apply (erule xor_red.Crypt_cong)
  apply (rule_tac A1="norm x1" and B1="norm x2" in Red_cong_trans)
  apply auto
  apply (rule normxor_reduce)
  apply (rule normed_norm)+
done

subsubsection {* fparts/subterm and norm interaction *}

lemma fsubterms_norm_NONCE:
  "\<lbrakk> NONCE C N \<in> fsubterms {norm B} \<rbrakk> \<Longrightarrow> NONCE C N \<in> fsubterms {B}"
  apply (rule fsubterms_reduce_NONCE)
  prefer 2
  apply assumption
  apply (rule norm_reduce)
done

lemma fsubterms_norm_KEY:
  "\<lbrakk> KEY k \<in> fsubterms {norm B} \<rbrakk> \<Longrightarrow> KEY k \<in> fsubterms {B}"
  apply (rule fsubterms_reduce_KEY)
  prefer 2
  apply assumption
  apply (rule norm_reduce)
done

lemma fsubterms_norm_AGENT:
  "\<lbrakk> AGENT C \<in> fsubterms {norm B} \<rbrakk> \<Longrightarrow> AGENT C \<in> fsubterms {B}"
  apply (rule fsubterms_reduce_AGENT)
  prefer 2
  apply assumption
  apply (rule norm_reduce)
done

lemma fparts_norm_KEY:
  "\<lbrakk> KEY k \<in> fparts {norm B} \<rbrakk> \<Longrightarrow> KEY k \<in> fparts {B}"
  apply (rule fparts_reduce_KEY)
  prefer 2
  apply assumption
  apply (rule norm_reduce)
done

lemma fparts_norm_NONCE:
  "\<lbrakk> NONCE a na \<in> fparts {norm B} \<rbrakk> \<Longrightarrow> NONCE a na \<in> fparts {B}"
  apply (rule fparts_reduce_NONCE)
  prefer 2
  apply assumption
  apply (rule norm_reduce)
done

lemma fsubterms_norm_CRYPT:
  "\<lbrakk> CRYPT k m \<in> fsubterms {norm X} \<rbrakk> \<Longrightarrow> \<exists> m'. CRYPT k m' \<in> fsubterms {X} \<and> norm m' = m"
  apply (subgoal_tac "X ~> norm X") prefer 2
  apply (rule norm_reduce) 
  apply (drule fsubterms_reduce_CRYPT)
  apply assumption
  apply auto
  apply (rule_tac x=msig' in exI)
  apply (rule conjI)
  apply force
  apply (drule xor_red_imp_xor_eq)
  apply (drule equiv_imp_norm)
  apply (subgoal_tac "normed m")
  apply (force simp add: norm_normed_id)
  apply (subgoal_tac "m \<in> fsubterms {norm X}")
  apply (erule normed_fsubterms)
  apply (rule normed_norm)
  apply (erule fsubterms.Ctext)
done

lemma fsubterms_norm_HASH:
  "\<lbrakk> HASH m \<in> fsubterms {norm X} \<rbrakk> \<Longrightarrow> \<exists> m'. HASH m' \<in> fsubterms {X} \<and> norm m' = m"
  apply (subgoal_tac "X ~> norm X") prefer 2
  apply (rule norm_reduce) 
  apply (drule fsubterms_reduce_HASH)
  apply assumption
  apply auto
  apply (rule_tac x=m' in exI)
  apply (rule conjI)
  apply force
  apply (drule xor_red_imp_xor_eq)
  apply (drule equiv_imp_norm)
  apply (subgoal_tac "normed m")
  apply (force simp add: norm_normed_id)
  apply (subgoal_tac "m \<in> fsubterms {norm X}")
  apply (erule normed_fsubterms)
  apply (rule normed_norm)
  apply (erule fsubterms.Hash)
done

lemma fsubterms_norm_MPAIR:
  "\<lbrakk> MPAIR a b \<in> fsubterms {norm X} \<rbrakk> \<Longrightarrow> \<exists> a' b'. MPAIR a' b' \<in> fsubterms {X} \<and> norm a' = a \<and> norm b' = b"
  apply (subgoal_tac "X ~> norm X") prefer 2
  apply (rule norm_reduce) 
  apply (drule fsubterms_reduce_MPAIR)
  apply assumption
  apply auto
  apply (rule_tac x=a' in exI, rule_tac x=b' in exI)
  apply (rule conjI)
  apply auto
  apply (drule xor_red_imp_xor_eq)
  apply (drule equiv_imp_norm)
  apply (subgoal_tac "normed a")
  apply (force simp add: norm_normed_id)
  apply (subgoal_tac "a \<in> fsubterms {norm X}")
  apply (erule normed_fsubterms)
  apply (rule normed_norm)
  apply (erule fsubterms.Fst)
  apply (drule xor_red_imp_xor_eq) back
  apply (drule equiv_imp_norm)
  apply (subgoal_tac "normed b")
  apply (force simp add: norm_normed_id)
  apply (subgoal_tac "b \<in> fsubterms {norm X}")
  apply (erule normed_fsubterms)
  apply (rule normed_norm)
  apply (erule fsubterms.Snd)
done

subsection {* message derivation *}

inductive_set
  DM :: "agent \<Rightarrow> msg set => msg set" 
  for A :: agent and H :: "msg set" where
    Inj [intro,simp] :    "X \<in> H ==> X \<in> DM A H"
  | Fst:     "MPair X Y \<in> DM A H ==> X \<in> DM A H"
  | Snd:     "MPair X Y \<in> DM A H ==> Y \<in> DM A H"
  | Nonce  [intro]:  "Nonce A n \<in> DM A H"
  | Agent  [intro]:   "Agent agt \<in> DM A H"
  | Number [intro]:   "Number n  \<in> DM A H"
  | Real   [intro]:   "Real n  \<in> DM A H"
  | Hash   [intro]:   "X \<in> DM A H ==> Hash X \<in> DM A H"
  | MPair  [intro]:   "[|X \<in> DM A H;  Y \<in> DM A H|] ==> MPair X Y  \<in> DM A H"
  | Crypt  [intro]:   "[|X \<in> DM A H;  Key(K) \<in> DM A H|] ==> Crypt K X \<in> DM A H"
  | Xor    [intro]:   "[|X \<in> DM A H;  Y \<in> DM A H|] ==> Xor X Y \<in> DM A H"
  | Decrypt:
    "[|Crypt K X \<in> DM A H; Key(invKey K) \<in>  DM A H|]
    ==> X \<in> DM A H"


lemmas constructor_defs = Nonce_def Number_def Key_def Agent_def Hash_def
                          MPair_def Crypt_def Xor_def Real_def Zero_def

subsubsection {* Freeness of all constructors besides Xor *}

lemma Nonce_Number_ineq: "Nonce a na \<noteq> Number n"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Nonce_Key_ineq: "Nonce a na \<noteq> Key k"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Nonce_Zero_ineq: "Nonce a na \<noteq> Zero"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Nonce_Agent_ineq: "Nonce a na \<noteq> Agent b"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Nonce_Real_ineq: "Nonce a na \<noteq> Real b"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Nonce_Hash_ineq: "Nonce a na \<noteq> Hash h"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Nonce_MACM_ineq: "Nonce a na \<noteq> Hash[k] x"
by (auto simp add: constructor_defs MACM_def dest!: Abs_eq_normed)

lemma Nonce_MPair_ineq: "Nonce a na \<noteq> MPair x y"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Nonce_Crypt_ineq: "Nonce a na \<noteq> Crypt k m"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Key_Number_ineq: "Key k \<noteq> Number n"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Key_Zero_ineq: "Key k \<noteq> Zero"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Key_Agent_ineq: "Key k \<noteq> Agent b"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Key_Real_ineq: "Key k \<noteq> Real b"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Key_Hash_ineq: "Key k \<noteq> Hash h"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Key_MACM_ineq: "Key k \<noteq> Hash[kh] h"
by (auto simp add: constructor_defs MACM_def dest!: Abs_eq_normed)

lemma Key_MPair_ineq: "Key k \<noteq> MPair x y"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Key_Crypt_ineq: "Key k' \<noteq> Crypt k m"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Crypt_Number_ineq: "Crypt k m \<noteq> Number n"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Crypt_Zero_ineq: "Crypt k m \<noteq> Zero"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Crypt_Agent_ineq: "Crypt k m \<noteq> Agent b"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Crypt_Real_ineq: "Crypt k m \<noteq> Real b"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Crypt_Hash_ineq: "Crypt k m \<noteq> Hash h"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Crypt_MACM_ineq: "Crypt k m \<noteq> Hash[hk] h"
by (auto simp add: constructor_defs MACM_def dest!: Abs_eq_normed)

lemma Crypt_MPair_ineq: "Crypt k m \<noteq> MPair x y"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Number_Agent_ineq: "Number n \<noteq> Agent b"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Number_Real_ineq: "Number n \<noteq> Real b"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Number_Hash_ineq: "Number n \<noteq> Hash h"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Number_Zero_ineq: "Number n \<noteq> Zero"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Number_MACM_ineq: "Number n \<noteq> Hash[hk] h"
by (auto simp add: constructor_defs MACM_def dest!: Abs_eq_normed)

lemma Number_MPair_ineq: "Number n \<noteq> MPair x y"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Agent_Real_ineq: "Agent a \<noteq> Real b"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Agent_Zero_ineq: "Agent a \<noteq> Zero"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Agent_Hash_ineq: "Agent a \<noteq> Hash h"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Agent_MACM_ineq: "Agent a \<noteq> Hash[hk] h"
by (auto simp add: constructor_defs MACM_def dest!: Abs_eq_normed)

lemma Agent_MPair_ineq: "Agent a \<noteq> MPair x y"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Real_Hash_ineq: "Real a \<noteq> Hash h"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Real_MACM_ineq: "Real a \<noteq> Hash[hk] h"
by (auto simp add: constructor_defs MACM_def dest!: Abs_eq_normed)

lemma Real_MPair_ineq: "Real a \<noteq> MPair x y"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Real_Zero_ineq: "Real a \<noteq> Zero"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Hash_MPair_ineq: "Hash h \<noteq> MPair x y"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Hash_Zero_ineq: "Hash h \<noteq> Zero"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma MACM_Hash_ineq: "Hash[hk] m \<noteq> Hash h"
by (auto simp add: constructor_defs MACM_def dest!: Abs_eq_normed)

lemmas constructors_ineq = Nonce_Number_ineq Nonce_Key_ineq Nonce_Agent_ineq Nonce_Real_ineq Nonce_Zero_ineq
                           Nonce_Hash_ineq Nonce_MACM_ineq Nonce_MPair_ineq Nonce_Crypt_ineq
                           Key_Number_ineq Key_Agent_ineq Key_Real_ineq Key_Hash_ineq Key_Zero_ineq
                           Key_MACM_ineq Key_MPair_ineq Key_Crypt_ineq Crypt_Number_ineq Crypt_Zero_ineq
                           Crypt_Agent_ineq Crypt_Real_ineq Crypt_Hash_ineq Crypt_MACM_ineq
                           Crypt_MPair_ineq Number_Agent_ineq Number_Real_ineq Number_Hash_ineq Number_Zero_ineq
                           Number_MACM_ineq Number_MPair_ineq Agent_Real_ineq Agent_Hash_ineq Agent_Zero_ineq
                           Agent_MACM_ineq Agent_MPair_ineq Real_Hash_ineq Real_MACM_ineq Real_Zero_ineq
                           Real_MPair_ineq Hash_MPair_ineq Hash_Zero_ineq MACM_Hash_ineq

declare constructors_ineq[iff]
declare constructors_ineq[symmetric,iff]

lemma Nonce_inject[dest!]: "Nonce a na = Nonce b nb \<Longrightarrow> a = b \<and> na = nb"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Key_inject[dest!]: "Key ka = Key kb \<Longrightarrow> ka = kb"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Agent_inject[dest!]: "Agent a = Agent b \<Longrightarrow> a = b"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Number_inject[dest!]: "Number a = Number b \<Longrightarrow> a = b"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Real_inject[dest!]: "Real a = Real b \<Longrightarrow> a = b"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

lemma Rep_msg_inj[dest]: "Rep_msg a = Rep_msg b \<Longrightarrow> a = b"
  apply (drule_tac f=Abs_msg in arg_cong)
  apply (auto simp add: Rep_msg_inverse)
done

lemma Hash_inject[dest!]: "Hash a = Hash b \<Longrightarrow> a = b"
  apply (auto simp add: constructor_defs dest!: Abs_eq_normed Rep_msg_inj)
done

lemma MPair_inject[dest!]: "MPair a b = MPair c d \<Longrightarrow> a = c \<and> b = d"
  apply (auto simp add: constructor_defs dest!: Abs_eq_normed Rep_msg_inj)
done

lemma Crypt_inject[dest!]: "Crypt ka ma = Crypt kb mb \<Longrightarrow> ka = kb \<and> ma = mb"
  apply (auto simp add: constructor_defs dest!: Abs_eq_normed Rep_msg_inj)
done

lemma parts_mono_elem:
  "\<lbrakk> X \<in> parts H; H \<subseteq> G \<rbrakk> \<Longrightarrow> X \<in> parts G"
  apply (drule parts.mono)
by (erule rev_subsetD)

lemma subterms_mono_elem:
  "\<lbrakk> X \<in> subterms H; H \<subseteq> G \<rbrakk> \<Longrightarrow> X \<in> subterms G"
  apply (drule subterms.mono)
by (erule rev_subsetD)

lemma Rep_Abs_norm[simp]: "Rep_msg (Abs_msg (norm x)) = norm x"
  apply (subgoal_tac "normed (norm x)") prefer 2
  apply (rule normed_norm)
  apply (simp only: Abs_msg_normed)
done

subsubsection {* interaction of DM with subterms/parts *}

lemma nonce_DM_subterms_nonce:
   "\<lbrakk> Nonce B NB \<in> subterms (DM A H); A \<noteq> B \<rbrakk>
  \<Longrightarrow> Nonce B NB \<in> subterms H"
  apply (drule subterms.singleton)
  apply auto
  apply (erule rev_mp)
  apply (rotate_tac 1)
  apply (erule rev_mp)
  apply (erule DM.induct)
  apply (auto elim: subterms_mono_elem)
  apply (auto simp add: subterms_def)
  apply (subgoal_tac "m=NONCE B NB") prefer 2
  apply (rule sym)
  apply (simp add: Nonce_def)
  apply (drule normed_fsubterms)
  apply force
  apply force
  apply (rule_tac x="NONCE B NB" in exI)
  apply (unfold Xor_def)
  apply (simp only: Rep_Abs_norm)
  apply (drule fsubterms_norm_NONCE)
  apply auto
  apply (drule fsubterms_singleton)
  apply auto
done

lemma nonce_DM_parts_nonce:
   "\<lbrakk> Nonce B NB \<in> parts (DM A H); A \<noteq> B \<rbrakk>
  \<Longrightarrow> Nonce B NB \<in> parts H"
  apply (drule parts.singleton)
  apply auto
  apply (erule rev_mp)
  apply (rotate_tac 1)
  apply (erule rev_mp)
  apply (erule DM.induct)
  apply (auto elim: parts_mono_elem)
  apply (auto simp add: parts_def)
  apply (subgoal_tac "m=NONCE B NB") prefer 2
  apply (rule sym)
  apply (simp add: Nonce_def)
  apply (drule normed_fparts)
  apply force
  apply force
  apply (rule_tac x="NONCE B NB" in exI)
  apply (unfold Xor_def)
  apply (simp only: Rep_Abs_norm)
  apply (drule fparts_norm_NONCE)
  apply auto
  apply (drule fparts_singleton)
  apply auto
done

lemma key_DM_parts_key:
   "\<lbrakk> Key k \<in> parts (DM A H) \<rbrakk>
  \<Longrightarrow> Key k \<in> parts H"
  apply (drule parts.singleton)
  apply auto
  apply (rotate_tac 1)
  apply (erule rev_mp)
  apply (erule DM.induct)
  apply (auto elim: parts_mono_elem)
  apply (auto simp add: parts_def)
  apply (subgoal_tac "m=KEY k") prefer 2
  apply (rule sym)
  apply (simp add: Key_def)
  apply (drule normed_fparts)
  apply force
  apply force
  apply (rule_tac x="KEY k" in exI)
  apply (unfold Xor_def)
  apply (simp only: Rep_Abs_norm)
  apply (drule fparts_norm_KEY)
  apply auto
  apply (drule fparts_singleton)
  apply auto
done

declare normed_norm[iff]

lemma crypt_DM_parts_crypt_key:
   "\<lbrakk> Crypt k m \<in> subterms (DM A H) \<rbrakk>
  \<Longrightarrow> Crypt k m \<in> subterms H \<or> Key k \<in> parts H "
  apply (drule subterms.singleton)
  apply auto
  apply (rotate_tac 1)
  apply (erule rev_mp)
  apply (erule rev_mp)
  apply (erule DM.induct)
  apply (auto elim: subterms_mono_elem)
  apply (rotate_tac 2)
  apply (erule contrapos_np)
  apply (rule_tac A=A in key_DM_parts_key)
  apply force
  apply (simp add: subterms_def)
  apply (unfold Xor_def)
  apply (elim exE conjE)
  apply (simp only: Rep_Abs_norm)
  apply (simp only: Crypt_def)
  apply (subgoal_tac "normed (CRYPT k (Rep_msg m))") prefer 2
  apply force
  apply (subgoal_tac "normed ma") prefer 2
  apply (frule normed_fsubterms)
  apply force
  apply force
  apply (drule  Abs_eq_normed)
  apply force
  apply force
  apply (subgoal_tac "CRYPT k (Rep_msg m)
                      \<in> fsubterms {MessageTheoryXor.norm (Rep_msg X \<oplus> Rep_msg Ya)}")
  apply (drule fsubterms_norm_CRYPT) prefer 2
  apply force
  apply auto
  apply (drule fsubterms.singleton) back
  apply auto
  apply (subgoal_tac "norm m' = m'") prefer 2
  apply (rule norm_normed_id)
  apply (rule normed_fsubterms)
  apply (erule fsubterms.Ctext)
  apply force
  apply force
  apply (subgoal_tac "norm m' = m'") prefer 2
  apply (rule norm_normed_id)
  apply (rule normed_fsubterms)
  apply (erule fsubterms.Ctext)
  apply force
  apply force
done

lemma mac_DM_parts_mac_key:
   "\<lbrakk> Hash (MPair (Key k) m) \<in> subterms (DM A H) \<rbrakk>
  \<Longrightarrow> Hash (MPair (Key k) m) \<in> subterms H \<or> Key k \<in> parts H "
  apply (drule subterms.singleton)
  apply auto
  apply (rotate_tac 1)
  apply (erule rev_mp)
  apply (erule rev_mp)
  apply (erule DM.induct)
  apply (auto elim: subterms_mono_elem)
  apply (rotate_tac 1)
  apply (erule contrapos_np)
  apply (rule_tac A=A in key_DM_parts_key)
  apply (drule DM.Fst)
  apply (erule parts.inj)
  apply (simp add: subterms_def)
  apply (unfold Xor_def)
  apply (elim exE conjE)
  apply (simp only: Rep_Abs_norm)
  apply (simp only: Hash_def)
  apply (subgoal_tac "normed (HASH (Rep_msg \<lbrace>Key k, m\<rbrace>))") prefer 2
  apply force
  apply (subgoal_tac "normed ma") prefer 2
  apply (frule normed_fsubterms)
  apply force
  apply force
  apply (drule Abs_eq_normed)
  apply force
  apply force
  apply (rule_tac x="HASH (Rep_msg \<lbrace>Key k, m\<rbrace>)" in exI)
  apply (rule conjI)
  apply force
  apply (clarsimp simp only: Zero_def)
  apply (drule fsubterms_norm_HASH)
  apply (elim exE conjE)
  apply auto
  apply (drule fsubterms.singleton)
  apply auto
  apply (subgoal_tac "norm m' = m'") prefer 2
  apply (rule norm_normed_id)
  apply (rule normed_fsubterms)
  apply (erule fsubterms.Hash)
  apply force
  apply simp
  apply (subgoal_tac "norm m' = m'") prefer 2
  apply (rule norm_normed_id)
  apply (rule normed_fsubterms)
  apply (erule fsubterms.Hash)
  apply force
  apply force
done

inductive_set LowHamXor :: "msg set"
 where
    Agent:   "(Agent a) \<in> LowHamXor"
  | Number:  "(Number n) \<in> LowHamXor"
  | Real:    "(Real r) \<in> LowHamXor"
  | Zero:     "Zero \<in> LowHamXor"
  | Xor:     "\<lbrakk> a \<in> LowHamXor; b \<in> LowHamXor \<rbrakk> \<Longrightarrow> Xor a b \<in> LowHamXor"

lemma parts_Key_Xor: "Key k \<in> parts {Xor a b} \<Longrightarrow> Key k \<in> parts {a,b}"
  apply (simp add: parts_def Key_def)
  apply auto
  apply (unfold Xor_def)
  apply (subgoal_tac "(MessageTheoryXor.norm (Rep_msg a \<oplus> Rep_msg b)) \<in> msg")
  apply (simp only: Abs_msg_inverse) prefer 2
  apply (simp only: msg_def)
  apply (subgoal_tac "normed (norm (Rep_msg a \<oplus> Rep_msg b))") prefer 2
  apply (force simp add: normed_norm)
  apply force
  apply (rule_tac x = m in exI)
  apply (rule conjI)
  apply force
  apply (subgoal_tac "KEY k \<in> msg")
  apply (subgoal_tac "m \<in> msg")
  apply (simp only: Abs_msg_inject) defer
  apply (subgoal_tac "normed m")
  apply (force simp add: msg_def)
  apply (rule normed_fparts)
  apply assumption
  apply force
  apply (force simp add: msg_def)
  apply (subgoal_tac "m = KEY k") defer
  
  apply force
  apply (simp only:)
  apply (drule fparts_norm_KEY)
  apply force
  done

lemma subterms_Key_Xor: "Key k \<in> subterms {Xor a b} \<Longrightarrow> Key k \<in> subterms {a,b}"
  apply (simp add: subterms_def Key_def)
  apply auto
  apply (unfold Xor_def)
  apply (subgoal_tac "(MessageTheoryXor.norm (Rep_msg a \<oplus> Rep_msg b)) \<in> msg")
  apply (simp only: Abs_msg_inverse) prefer 2
  apply (simp only: msg_def)
  apply (subgoal_tac "normed (norm (Rep_msg a \<oplus> Rep_msg b))") prefer 2
  apply (force simp add: normed_norm)
  apply force
  apply (rule_tac x = m in exI)
  apply (rule conjI)
  apply force
  apply (subgoal_tac "KEY k \<in> msg")
  apply (subgoal_tac "m \<in> msg")
  apply (simp only: Abs_msg_inject) defer
  apply (subgoal_tac "normed m")
  apply (force simp add: msg_def)
  apply (rule normed_fsubterms)
  apply assumption
  apply force
  apply (force simp add: msg_def)
  apply (subgoal_tac "m = KEY k") defer
  
  apply force
  apply (simp only:)
  apply (drule fsubterms_norm_KEY)
  apply force
  done


lemma subterms_Nonce_Xor: "Nonce D ND \<in> subterms {Xor a b} \<Longrightarrow> Nonce D ND \<in> subterms {a,b}"
  apply (simp add: subterms_def Nonce_def)
  apply auto
  apply (unfold Xor_def)
  apply (subgoal_tac "(MessageTheoryXor.norm (Rep_msg a \<oplus> Rep_msg b)) \<in> msg")
  apply (simp only: Abs_msg_inverse) prefer 2
  apply (simp only: msg_def)
  apply (subgoal_tac "normed (norm (Rep_msg a \<oplus> Rep_msg b))") prefer 2
  apply (force simp add: normed_norm)
  apply force
  apply (rule_tac x = m in exI)
  apply (rule conjI)
  apply force
  apply (subgoal_tac "NONCE D ND \<in> msg")
  apply (subgoal_tac "m \<in> msg")
  apply (simp only: Abs_msg_inject) defer
  apply (subgoal_tac "normed m")
  apply (force simp add: msg_def)
  apply (rule normed_fsubterms)
  apply assumption
  apply force
  apply (force simp add: msg_def)
  apply (subgoal_tac "m = NONCE D ND") defer
  
  apply force
  apply (simp only:)
  apply (drule fsubterms_norm_NONCE)
  apply force
  done

lemma subterms_Hash_Xor: "Hash m \<in> subterms {Xor a b} \<Longrightarrow> Hash m \<in> subterms {a,b}"
  apply (simp add: subterms_def Hash_def)
  apply auto
  apply (unfold Xor_def)
  apply (subgoal_tac "(MessageTheoryXor.norm (Rep_msg a \<oplus> Rep_msg b)) \<in> msg")
  apply (simp only: Abs_msg_inverse) prefer 2
  apply (simp only: msg_def)
  apply (subgoal_tac "normed (norm (Rep_msg a \<oplus> Rep_msg b))") prefer 2
  apply (force simp add: normed_norm)
  apply force
  apply (rule_tac x = ma in exI)
  apply (rule conjI)
  apply force
  apply (subgoal_tac "HASH (Rep_msg m) \<in> msg")
  apply (subgoal_tac "ma \<in> msg")
  apply (simp only: Abs_msg_inject) defer
  apply (subgoal_tac "normed ma")
  apply (force simp add: msg_def)
  apply (rule normed_fsubterms)
  apply assumption
  apply force
  apply (force simp add: msg_def)
  apply (subgoal_tac "ma = HASH (Rep_msg m)") defer
  
  apply force
  apply (simp only:)
  apply (drule fsubterms_norm_HASH)
  apply auto
  apply (subgoal_tac "normed (HASH m')") prefer 2
  apply (drule fsubterms_singleton)
  apply auto
  apply (rule normed_fsubterms)
  apply force
  apply force
  apply (rule normed_fsubterms)
  apply force
  apply force
  apply (subgoal_tac "normed m'")
  apply (simp add: norm_normed_id)
  apply (subgoal_tac "m' = norm m'")
  apply force
  apply (drule norm_normed_id)
  apply force
  done



lemma subterms_Crypt_Xor: "Crypt c d \<in> subterms {Xor a b} \<Longrightarrow> Crypt c d \<in> subterms {a,b}"
  apply (simp add: subterms_def Crypt_def)
  apply auto
  apply (unfold Xor_def)
  apply (subgoal_tac "(MessageTheoryXor.norm (Rep_msg a \<oplus> Rep_msg b)) \<in> msg")
  apply (simp only: Abs_msg_inverse) prefer 2
  apply (simp only: msg_def)
  apply (subgoal_tac "normed (norm (Rep_msg a \<oplus> Rep_msg b))") prefer 2
  apply (force simp add: normed_norm)
  apply force
  apply (rule_tac x = m in exI)
  apply (rule conjI)
  apply force
  apply (subgoal_tac "CRYPT c (Rep_msg d) \<in> msg")
  apply (subgoal_tac "m \<in> msg")
  apply (simp only: Abs_msg_inject) defer
  apply (subgoal_tac "normed m")
  apply (force simp add: msg_def)
  apply (rule normed_fsubterms)
  apply assumption
  apply force
  apply (force simp add: msg_def)
  apply (subgoal_tac "m = CRYPT c (Rep_msg d)") defer
  
  apply force
  apply (simp only:)
  apply (drule fsubterms_norm_CRYPT)
  apply auto
  apply (subgoal_tac "normed (CRYPT c m')") prefer 2
  apply (drule fsubterms_singleton)
  apply auto
  apply (rule normed_fsubterms)
  apply force
  apply force
  apply (rule normed_fsubterms)
  apply force
  apply force
  apply (subgoal_tac "normed m'")
  apply (simp add: norm_normed_id)
  apply (subgoal_tac "m' = norm m'")
  apply force
  apply (drule norm_normed_id)
  apply force
  done

lemma parts_Zero[simp]: "parts {Zero} = {Zero}"
  apply (auto simp add: parts_def Zero_def)
  apply (subgoal_tac "normed ZERO", auto)+
done


lemma subterms_Zero[simp]: "subterms {Zero} = {Zero}"
  apply (auto simp add: subterms_def Zero_def)
  apply (subgoal_tac "normed ZERO", auto)+
done

lemma key_notin_parts_LowHam: "\<not> (Key k \<in> parts LowHamXor)"
proof -
  {
    fix x :: "msg"
    fix y :: "msg"
    assume "y \<in> LowHamXor" and "x \<in> parts {y}"
    hence "\<forall> k. x \<noteq> Key k"
      proof (induct y)
        case (Agent a)
        show ?case using prems by simp
      next
        case (Number r)
        show ?case using prems by simp
      next
        case Zero
        show ?case using prems by simp
      next
        case (Real r)
        show ?case using prems by simp
      next
        case (Xor a b)
        show ?case using prems
          apply auto
          apply (drule parts_Key_Xor)
          apply (drule_tac H="{a,b}" in parts.singleton)
          by auto
      qed
  }
  thus ?thesis apply -
    apply auto
    apply  (drule parts.singleton)
    apply auto
    done
qed

lemma key_notin_subterms_LowHam: "\<not> (Key k \<in> subterms LowHamXor)"
proof -
  {
    fix x :: "msg"
    fix y :: "msg"
    assume "y \<in> LowHamXor" and "x \<in> subterms {y}"
    hence "\<forall> k. x \<noteq> Key k"
      proof (induct y)
        case (Agent a)
        show ?case using prems by simp
      next
        case (Number r)
        show ?case using prems by simp
      next
        case Zero
        show ?case using prems by simp
      next
        case (Real r)
        show ?case using prems by simp
      next
        case (Xor a b)
        show ?case using prems
          apply auto
          apply (drule subterms_Key_Xor)
          apply (drule_tac H="{a,b}" in subterms.singleton)
          by auto
      qed
  }
  thus ?thesis apply -
    apply auto
    apply  (drule subterms.singleton)
    apply auto
    done
qed

lemma nonce_notin_subterms_LowHam: "\<not> (Nonce D ND \<in> subterms LowHamXor)"
proof -
  {
    fix x :: "msg"
    fix y :: "msg"
    assume "y \<in> LowHamXor" and "x \<in> subterms {y}"
    hence "\<forall> D ND. x \<noteq> Nonce D ND"
      proof (induct y)
        case (Agent a)
        show ?case using prems by simp
      next
        case (Number r)
        show ?case using prems by simp
      next
        case Zero
        show ?case using prems by simp
      next
        case (Real r)
        show ?case using prems by simp
      next
        case (Xor a b)
        show ?case using prems
          apply auto
          apply (drule subterms_Nonce_Xor)
          apply (drule_tac H="{a,b}" in subterms.singleton)
          by auto
      qed
  }
  thus ?thesis apply -
    apply auto
    apply  (drule subterms.singleton)
    apply auto
    done
qed


lemma hash_notin_subterms_LowHam: "\<not> (Hash m \<in> subterms LowHamXor)"
proof -
  {
    fix x :: "msg"
    fix y :: "msg"
    assume "y \<in> LowHamXor" and "x \<in> subterms {y}"
    hence "\<forall> D ND. x \<noteq> Hash m"
      proof (induct y)
        case (Agent a)
        show ?case using prems by simp
      next
        case Zero
        show ?case using prems by simp
      next
        case (Number r)
        show ?case using prems by simp
      next
        case (Real r)
        show ?case using prems by simp
      next
        case (Xor a b)
        show ?case using prems
          apply auto
          apply (drule subterms_Hash_Xor)
          apply (drule_tac H="{a,b}" in subterms.singleton)
          by auto
      qed
  }
  thus ?thesis apply -
    apply auto
    apply  (drule subterms.singleton)
    apply auto
    done
qed

lemma crypt_notin_subterms_LowHam: "\<not> (Crypt m m' \<in> subterms LowHamXor)"
proof -
  {
    fix x :: "msg"
    fix y :: "msg"
    assume "y \<in> LowHamXor" and "x \<in> subterms {y}"
    hence "\<forall> D ND. x \<noteq> Crypt m m'"
      proof (induct y)
        case (Agent a)
        show ?case using prems by simp
      next
        case (Number r)
        show ?case using prems by simp
      next
        case Zero
        show ?case using prems by simp
      next
        case (Real r)
        show ?case using prems by simp
      next
        case (Xor a b)
        show ?case using prems
          apply auto
          apply (drule subterms_Crypt_Xor)
          apply (drule_tac H="{a,b}" in subterms.singleton)
          by auto
      qed
  }
  thus ?thesis apply -
    apply auto
    apply  (drule subterms.singleton)
    apply auto
    done
qed

fun
  fcomponents :: "fmsg \<Rightarrow> fmsg set"
 where
   "fcomponents (MPAIR a b) = fcomponents a \<union> fcomponents b"
 | "fcomponents m           = {m}"


definition
  components :: "msg set \<Rightarrow> msg set"
 where
  "components H = { Abs_msg m | m n . m \<in> fcomponents (Rep_msg n) \<and> n \<in> H}"


lemma norm_Rep[simp]:
  "norm (Rep_msg m) = Rep_msg m"
  apply (subgoal_tac "normed (Rep_msg m)")
  apply (auto simp add: norm_normed_id)
done


lemma Xor_Zero:  "Xor a Zero = a"
  apply (auto simp add: Xor_def Zero_def)
  apply (subgoal_tac "normed ZERO")
  apply (simp add: Abs_msg_inverse)
  apply auto
  apply (subgoal_tac "normed (Rep_msg a)") prefer 2
  apply force
  apply (simp add: norm_normed_id)
  apply (simp add: Rep_msg_inverse)
done

lemma Xor_comm: "Xor A B = Xor B A"
  apply (auto simp add: Xor_def normxor_com)
  done

lemma Xor_assoc: " Xor (Xor A B) C = Xor A (Xor B C)"
  apply (auto simp add: Xor_def)
  apply (subgoal_tac "(Rep_msg A  \<otimes>  Rep_msg B) \<in> msg")
  apply (simp only: Abs_msg_inverse) prefer 2
  apply (auto simp add: msg_def)
  apply (rule normed_normxor)
  apply auto
  apply (subgoal_tac "(Rep_msg B  \<otimes>  Rep_msg C) \<in> msg")
  apply (simp only: Abs_msg_inverse) prefer 2
  apply (auto simp add: msg_def)
  apply (rule normed_normxor)
  apply auto
  apply (simp add: normxor_assoc2)
  done

lemma Xor_comm2: " Xor A (Xor B C) = Xor B (Xor A C)"
  apply (subst Xor_assoc[THEN sym])
  apply (simp add: Xor_comm[of A B])
  apply (simp add: Xor_assoc)
  done

lemma Xor_reduce[simp]: "Xor A (Xor A B) = B"
  apply (auto simp add: Xor_def)
  apply (subgoal_tac "(Rep_msg A  \<otimes>  Rep_msg B) \<in> msg")
  apply (simp only: Abs_msg_inverse) prefer 2
  apply (auto simp add: msg_def)
  apply (rule normed_normxor)
  apply auto
  apply (subgoal_tac "Rep_msg A \<otimes> Rep_msg A \<otimes> Rep_msg B
                     = norm (Rep_msg A \<oplus> Rep_msg A \<oplus> Rep_msg B)") prefer 2
  apply force
  apply (subgoal_tac "norm (Rep_msg A \<oplus> Rep_msg A \<oplus> Rep_msg B)
                     \<approx> Rep_msg A \<oplus> Rep_msg A \<oplus> Rep_msg B") prefer 2
  apply (rule norm_equiv[THEN xor_eq.symm])
  apply (subgoal_tac "Rep_msg A \<oplus> Rep_msg A \<oplus> Rep_msg B \<approx> Rep_msg B") prefer 2
  apply simp
  apply (rule_tac Y="ZERO \<oplus> Rep_msg B" in  xor_eq.trans)
  apply (rule xor_eq.trans[OF xor_eq.Xor_assoc])
  apply (rule xor_eq.Xor_cong)
  apply force
  apply force
  apply force
  apply (simp add: equiv_norm)
  apply (simp only: Rep_msg_inverse) 
  done

lemma Xor_reduce2[simp]: "Xor A (Xor B A) = B"
  apply (simp add: Xor_comm2 Xor_comm Xor_assoc)
  done

lemmas Xor_rewrite = Xor_assoc Xor_comm Xor_comm2



lemma fcomponents_imp_fparts: "x \<in> fcomponents m \<Longrightarrow> x \<in> fparts {m}"
  apply (induct m)
  apply auto
  apply (drule_tac H="{m1,m2}" in fparts.trans)
  apply auto
  apply (drule_tac H="{m1,m2}" in fparts.trans)
  apply auto
  done

lemma A1: "x \<in> components S \<Longrightarrow> x \<in> parts S"
  apply (auto simp add: components_def parts_def)
  apply (rule_tac x=m in exI)
  apply auto
  apply (drule fcomponents_imp_fparts)
  apply (drule_tac H="Rep_msg`S" in fparts.trans)
  apply auto
  done

lemma key_fcomponents_fparts:
  "KEY k \<in> fparts {m} \<Longrightarrow> \<exists>n\<in>fcomponents m. KEY k \<in> fparts {n}"
  apply (induct m)
  apply auto
  apply (drule fparts.singleton)
  apply auto
done

lemma normed_fcomponents:
  "\<lbrakk> Y \<in> fcomponents X; normed X \<rbrakk> \<Longrightarrow> normed Y"
  apply (induct X)
  apply auto
  apply (auto elim: normed_MPAIR)
  done

lemma A2: "Key k \<in> parts S \<Longrightarrow> \<exists>m\<in>components S. Key k \<in> parts {m}"
  apply (auto simp add: Key_def parts_def components_def)
  apply (drule fparts.singleton)
  apply auto
  apply (frule normed_fparts)
  apply force
  apply (subgoal_tac "KEY k \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (subgoal_tac "m \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (drule_tac f=Rep_msg in HOL.arg_cong)
  apply (auto simp add: Abs_msg_inverse)
  apply (drule key_fcomponents_fparts)
  apply auto
  apply (rule_tac x="Abs_msg n" in exI)
  apply auto
  apply (rule_tac x="KEY k" in exI)
  apply (subgoal_tac "n \<in> msg")
  apply (auto dest: normed_fcomponents simp add: Abs_msg_inverse msg_def)
  done

lemma nonce_fcomponents_fsubterms:
  "NONCE A NA \<in> fsubterms {m} \<Longrightarrow> \<exists>n\<in>fcomponents m. NONCE A NA \<in> fsubterms {n}"
  apply (induct m)
  apply auto
  apply (drule fsubterms.singleton)
  apply auto
done

lemma hash_fcomponents_fsubterms:
  "HASH c \<in> fsubterms {m} \<Longrightarrow> \<exists>n\<in>fcomponents m. HASH c \<in> fsubterms {n}"
  apply (induct m)
  apply auto
  apply (drule fsubterms.singleton)
  apply auto
done

lemma crypt_fcomponents_fsubterms:
  "CRYPT K M \<in> fsubterms {m} \<Longrightarrow> \<exists>n\<in>fcomponents m. CRYPT K M \<in> fsubterms {n}"
  apply (induct m)
  apply auto
  apply (drule fsubterms.singleton)
  apply auto
done

lemma A3: "Nonce A N \<in> subterms S \<Longrightarrow> \<exists>m\<in>components S. Nonce A N \<in> subterms {m}"
  apply (auto simp add: Nonce_def subterms_def components_def)
  apply (drule fsubterms.singleton)
  apply auto
  apply (frule normed_fsubterms)
  apply force
  apply (subgoal_tac "NONCE A N \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (subgoal_tac "m \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (drule_tac f=Rep_msg in HOL.arg_cong)
  apply (auto simp add: Abs_msg_inverse)
  apply (drule nonce_fcomponents_fsubterms)
  apply auto
  apply (rule_tac x="Abs_msg n" in exI)
  apply auto
  apply (rule_tac x="NONCE A N" in exI)
  apply (subgoal_tac "n \<in> msg")
  apply (auto dest: normed_fcomponents simp add: Abs_msg_inverse msg_def)
  done

lemma A4: "Hash c \<in> subterms S \<Longrightarrow> \<exists>m\<in>components S. Hash c \<in> subterms {m}"
  apply (auto simp add: Hash_def subterms_def components_def)
  apply (drule fsubterms.singleton)
  apply auto
  apply (frule normed_fsubterms)
  apply force
  apply (subgoal_tac "HASH (Rep_msg c) \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (subgoal_tac "m \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (drule_tac f=Rep_msg in HOL.arg_cong)
  apply (auto simp add: Abs_msg_inverse)
  apply (drule hash_fcomponents_fsubterms)
  apply auto
  apply (rule_tac x="Abs_msg n" in exI)
  apply auto
  apply (rule_tac x="HASH (Rep_msg c)" in exI)
  apply (subgoal_tac "n \<in> msg")
  apply (auto dest: normed_fcomponents simp add: Abs_msg_inverse msg_def)
  done

lemma A5: "Crypt k p \<in> subterms S \<Longrightarrow> \<exists>M\<in>components S. Crypt k p \<in> subterms {M}"
  apply (auto simp add: Crypt_def subterms_def components_def)
  apply (drule fsubterms.singleton)
  apply auto
  apply (frule normed_fsubterms)
  apply force
  apply (subgoal_tac "CRYPT k (Rep_msg p) \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (subgoal_tac "m \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (drule_tac f=Rep_msg in HOL.arg_cong)
  apply (auto simp add: Abs_msg_inverse)
  apply (drule crypt_fcomponents_fsubterms)
  apply auto
  apply (rule_tac x="Abs_msg n" in exI)
  apply auto
  apply (rule_tac x="CRYPT k (Rep_msg p)" in exI)
  apply (subgoal_tac "n \<in> msg")
  apply (auto dest: normed_fcomponents simp add: Abs_msg_inverse msg_def)
  done

interpretation MESSAGE_DERIVATION Crypt Nonce MPair Hash Number parts subterms DM LowHamXor Xor components Key
  apply (unfold_locales)
  apply (erule nonce_DM_subterms_nonce, force)
  apply (erule nonce_DM_parts_nonce, force)
  apply (erule key_DM_parts_key)
  apply (erule crypt_DM_parts_crypt_key)
  apply (erule mac_DM_parts_mac_key)

  apply (rule_tac x="Xor X Y" in bexI)
  apply (simp add: Xor_rewrite)
  apply simp
  apply (simp add: Xor_rewrite)

  apply (drule parts_Key_Xor)
  apply (drule parts.singleton)
  apply auto
  apply (drule_tac G=LowHamXor in parts_mono_elem)
  apply (auto simp add: key_notin_parts_LowHam)

  apply (drule subterms_Key_Xor)
  apply (drule subterms.singleton)
  apply auto
  apply (drule_tac G=LowHamXor in subterms_mono_elem)
  apply (auto simp add: key_notin_subterms_LowHam)

  apply (drule subterms_Nonce_Xor)
  apply (drule subterms.singleton)
  apply auto
  apply (drule_tac G=LowHamXor in subterms_mono_elem)
  apply (auto simp add: nonce_notin_subterms_LowHam)

  apply (drule subterms_Crypt_Xor)
  apply (drule subterms.singleton)
  apply auto
  apply (drule_tac G=LowHamXor in subterms_mono_elem)
  apply (auto simp add: crypt_notin_subterms_LowHam)


  apply (drule subterms_Hash_Xor)
  apply (drule subterms.singleton)
  apply auto
  apply (drule_tac G=LowHamXor in subterms_mono_elem)
  apply (auto simp add: hash_notin_subterms_LowHam)

  apply (auto intro: A1 A2 A3 A4 A5)
done

end