(*  Title:      HOL/Auth/Event
    ID:         $Id: Event.thy,v 1.36 2007/08/01 18:25:16 wenzelm Exp $
    Author:     Lawrence C Paulson, Cambridge University Computer Laboratory
                Benedikt Schmidt, ETH Zurich
    Copyright   1996  University of Cambridge
                2008-2009 ETH Zurich

Datatype of events; freshness

*)

header{*Theory of Events for Security Protocols*}

theory Event imports MessageTheory begin

datatype
  'msg event = Send  transmitter 'msg "'msg list"
             | Recv  receiver 'msg
             | Claim agent 'msg

types
  'msg trace = "(time * 'msg event) list"

text{* list.induct with time * event as elements *}

lemma trace_induct:
  "\<lbrakk>P []; \<And>t ev xs. P xs \<Longrightarrow> P ((t,ev) # xs)\<rbrakk> \<Longrightarrow> P xs"
by (rule list.induct, auto)

locale INITSTATE = MESSAGE_DERIVATION _ _ _ _ _ _ _ _ _ _ _ Key for Key :: "nat \<Rightarrow> 'msg" +
  (*Initial states of agents -- parameter of the construction*)
  fixes initState :: "agent => 'msg set"

context MESSAGE_DERIVATION begin

fun
  knows :: "[agent,'d trace] \<Rightarrow> 'd set"
 where
  knows_Nil:
  "knows A []     = {}"
| knows_Cons:
  "knows A (x#xs) =
     (case x of
        (t,Recv (Rx A' i) m) \<Rightarrow>
          if A=A' then insert m (knows A xs) else knows A xs
      | _ \<Rightarrow> knows A xs)"

subsection{*Function @{term knows}*}

(*Simplifying   
 parts(insert X (knows A evs)) = parts{X} \<union> parts(knows A evs).
  This version won't loop with the simplifier.*)

 (* standard does not work with locale *)
lemmas parts_insert_knows_A = parts.insert [of _ "knows A evs"]
lemmas subterms_insert_knows_A = subterms.insert [of _ "knows A evs"]

lemma knows_A_Send [simp]:
     "knows A ((t,Send (Tx A i) X L) # evs) = (knows A evs)"
by simp

lemma knows_A_Recv [simp]:
  "knows A ((t,Recv (Rx A i) X) # evs) = insert X (knows A evs)"
by simp

lemma knows_Recv_Other [simp]:
  "A \<noteq> A' \<Longrightarrow> knows A ((t,Recv (Rx A' i) X) # evs) = knows A evs"
by simp

lemma knows_subset_knows_Send:
  "knows A evs \<subseteq> knows A ((t,Send B X L) # evs)"
by (simp add: subset_insertI)

lemma knows_subset_knows_Claim:
  "knows A evs \<subseteq> knows A ((t,Claim B X) # evs)"
by force

lemma knows_subset_knows_Recv:
  "knows A evs \<subseteq> knows A ((t, Recv B X) # evs)"
by (simp add: subset_insertI)

text{* Everybody sees what is sent over the network *}

lemma Recv_imp_knows_A:
  assumes A: "(t,Recv (Rx A i) X) \<in> set evs" shows "X \<in> knows A evs" using A
  apply (induct evs)
  apply (simp_all (no_asm_simp) split add: event.split)
  apply (auto split: event.split)
done

end

text{*What the Agent knows is either initially known
      or included in a received message*}

definition (in INITSTATE)
  knowsI :: "[agent,'msg trace] \<Rightarrow> 'msg set" where
  "knowsI A tr = (knows A tr \<union> initState A)"

lemma (in INITSTATE) knowsI_A_imp_Recv_initState:
  assumes knowsx: "X \<in> knowsI A evs"
  shows "(\<exists> t i. (t, Recv (Rx A i) X) \<in> set evs) \<or> X \<in> initState A" using knowsx
proof (induct rule: trace_induct)
  case 1
  thus ?case by (auto simp add: knowsI_def)
next
  case (2 t ev xs)
  note prem = `X \<in> knowsI A ((t,ev) # xs)` and
       IH = `X \<in> knowsI A xs 
             \<Longrightarrow> (\<exists>t i. (t, Recv (Rx A i) X) \<in> set xs) \<or> X \<in> initState A`
  thus ?case
  proof cases
    assume "X \<in> knowsI A xs"
    with IH show ?thesis by (auto simp add: knowsI_def)
  next
    assume xn: "X \<notin> knowsI A xs"
    show ?case
    proof (cases ev)
      case (Send TA' X' L)
      with xn have "X \<notin> knowsI A ((t,ev) # xs)" by (auto simp add: knowsI_def)
      thus ?thesis using prem by contradiction
    next
      case Claim
      with prem have "X \<in> knowsI A xs" by (auto simp add: knowsI_def)
      with xn show ?thesis by contradiction
    next
      case (Recv RA' X')
      with prem and xn have xeq: "X = X'"
	by (auto split: split_if_asm simp add: knowsI_def)
      with prems xn have "\<exists>i. RA'=Rx A i"
	by (auto split: split_if_asm simp add: knowsI_def)
      with prems xeq show ?thesis by auto
    qed
  qed
qed

subsection{*Function @{term used}*}

context MESSAGE_DERIVATION begin

fun
  used :: "'msg trace \<Rightarrow> 'msg set"
 where
  used_Nil:
  "used []         = {}"
| used_Cons:
  "used ((_,ev) # evs) =
		     (case ev of
			Send T X L => subterms {X} \<union> used evs
		      | Recv T X   => used evs
		      | Claim A X  => used evs)"
    --{*The case for @{term Recv} seems anomalous, but @{term Recv} always
        follows @{term Send} in real protocols. *}

lemma Send_imp_used: "(t, Send A X L) \<in> set evs \<Longrightarrow> X \<in> used evs"
  apply (induct evs)
  apply (auto split: event.split)
done

lemma used_Send [simp]: "used ((t,Send A X L) # evs) = subterms{X} \<union> used evs"
by simp

lemma used_Claim [simp]: "used ((t,Claim A X) # evs) = used evs"
by simp

lemma used_Recv [simp]: "used ((t,Recv A X) # evs) = used evs"
by simp

lemma used_nil_subset: "used [] \<subseteq> used evs"
by simp

lemma Send_imp_parts_used:
  assumes a: "(t, Send A X L) \<in> set evs" and b: "Y \<in> subterms {X}"
  shows "Y \<in> used evs" using a b
proof (induct evs rule: trace_induct)
  case 1 thus ?case by (auto split: event.split)
next
  case (2 ts x xs)
  thus ?case  by (auto split: event.split)
qed

lemma used_Receive_nothing [simp]:
  "used ((t, Recv B m) # tr) = used tr"
by (auto simp add: used.simps split: event.split)

lemma subterms_set_used:
  assumes "(t, Send RA X L) \<in> set tr" and "Y \<in> subterms {X}"
  shows "Y \<in> used tr" using prems
proof (induct rule: trace_induct)
  case 1
  hence "False" by auto
  thus ?case by auto
next
  case (2 t' ev tr)
  show ?case
  proof cases
    assume "(t, Send RA X L) \<in> set tr"
    thus ?thesis using prems by (auto split: event.split)
  next
    assume "(t, Send RA X L) \<notin> set tr"
    with prems have "(t, Send RA X L) = (t',ev)" by auto
    thus ?thesis using `Y \<in> subterms {X}` by auto
  qed
qed

end

context INITSTATE begin

definition
  usedI :: "'msg trace \<Rightarrow> 'msg set" where
  "usedI tr = used tr \<union> (UN B. subterms (initState B))"

lemma initState_into_used: "X \<in> subterms (initState B) ==> X \<in> usedI evs"
  apply (auto simp add: usedI_def)
done

lemma usedI_Send [simp]:
  "usedI ((t,Send A X L) # evs) = subterms{X} \<union> usedI evs"
  apply (simp add: usedI_def used.simps)
by auto

lemma usedI_Claim [simp]: "usedI ((t,Claim A X) # evs) = usedI evs"
by (simp add: usedI_def used.simps)

lemma usedI_Recv [simp]: "usedI ((t,Recv A X) # evs) = usedI evs"
by (simp add: usedI_def used.simps)

lemma usedI_nil_subset: "usedI [] \<subseteq> usedI evs"
  apply (simp add: usedI_def)
done

lemma knowsI_subset_knows_Cons: "knowsI A evs \<subseteq> knowsI A (e # evs)"
by (induct e, auto simp: knowsI_def knows.simps split: event.split)

lemma initState_subset_knowsI: "initState A \<subseteq> knowsI A evs"
  apply (auto simp add: knowsI_def)
done

end

lemma (in MESSAGE_DERIVATION) knows_subset_knows_Cons:
  "knows A evs \<subseteq> knows A (e # evs)"
by (induct e, auto simp: knows.simps split: event.split)


lemma (in MESSAGE_DERIVATION) Send_imp_used_parts:
  "(Y \<in> subterms {X} \<and> (t, Send A X L) \<in> set evs)
   \<Longrightarrow> Y \<in> used evs"
  apply (induct evs)
  apply (auto split: event.split)
done

lemma (in MESSAGE_DERIVATION) Used_imp_send_parts:
  "Y \<in> used evs \<Longrightarrow> (\<exists> X t A L. Y \<in> subterms {X} \<and> (t, Send A X L) \<in> set evs)"
  apply (induct evs)
  apply (auto split: event.split)
  apply (case_tac b)
  apply auto
done

lemma (in MESSAGE_DERIVATION) used_order_irrev:
  assumes a: "set X = set Y"
  shows "used X = used Y" using a
  apply auto
  apply (drule Used_imp_send_parts)
  apply (elim exE)
  apply (rule Send_imp_used_parts)
  apply auto
  apply (drule Used_imp_send_parts)
  apply (elim exE)
  apply (rule Send_imp_used_parts)
  apply auto
done

lemma (in MESSAGE_DERIVATION) used_mono:
  assumes a: "set X \<subseteq> set Y" and b: "x \<in> used X"
  shows "x \<in> used Y" using a b
  apply -
  apply (drule Used_imp_send_parts)
  apply (elim exE)
  apply (rule Send_imp_used_parts)
  apply auto
done

lemma (in INITSTATE) usedI_mono:
  assumes a: "set X \<subseteq> set Y" and b: "x \<in> usedI X"
  shows "x \<in> usedI Y" using a b
  apply (auto simp add: usedI_def)
  apply (rule used_mono)
  apply auto
done

lemma (in MESSAGE_DERIVATION) used_time_irrev:
  assumes a: "snd`(set X) = snd`(set Y)"
  shows "used X = used Y" using a apply -
  apply auto
  apply (drule Used_imp_send_parts)
  apply (elim exE)
  apply (subgoal_tac "Send A Xa L \<in> snd`set Y") prefer 2
  apply force
  apply (subgoal_tac "\<exists> ty. (ty, Send A Xa L) \<in> set Y") prefer 2
  apply force
  apply (elim exE)
  apply (rule Send_imp_used_parts)
  apply auto
  apply (drule Used_imp_send_parts)
  apply (elim exE)
  apply (subgoal_tac "Send A Xa L \<in> snd`set X") prefer 2
  apply force
  apply (subgoal_tac "\<exists> tx. (tx, Send A Xa L) \<in> set X") prefer 2
  apply force
  apply (elim exE)
  apply (rule Send_imp_used_parts)
  apply auto
done

lemma (in INITSTATE) usedI_time_irrev:
  assumes a: "snd`(set X) = snd`(set Y)"
  shows "usedI X = usedI Y" using a
  apply (simp add: usedI_def)
  apply (rule set_un_eq)
  apply (rule used_time_irrev)
  apply auto
done

lemma (in MESSAGE_DERIVATION) used_mono_snd:
  assumes a: "snd`(set X) \<subseteq> snd`(set Y)" and
          b: "x \<in> used X"
  shows "x \<in> used Y" using a b
proof -
  let ?U = "map (\<lambda> (t,ev). (0::real,ev)) X"  and ?V = "map (\<lambda> (t,ev). (0::real,ev)) Y"
  have ux: "snd`(set ?U) = snd`(set X)" apply (auto intro: Set.rev_image_eqI) done
  have vy: "snd`(set ?V) = snd`(set Y)" apply (auto intro: Set.rev_image_eqI) done
  from a have a: "(set ?U) \<subseteq> (set ?V)" using prems apply auto
    apply (subgoal_tac "\<exists> t. (t,ba) \<in> set Y") defer
    apply force
    apply (erule exE) apply auto done
  from ux `x \<in> used X` have "x \<in> used ?U" apply -
    apply (simp only: used_time_irrev [where X=X and Y="?U"]) done
  with a have "x \<in> used ?V" apply - apply (rule used_mono [where X="?U" and Y="?V"])
    by auto
  with vy [THEN sym] show ?thesis apply -
    apply (simp only: used_time_irrev [where X="Y" and Y="?V"]) done
qed

lemma (in INITSTATE) usedI_mono_snd:
  "[| snd`(set X) \<subseteq> snd`(set Y); x \<in> usedI X|] ==> x \<in> usedI Y"
  apply (auto simp add: usedI_def)
  apply (rule used_mono_snd)
  apply auto
done

end