(*  Title:      HOL/Auth/Message
    ID:         $Id: Message.thy,v 1.47 2007/08/01 18:25:16 wenzelm Exp $
    Author:     Lawrence C Paulson, Cambridge University Computer Laboratory
                Benedikt Schmidt, ETH Zurich
    Copyright   1996  University of Cambridge
                2008-2012 ETH Zurich
Datatypes of agents and messages;
Inductive relations "parts", "analz" and "synth"
*)

header{*Message Theory for Free Algebra*}

theory MessageTheoryFree
imports
  Event (* XXX: Hack to get lemmas added to MESSAGE_DERIVATION in Event *)
  MessageTheory
begin

datatype
     msg = Agent  agent	     --{*Agent names*}
         | Number int        --{*Ordinary integers *}
         | Real   real       --{*Real Numbers, used for times, locations, ..*}
         | Nonce  agent nat 
             --{*Unguessable nonces, tagged with agent to prevent collisions*}
         | Key    key        --{*Crypto keys*}
	 | Hash   msg        --{*Hashing*}
	 | MPair  msg msg    --{*Compound messages*}
	 | Crypt  key msg    --{*Encryption, public- or shared-key*}

syntax
  "@MTuple"      :: "['a, args] => 'a * 'b"       ("(2\<lbrace>_,/ _\<rbrace>)")
  "@HMac"        :: "['a,'a] \<Rightarrow> 'a" ("(4Hash[_] /_)" [0, 1000])

syntax (xsymbols)
  "@MTuple"      :: "['a, args] => 'a * 'b"       ("(2\<lbrace>_,/ _\<rbrace>)")
  "@HMac"        :: "['a,'a] \<Rightarrow> 'a" ("(4Hash[_] /_)" [0, 1000])

interpretation MESSAGE_THEORY_DATA Key Crypt Nonce MPair Hash Number
done

translations
  "\<lbrace>x, y, z\<rbrace>"   == "\<lbrace>x, \<lbrace>y, z\<rbrace>\<rbrace>"
  "\<lbrace>x, y\<rbrace>"      == "MPair x y"
  "Hash[X] Y"  ==  "MESSAGE_THEORY_DATA.MACM MPair Hash X Y"


text{*Equations hold because constructors are injective.*}
lemma Honest_image_eq [simp]: "(Honest x \<in> Honest`A) = (x:A)"
by auto


text{*Equations hold because constructors are injective.*}
lemma Intruder_image_eq [simp]: "(Intruder x \<in> Intruder`A) = (x:A)"
by auto

lemma Key_image_eq [simp]: "(Key x \<in> Key`A) = (x\<in>A)"
by auto

lemma Nonce_Key_image_eq [simp]: "(Nonce B x \<notin> Key`A)"
by auto


definition
  keysFor :: "msg set => key set"
    --{*Keys useful to decrypt elements of a message set*}
 where
  "keysFor H == invKey ` {K. \<exists>X. Crypt K X \<in> H}"

lemma invKey_eq [simp]: "(invKey K = invKey K') = (K=K')"
apply safe
apply (drule_tac f = invKey in arg_cong, simp)
done

subsection{*keysFor operator*}

lemma keysFor_empty [simp]: "keysFor {} = {}"
by (unfold keysFor_def, blast)

lemma keysFor_Un [simp]: "keysFor (H \<union> H') = keysFor H \<union> keysFor H'"
by (unfold keysFor_def, blast)

lemma keysFor_UN [simp]: "keysFor (\<Union>i\<in>A. H i) = (\<Union>i\<in>A. keysFor (H i))"
by (unfold keysFor_def, blast)

text{*Monotonicity*}
lemma keysFor_mono: "G \<subseteq> H ==> keysFor(G) \<subseteq> keysFor(H)"
by (unfold keysFor_def, blast)

lemma keysFor_insert_Agent [simp]: "keysFor (insert (Agent A) H) = keysFor H"
by (unfold keysFor_def, auto)

lemma keysFor_insert_Nonce [simp]: "keysFor (insert (Nonce B N) H) = keysFor H"
by (unfold keysFor_def, auto)

lemma keysFor_insert_Number [simp]: "keysFor (insert (Number N) H) = keysFor H"
by (unfold keysFor_def, auto)

lemma keysFor_insert_Real [simp]: "keysFor (insert (Real N) H) = keysFor H"
by (unfold keysFor_def, auto)

lemma keysFor_insert_Key [simp]: "keysFor (insert (Key K) H) = keysFor H"
by (unfold keysFor_def, auto)

lemma keysFor_insert_Hash [simp]: "keysFor (insert (Hash X) H) = keysFor H"
by (unfold keysFor_def, auto)

lemma keysFor_insert_MPair [simp]: "keysFor (insert \<lbrace>X,Y\<rbrace> H) = keysFor H"
by (unfold keysFor_def, auto)

lemma keysFor_insert_Crypt [simp]: 
    "keysFor (insert (Crypt K X) H) = insert (invKey K) (keysFor H)"
by (unfold keysFor_def, auto)

lemma keysFor_image_Key [simp]: "keysFor (Key`E) = {}"
by (unfold keysFor_def, auto)

lemma Crypt_imp_invKey_keysFor: "Crypt K X \<in> H ==> invKey K \<in> keysFor H"
by (unfold keysFor_def, blast)


subsection{*Inductive relation "parts"*}

inductive_set
  parts :: "msg set => msg set"
  for H :: "msg set "
  where
    Inj [intro]: "X \<in> H \<Longrightarrow> X \<in> parts H"
  | Fst:         "\<lbrace>X,Y\<rbrace>   \<in> parts H ==> X \<in> parts H"
  | Snd:         "\<lbrace>X,Y\<rbrace>   \<in> parts H ==> Y \<in> parts H"
  | Body:        "Crypt K X \<in> parts H ==> X \<in> parts H"


interpretation MESSAGE_THEORY_PARTS Crypt Nonce MPair Hash Number Key parts
  apply (unfold_locales)
  apply (erule parts.Inj)
  apply (erule_tac parts.induct)
  apply (auto dest: parts.Body parts.Fst parts.Snd)
  apply (erule parts.induct)
  apply (auto dest: parts.Body parts.Fst parts.Snd)
  apply (erule parts.induct)
  apply (auto dest: parts.Body parts.Fst parts.Snd)
done

subsubsection{*Rewrite rules for pulling out atomic messages *}


lemma parts_insert_Agent [simp]:
     "parts (insert (Agent agt) H) = insert (Agent agt) (parts H)"
  apply (rule parts.insert_eq_I) 
  apply (erule parts.induct, auto dest: parts.Fst parts.Snd parts.Body) 
done

lemma parts_insert_Nonce [simp]:
     "parts (insert (Nonce B N) H) = insert (Nonce B N) (parts H)"
apply (rule parts.insert_eq_I) 
apply (erule parts.induct, auto dest: parts.Fst parts.Snd parts.Body) 
done

lemma parts_insert_Number [simp]:
     "parts (insert (Number N) H) = insert (Number N) (parts H)"
apply (rule parts.insert_eq_I) 
apply (erule parts.induct, auto dest: parts.Fst parts.Snd parts.Body) 
done

lemma parts_insert_Real [simp]:
     "parts (insert (Real N) H) = insert (Real N) (parts H)"
apply (rule parts.insert_eq_I)
apply (erule parts.induct, auto dest: parts.Fst parts.Snd parts.Body) 
done

lemma parts_insert_Key [simp]:
     "parts (insert (Key K) H) = insert (Key K) (parts H)"
apply (rule parts.insert_eq_I) 
apply (erule parts.induct, auto dest: parts.Fst parts.Snd parts.Body) 
done

lemma parts_insert_Hash [simp]:
     "parts (insert (Hash X) H) = insert (Hash X) (parts H)"
apply (rule parts.insert_eq_I) 
apply (erule parts.induct, auto dest: parts.Fst parts.Snd parts.Body) 
done

lemma parts_insert_Crypt [simp]:
     "parts (insert (Crypt K X) H) = insert (Crypt K X) (parts (insert X H))"
apply (rule equalityI)
apply (rule subsetI)
apply (erule parts.induct, auto dest: parts.Fst parts.Snd parts.Body) 
apply (blast intro: parts.Body)
done

lemma parts_insert_MPair [simp]:
     "parts (insert \<lbrace>X,Y\<rbrace> H) =  
          insert \<lbrace>X,Y\<rbrace> (parts (insert X (insert Y H)))"
apply (rule equalityI)
apply (rule subsetI)
apply (erule parts.induct, auto dest: parts.Fst parts.Snd parts.Body) 
apply (blast intro: parts.Fst parts.Snd)+
done

lemma parts_image_Key [simp]: "parts (Key`N) = Key`N"
apply auto
apply (erule parts.induct, auto)
done

subsection{*Inductive relation "subterms"*}

inductive_set
  subterms :: "msg set => msg set"
  for H :: "msg set "
  where
    Inj [intro]: "X \<in> H \<Longrightarrow> X \<in> subterms H"
  | Fst:         "\<lbrace>X,Y\<rbrace>   \<in> subterms H ==> X \<in> subterms H"
  | Snd:         "\<lbrace>X,Y\<rbrace>   \<in> subterms H ==> Y \<in> subterms H"
  | Body:        "Crypt K X \<in> subterms H ==> X \<in> subterms H"
  | Hash:        "Hash X \<in> subterms H ==> X \<in> subterms H"


interpretation MESSAGE_THEORY_SUBTERM Crypt Nonce MPair Hash Number parts Key subterms
  apply (unfold_locales)
  apply (erule subterms.Inj)
  apply (erule_tac subterms.induct)
  apply (auto dest: subterms.Body subterms.Fst subterms.Snd subterms.Hash)
  apply (erule subterms.induct)
  apply (auto dest: subterms.Body subterms.Fst subterms.Snd  subterms.Hash)
  apply (erule subterms.induct)
  apply (auto dest: subterms.Body subterms.Fst subterms.Snd  subterms.Hash)
  apply (erule parts.induct)
  apply (auto dest: subterms.Body subterms.Fst subterms.Snd  subterms.Hash)
done

subsubsection{*Rewrite rules for pulling out atomic messages *}


lemma subterms_insert_Agent [simp]:
     "subterms (insert (Agent agt) H) = insert (Agent agt) (subterms H)"
apply (rule subterms.insert_eq_I) 
apply (erule subterms.induct, auto dest: subterms.Fst subterms.Snd subterms.Body subterms.Hash)
done

lemma subterms_insert_Nonce [simp]:
     "subterms (insert (Nonce B N) H) = insert (Nonce B N) (subterms H)"
apply (rule subterms.insert_eq_I) 
apply (erule subterms.induct, auto dest: subterms.Fst subterms.Snd subterms.Body subterms.Hash)
done

lemma subterms_insert_Number [simp]:
     "subterms (insert (Number N) H) = insert (Number N) (subterms H)"
apply (rule subterms.insert_eq_I)
apply (erule subterms.induct, auto dest: subterms.Fst subterms.Snd subterms.Body subterms.Hash)
done

lemma subterms_insert_Real [simp]:
     "subterms (insert (Real N) H) = insert (Real N) (subterms H)"
apply (rule subterms.insert_eq_I)
apply (erule subterms.induct, auto dest: subterms.Fst subterms.Snd subterms.Body subterms.Hash)
done

lemma subterms_insert_Key [simp]:
     "subterms (insert (Key K) H) = insert (Key K) (subterms H)"
apply (rule subterms.insert_eq_I)
apply (erule subterms.induct, auto dest: subterms.Fst subterms.Snd subterms.Body subterms.Hash)
done

lemma subterms_insert_Hash [simp]:
     "subterms (insert (Hash X) H) = insert (Hash X) (subterms (insert X H))"
apply (rule equalityI)
apply (rule subsetI)
apply (erule subterms.induct, auto dest: subterms.Fst subterms.Snd subterms.Body subterms.Hash) 
apply (blast intro: subterms.Hash)
done

lemma subterms_insert_Crypt [simp]:
     "subterms (insert (Crypt K X) H) = insert (Crypt K X) (subterms (insert X H))"
apply (rule equalityI)
apply (rule subsetI)
apply (erule subterms.induct, auto dest: subterms.Fst subterms.Snd subterms.Body subterms.Hash)
apply (blast intro: subterms.Body)
done

lemma subterms_insert_MPair [simp]:
     "subterms (insert \<lbrace>X,Y\<rbrace> H) =  
          insert \<lbrace>X,Y\<rbrace> (subterms (insert X (insert Y H)))"
apply (rule equalityI)
apply (rule subsetI)
apply (erule subterms.induct, auto dest: subterms.Fst subterms.Snd subterms.Body subterms.Hash)
apply (blast intro: subterms.Fst subterms.Snd)+
done

lemma subterms_image_Key [simp]: "subterms (Key`N) = Key`N"
apply auto
apply (erule subterms.induct, auto)
done


subsection{* The DM (Dolev-Yao) Operator*}

text{* The inductively defined operator DM takes an agent name and a
set of messages (typically the current knowledge of the agent) as an
argument. It returns the closure of the given set of messages under
pairing, decomposition of pairs, hashing, encryption and
decryption. Furthermore it adds all terms constructable by an agent,
such as other agents' names, numbers, etc. However, neither keys nor
only nonces associated with other agents can be generated. *}

inductive_set
  DM :: "agent \<Rightarrow> msg set => msg set" 
  for A :: agent and H :: "msg set" where
    Inj [intro,simp] :    "X \<in> H ==> X \<in> DM A H"
  | Fst:     "\<lbrace>X,Y\<rbrace> \<in> DM A H ==> X \<in> DM A H"
  | Snd:     "\<lbrace>X,Y\<rbrace> \<in> DM A H ==> Y \<in> DM A H"
  | Nonce  [intro]:  "Nonce A n \<in> DM A H"
  | Agent  [intro]:   "Agent agt \<in> DM A H"
  | Number [intro]:   "Number n  \<in> DM A H"
  | Real   [intro]:   "Real n  \<in> DM A H"
  | Hash   [intro]:   "X \<in> DM A H ==> Hash X \<in> DM A H"
  | MPair  [intro]:   "[|X \<in> DM A H;  Y \<in> DM A H|] ==> \<lbrace>X,Y\<rbrace> \<in> DM A H"
  | Crypt  [intro]:   "[|X \<in> DM A H;  Key(K) \<in> DM A H|] ==> Crypt K X \<in> DM A H"
  | Decrypt:
    "[|Crypt K X \<in> DM A H; Key(invKey K) \<in>  DM A H|]
    ==> X \<in> DM A H"

lemma DM_mono:
  "G \<subseteq> H \<Longrightarrow> DM A G \<subseteq> DM A H"
  apply (auto)
  apply (erule DM.induct)
  apply (auto intro: DM.Fst DM.Snd DM.Decrypt)
done

lemma DM_inj:
  "X \<subseteq> DM A X"
  apply (auto)
done

interpretation MESSAGE_THEORY_DM Crypt Nonce MPair Hash Number parts subterms Key DM
apply (unfold_locales)
done

subsection{*Inductive relation "analz"*}

text{*Inductive definition of "analz" -- what can be broken down from a set of
    messages, including keys.  A form of downward closure.  Pairs can
    be taken apart; messages decrypted with known keys.  *}

inductive_set
  analz :: "msg set => msg set" for H :: "msg set" where
    Inj [intro,simp] :    "X \<in> H ==> X \<in> analz H"
  | Fst:     "\<lbrace>X,Y\<rbrace> \<in> analz H ==> X \<in> analz H"
  | Snd:     "\<lbrace>X,Y\<rbrace> \<in> analz H ==> Y \<in> analz H"
  | Decrypt [dest]:
    "[|Crypt K X \<in> analz H; Key(invKey K): analz H|]
    ==> X \<in> analz H"


text{*Monotonicity; Lemma 1 of Lowe's paper*}
lemma analz_mono: "G\<subseteq>H ==> analz(G) \<subseteq> analz(H)"
apply auto
apply (erule analz.induct) 
apply (auto dest: analz.Fst analz.Snd) 
done

text{*Making it safe speeds up proofs*}
lemma MPair_analz [elim!]:
     "[| \<lbrace>X,Y\<rbrace> \<in> analz H;        
             [| X \<in> analz H; Y \<in> analz H |] ==> P   
          |] ==> P"
by (blast dest: analz.Fst analz.Snd)

lemma analz_increasing: "H \<subseteq> analz(H)"
by blast

lemma analz_subset_parts: "analz H \<subseteq> parts H"
apply (rule subsetI)
apply (erule analz.induct)
apply (auto dest: parts.Fst parts.Snd parts.Body)
done 

lemmas analz_into_parts = analz_subset_parts [THEN subsetD]

lemmas not_parts_not_analz = analz_subset_parts [THEN contra_subsetD]

lemma parts_analz [simp]: "parts (analz H) = parts H"
apply (rule equalityI)
apply (rule analz_subset_parts [THEN parts.mono, THEN subset_trans], simp)
apply (blast intro: analz_increasing [THEN parts.mono, THEN subsetD])
done

lemma analz_parts [simp]: "analz (parts H) = parts H"
apply auto
apply (erule analz.induct, auto dest: parts.Fst parts.Snd parts.Body)
done

lemma analz_subset_subterms: "analz H \<subseteq> subterms H"
apply (insert parts_subset_subterms[of H] analz_subset_parts)
by auto

lemmas analz_into_subterms = analz_subset_subterms [THEN subsetD]

lemmas not_subterms_not_analz = analz_subset_subterms [THEN contra_subsetD]

lemma subterms_analz [simp]: "subterms (analz H) = subterms H"
apply (rule equalityI)
apply (rule analz_subset_subterms [THEN subterms.mono, THEN subset_trans], simp)
apply (blast intro: analz_increasing [THEN subterms.mono, THEN subsetD])
done

lemma analz_subterms [simp]: "analz (subterms H) = subterms H"
apply auto
apply (erule analz.induct, auto dest: subterms.Fst subterms.Snd subterms.Body)
done

lemmas analz_insertI = subset_insertI [THEN analz_mono, THEN [2] rev_subsetD, standard]

subsubsection{*General equational properties *}

lemma analz_empty [simp]: "analz{} = {}"
apply safe
apply (erule analz.induct, blast+)
done

text{*Converse fails: we can analz more from the union than from the 
  separate parts, as a key in one might decrypt a message in the other*}
lemma analz_Un: "analz(G) \<union> analz(H) \<subseteq> analz(G \<union> H)"
by (intro Un_least analz_mono Un_upper1 Un_upper2)

lemma analz_insert: "insert X (analz H) \<subseteq> analz(insert X H)"
by (blast intro: analz_mono [THEN [2] rev_subsetD])

subsubsection{*Rewrite rules for pulling out atomic messages *}

lemmas analz_insert_eq_I = equalityI [OF subsetI analz_insert]

lemma analz_insert_Agent [simp]:
     "analz (insert (Agent agt) H) = insert (Agent agt) (analz H)"
apply (rule analz_insert_eq_I) 
apply (erule analz.induct, auto) 
done

lemma analz_insert_Nonce [simp]:
     "analz (insert (Nonce B N) H) = insert (Nonce B N) (analz H)"
apply (rule analz_insert_eq_I) 
apply (erule analz.induct, auto) 
done

lemma analz_insert_Number [simp]:
     "analz (insert (Number N) H) = insert (Number N) (analz H)"
apply (rule analz_insert_eq_I) 
apply (erule analz.induct, auto) 
done

lemma analz_insert_Real [simp]:
     "analz (insert (Real N) H) = insert (Real N) (analz H)"
apply (rule analz_insert_eq_I) 
apply (erule analz.induct, auto) 
done

lemma analz_insert_Hash [simp]:
     "analz (insert (Hash X) H) = insert (Hash X) (analz H)"
apply (rule analz_insert_eq_I) 
apply (erule analz.induct, auto) 
done

text{*Can only pull out Keys if they are not needed to decrypt the rest*}
lemma analz_insert_Key [simp]: 
    "K \<notin> keysFor (analz H) ==>   
          analz (insert (Key K) H) = insert (Key K) (analz H)"
apply (unfold keysFor_def)
apply (rule analz_insert_eq_I) 
apply (erule analz.induct, auto) 
done

lemma analz_insert_MPair [simp]:
     "analz (insert \<lbrace>X,Y\<rbrace> H) =  
          insert \<lbrace>X,Y\<rbrace> (analz (insert X (insert Y H)))"
apply (rule equalityI)
apply (rule subsetI)
apply (erule analz.induct, auto)
apply (erule analz.induct)
apply (blast intro: analz.Fst analz.Snd)+
done

text{*Can pull out enCrypted message if the Key is not known*}
lemma analz_insert_Crypt:
     "Key (invKey K) \<notin> analz H 
      ==> analz (insert (Crypt K X) H) = insert (Crypt K X) (analz H)"
apply (rule analz_insert_eq_I) 
apply (erule analz.induct, auto)
done

lemma lemma1: "Key (invKey K) \<in> analz H ==>   
               analz (insert (Crypt K X) H) \<subseteq>  
               insert (Crypt K X) (analz (insert X H))"
apply (rule subsetI)
apply (erule_tac x = x in analz.induct, auto)
done

lemma lemma2: "Key (invKey K) \<in> analz H ==>   
               insert (Crypt K X) (analz (insert X H)) \<subseteq>  
               analz (insert (Crypt K X) H)"
apply auto
apply (erule_tac x = x in analz.induct, auto)
apply (blast intro: analz_insertI analz.Decrypt)
done

lemma analz_insert_Decrypt:
     "Key (invKey K) \<in> analz H ==>   
               analz (insert (Crypt K X) H) =  
               insert (Crypt K X) (analz (insert X H))"
by (intro equalityI lemma1 lemma2)

text{*Case analysis: either the message is secure, or it is not! Effective,
but can cause subgoals to blow up! Use with @{text "split_if"}; apparently
@{text "split_tac"} does not cope with patterns such as @{term"analz (insert
(Crypt K X) H)"} *} 
lemma analz_Crypt_if [simp]:
     "analz (insert (Crypt K X) H) =                 
          (if (Key (invKey K) \<in> analz H)                 
           then insert (Crypt K X) (analz (insert X H))  
           else insert (Crypt K X) (analz H))"
by (simp add: analz_insert_Crypt analz_insert_Decrypt)


text{*This rule supposes "for the sake of argument" that we have the key.*}
lemma analz_insert_Crypt_subset:
     "analz (insert (Crypt K X) H) \<subseteq>   
           insert (Crypt K X) (analz (insert X H))"
apply (rule subsetI)
apply (erule analz.induct, auto)
done


lemma analz_image_Key [simp]: "analz (Key`N) = Key`N"
apply auto
apply (erule analz.induct, auto)
done


subsubsection{*Idempotence and transitivity *}

lemma analz_analzD [dest!]: "X\<in> analz (analz H) ==> X\<in> analz H"
by (erule analz.induct, blast+)

lemma analz_idem [simp]: "analz (analz H) = analz H"
by blast

lemma analz_subset_iff [simp]: "(analz G \<subseteq> analz H) = (G \<subseteq> analz H)"
apply (rule iffI)
apply (iprover intro: subset_trans analz_increasing)  
apply (frule analz_mono, simp) 
done

lemma analz_trans: "[| X\<in> analz G;  G \<subseteq> analz H |] ==> X\<in> analz H"
by (drule analz_mono, blast)

text{*Cut; Lemma 2 of Lowe*}
lemma analz_cut: "[| Y\<in> analz (insert X H);  X\<in> analz H |] ==> Y\<in> analz H"
by (erule analz_trans, blast)

(*Cut can be proved easily by induction on
   "Y: analz (insert X H) ==> X: analz H --> Y: analz H"
*)

text{*This rewrite rule helps in the simplification of messages that involve
  the forwarding of unknown components (X).  Without it, removing occurrences
  of X can be very complicated. *}
lemma analz_insert_eq: "X\<in> analz H ==> analz (insert X H) = analz H"
by (blast intro: analz_cut analz_insertI)


text{*A congruence rule for "analz" *}

lemma analz_subset_cong:
     "[| analz G \<subseteq> analz G'; analz H \<subseteq> analz H' |] 
      ==> analz (G \<union> H) \<subseteq> analz (G' \<union> H')"
apply simp
apply (iprover intro: conjI subset_trans analz_mono Un_upper1 Un_upper2) 
done

lemma analz_cong:
     "[| analz G = analz G'; analz H = analz H' |] 
      ==> analz (G \<union> H) = analz (G' \<union> H')"
by (intro equalityI analz_subset_cong, simp_all) 

lemma analz_insert_cong:
     "analz H = analz H' ==> analz(insert X H) = analz(insert X H')"
by (force simp only: insert_def intro!: analz_cong)

text{*If there are no pairs or encryptions then analz does nothing*}
lemma analz_trivial:
     "[| \<forall>X Y. \<lbrace>X,Y\<rbrace> \<notin> H;  \<forall>X K. Crypt K X \<notin> H |] ==> analz H = H"
apply safe
apply (erule analz.induct, blast+)
done

lemma analz_UN_analz_lemma:
     "X\<in> analz (\<Union>i\<in>A. analz (H i)) ==> X\<in> analz (\<Union>i\<in>A. H i)"
apply (erule analz.induct)
apply (blast intro: analz_mono [THEN [2] rev_subsetD])+
done

lemma analz_UN_analz [simp]: "analz (\<Union>i\<in>A. analz (H i)) = analz (\<Union>i\<in>A. H i)"
by (blast intro: analz_UN_analz_lemma analz_mono [THEN [2] rev_subsetD])

subsection{*Inductive relation "synth"*}

text{*Inductive definition of "synth" -- what can be built up from a set of
    messages.  A form of upward closure.  Pairs can be built, messages
    encrypted with known keys.  Agent names are public domain.
    Numbers can be guessed, but Nonces cannot be.  *}

inductive_set
  synth :: "msg set => msg set"
  for H :: "msg set"
  where
    Inj    [intro]:   "X \<in> H ==> X \<in> synth H"
  | Agent  [intro]:   "Agent agt \<in> synth H"
  | Number [intro]:   "Number n  \<in> synth H"
  | Real   [intro]:   "Real n  \<in> synth H"
  | Hash   [intro]:   "X \<in> synth H ==> Hash X \<in> synth H"
  | MPair  [intro]:   "[|X \<in> synth H;  Y \<in> synth H|] ==> \<lbrace>X,Y\<rbrace> \<in> synth H"
  | Crypt  [intro]:   "[|X \<in> synth H;  Key(K) \<in> H|] ==> Crypt K X \<in> synth H"

text{*Monotonicity*}
lemma synth_mono: "G\<subseteq>H ==> synth(G) \<subseteq> synth(H)"
  by (auto, erule synth.induct, auto)

text{*NO @{text Agent_synth}, as any Agent name can be synthesized.  
  The same holds for @{term Number}*}
inductive_cases Nonce_synth [elim!]: "Nonce B n \<in> synth H"
inductive_cases Key_synth   [elim!]: "Key K \<in> synth H"
inductive_cases Hash_synth  [elim!]: "Hash X \<in> synth H"
inductive_cases MPair_synth [elim!]: "\<lbrace>X,Y\<rbrace> \<in> synth H"
inductive_cases Crypt_synth [elim!]: "Crypt K X \<in> synth H"
inductive_cases Number_synth [elim!]: "Number n \<in> synth H"
inductive_cases Real_synth [elim!]: "Real t \<in> synth H"

lemma synth_increasing: "H \<subseteq> synth(H)"
by blast

subsubsection{*Unions *}

text{*Converse fails: we can synth more from the union than from the 
  separate parts, building a compound message using elements of each.*}
lemma synth_Un: "synth(G) \<union> synth(H) \<subseteq> synth(G \<union> H)"
by (intro Un_least synth_mono Un_upper1 Un_upper2)

lemma synth_insert: "insert X (synth H) \<subseteq> synth(insert X H)"
by (blast intro: synth_mono [THEN [2] rev_subsetD])

subsubsection{*Idempotence and transitivity *}

lemma synth_synthD [dest!]: "X\<in> synth (synth H) ==> X\<in> synth H"
by (erule synth.induct, blast+)

lemma synth_idem: "synth (synth H) = synth H"
by blast

lemma synth_subset_iff [simp]: "(synth G \<subseteq> synth H) = (G \<subseteq> synth H)"
apply (rule iffI)
apply (iprover intro: subset_trans synth_increasing)  
apply (frule synth_mono, simp add: synth_idem) 
done

lemma synth_trans: "[| X\<in> synth G;  G \<subseteq> synth H |] ==> X\<in> synth H"
by (drule synth_mono, blast)

text{*Cut; Lemma 2 of Lowe*}
lemma synth_cut: "[| Y\<in> synth (insert X H);  X\<in> synth H |] ==> Y\<in> synth H"
by (erule synth_trans, blast)

lemma Agent_synth [simp]: "Agent A \<in> synth H"
by blast

lemma Nonce_synth_eq [simp]: "(Nonce A n \<in> synth H) = (Nonce A n \<in> H)"
by blast

lemma Key_synth_eq [simp]: "(Key K \<in> synth H) = (Key K \<in> H)"
by blast

lemma Crypt_synth_eq [simp]:
     "Key K \<notin> H ==> (Crypt K X \<in> synth H) = (Crypt K X \<in> H)"
by blast


lemma keysFor_synth [simp]: 
    "keysFor (synth H) = keysFor H \<union> invKey`{K. Key K \<in> H}"
by (unfold keysFor_def, blast)


subsubsection{*Combinations of parts, analz and synth *}

lemma parts_synth [simp]: "parts (synth H) = parts H \<union> synth H"
apply (rule equalityI)
apply (rule subsetI)
apply (erule parts.induct)
apply (blast intro: synth_increasing [THEN parts.mono, THEN subsetD] 
                    parts.Fst parts.Snd parts.Body)+
done

lemma subterms_synth [simp]: "subterms (synth H) = subterms H \<union> synth H"
apply (rule equalityI)
apply (rule subsetI)
apply (erule subterms.induct)
apply force
apply (blast intro: synth_increasing [THEN subterms.mono, THEN subsetD]
                    subterms.Fst subterms.Snd subterms.Body subterms.Hash)+
done

lemma analz_analz_Un [simp]: "analz (analz G \<union> H) = analz (G \<union> H)"
apply (intro equalityI analz_subset_cong)+
apply simp_all
done

lemma analz_synth_Un [simp]: "analz (synth G \<union> H) = analz (G \<union> H) \<union> synth G"
apply (rule equalityI)
apply (rule subsetI)
apply (erule analz.induct)
prefer 5 apply (blast intro: analz_mono [THEN [2] rev_subsetD])
apply (blast intro: analz.Fst analz.Snd analz.Decrypt)+
done

lemma analz_synth [simp]: "analz (synth H) = analz H \<union> synth H"
apply (cut_tac H = "{}" in analz_synth_Un)
apply (simp (no_asm_use))
done

lemma parts_analz_parts_knowsI:
  assumes A: "X \<in> analz (knowsI A tr)"
  shows   "\<exists> Y \<in> knowsI A tr. X \<in> parts {Y}" using A
proof -
  from A have "X \<in> parts (knowsI A tr)"
    apply - apply (rule analz_into_parts) .
  hence "X \<in> parts (knowsI A tr)" by auto
  hence "\<exists> Y \<in> knowsI A tr. X \<in> parts {Y}" by (auto intro: parts.singleton)
  thus ?thesis by auto
qed

lemma subset_analz_parts_knowsI:
  assumes A: "X \<in> analz (knowsI A tr)"
  shows   "\<exists> Y \<in> knowsI A tr. X \<in> subterms {Y}" using A
apply -
apply (drule parts_analz_parts_knowsI)
apply (auto intro: parts_in_subterms)
done

subsubsection{*For reasoning about the Fake rule in traces *}

lemma parts_insert_subset_Un: "X\<in> G ==> parts(insert X H) \<subseteq> parts G \<union> parts H"
by (rule subset_trans [OF parts.mono parts.Un_subset2], blast)

lemma subterms_insert_subset_Un: "X\<in> G ==> subterms(insert X H) \<subseteq> subterms G \<union> subterms H"
by (rule subset_trans [OF subterms.mono subterms.Un_subset2], blast)

text{*More specifically for Fake.  Very occasionally we could do with a version
  of the form  @{term"parts{X} \<subseteq> synth (analz H) \<union> parts H"} *}
lemma Fake_parts_insert:
     "X \<in> synth (analz H) ==>  
      parts (insert X H) \<subseteq> synth (analz H) \<union> parts H"
apply (drule parts_insert_subset_Un)
apply (simp (no_asm_use))
apply blast
done

lemma Fake_parts_insert_in_Un:
     "[|Z \<in> parts (insert X H);  X: synth (analz H)|] 
      ==> Z \<in>  synth (analz H) \<union> parts H";
by (blast dest: Fake_parts_insert  [THEN subsetD, dest])

text{*@{term H} is sometimes @{term"Key ` KK \<union> spies evs"}, so can't put 
      @{term "G=H"}.*}
lemma Fake_analz_insert:
     "X\<in> synth (analz G) ==>  
      analz (insert X H) \<subseteq> synth (analz G) \<union> analz (G \<union> H)"
apply (rule subsetI)
apply (subgoal_tac "x \<in> analz (synth (analz G) \<union> H) ")
prefer 2 apply (blast intro: analz_mono [THEN [2] rev_subsetD] analz_mono [THEN synth_mono, THEN [2] rev_subsetD])
apply (simp (no_asm_use))
apply blast
done

lemma analz_conj_parts [simp]:
     "(X \<in> analz H & X \<in> parts H) = (X \<in> analz H)"
by (blast intro: analz_subset_parts [THEN subsetD])

lemma analz_disj_parts [simp]:
     "(X \<in> analz H | X \<in> parts H) = (X \<in> parts H)"
by (blast intro: analz_subset_parts [THEN subsetD])

lemma analz_conj_subterms [simp]:
     "(X \<in> analz H & X \<in> subterms H) = (X \<in> analz H)"
by (blast intro: analz_subset_subterms [THEN subsetD])

lemma analz_disj_subterms [simp]:
     "(X \<in> analz H | X \<in> subterms H) = (X \<in> subterms H)"
by (blast intro: analz_subset_subterms [THEN subsetD])

text{*Without this equation, other rules for synth and analz would yield
  redundant cases*}
lemma MPair_synth_analz [iff]:
     "(\<lbrace>X,Y\<rbrace> \<in> synth (analz H)) =  
      (X \<in> synth (analz H) & Y \<in> synth (analz H))"
by blast

lemma Crypt_synth_analz:
     "[| Key K \<in> analz H;  Key (invKey K) \<in> analz H |]  
       ==> (Crypt K X \<in> synth (analz H)) = (X \<in> synth (analz H))"
by blast

lemma Hash_synth_analz [simp]:
     "X \<notin> synth (analz H)  
      ==> (Hash\<lbrace>X,Y\<rbrace> \<in> synth (analz H)) = (Hash\<lbrace>X,Y\<rbrace> \<in> analz H)"
by blast

subsection{*MACM: a combination of Hash and MPair*}

subsubsection{*Freeness *}

lemma Agent_neq_MACM: "Agent A ~= Hash[X] Y"
by (unfold MACM_def, simp)

lemma Nonce_neq_MACM: "Nonce A N ~= Hash[X] Y"
by (unfold MACM_def, simp)

lemma Number_neq_MACM: "Number N ~= Hash[X] Y"
by (unfold MACM_def, simp)

lemma Key_neq_MACM: "Key K ~= Hash[X] Y"
by (unfold MACM_def, simp)

lemma Hash_neq_MACM: "Hash Z ~= Hash[X] Y"
by (unfold MACM_def, simp)

lemma Crypt_neq_MACM: "Crypt K X' ~= Hash[X] Y"
by (unfold MACM_def, simp)

lemmas MACM_neqs = Agent_neq_MACM Nonce_neq_MACM Number_neq_MACM 
                    Key_neq_MACM Hash_neq_MACM Crypt_neq_MACM

declare MACM_neqs [iff]
declare MACM_neqs [symmetric, iff]

lemma MACM_eq [iff]: "(Hash[X'] Y' = Hash[X] Y) = (X' = X & Y'=Y)"
by (simp add: MACM_def)

lemma MPair_eq_MACM [iff]:
     "(\<lbrace>X',Y'\<rbrace> = Hash[X] Y) = (X' = Hash\<lbrace>X,Y\<rbrace> & Y'=Y)"
by (simp add: MACM_def)

lemma MACM_eq_MPair [iff]:
     "(Hash[X] Y = \<lbrace>X',Y'\<rbrace>) = (X' = Hash\<lbrace>X,Y\<rbrace> & Y'=Y)"
by (auto simp add: MACM_def)


subsubsection{*Specialized laws, proved in terms of those for Hash and MPair *}

lemma keysFor_insert_MACM [simp]: "keysFor (insert (Hash[X] Y) H) = keysFor H"
by (simp add: MACM_def)

lemma parts_insert_MACM [simp]: 
    "parts (insert (Hash[X] Y) H) =  
     insert (Hash[X] Y) (insert (Hash\<lbrace>X,Y\<rbrace>) (parts (insert Y H)))"
by (simp add: MACM_def)

lemma subterms_insert_MACM [simp]: 
    "subterms (insert (Hash[X] Y) H) =  
     insert (Hash[X] Y) (insert (Hash\<lbrace>X,Y\<rbrace>) (insert \<lbrace>X,Y\<rbrace> (subterms (insert X (insert Y H)))))"
by (simp add: MACM_def)

lemma analz_insert_MACM [simp]: 
    "analz (insert (Hash[X] Y) H) =  
     insert (Hash[X] Y) (insert (Hash\<lbrace>X,Y\<rbrace>) (analz (insert Y H)))"
by (simp add: MACM_def)

lemma MACM_synth_analz [simp]:
     "X \<notin> synth (analz H)  
    ==> (Hash[X] Y \<in> synth (analz H)) =  
        (Hash \<lbrace>X, Y\<rbrace> \<in> analz H & Y \<in> synth (analz H))"
by (simp add: MACM_def)

(*
text{*We do NOT want Crypt... messages broken up in protocols!!*}
declare parts.Body [rule del]
*)

lemma Crypt_notin_image_Key [simp]: "Crypt K X \<notin> Key ` A"
by auto

lemma Hash_notin_image_Key [simp] :"Hash X \<notin> Key ` A"
by auto

lemma synth_analz_mono: "G\<subseteq>H ==> synth (analz(G)) \<subseteq> synth (analz(H))"
by (iprover intro: synth_mono analz_mono)

lemma Fake_analz_eq [simp]:
     "X \<in> synth(analz H) ==> synth (analz (insert X H)) = synth (analz H)"
apply (drule Fake_analz_insert[of _ _ "H"])
apply (simp add: synth_increasing[THEN Un_absorb2])
apply (drule synth_mono)
apply (simp add: synth_idem)
apply (rule equalityI)
apply (simp add: );
apply (rule synth_analz_mono, blast)   
done

text{*Two generalizations of @{text analz_insert_eq}*}
lemma gen_analz_insert_eq [rule_format]:
     "X \<in> analz H ==> ALL G. H \<subseteq> G --> analz (insert X G) = analz G";
by (blast intro: analz_cut analz_insertI analz_mono [THEN [2] rev_subsetD])

lemma synth_analz_insert_eq [rule_format]:
     "X \<in> synth (analz H) 
      ==> ALL G. H \<subseteq> G --> (Key K \<in> analz (insert X G)) = (Key K \<in> analz G)";
apply (erule synth.induct) 
apply (simp_all add: gen_analz_insert_eq subset_trans [OF _ subset_insertI]) 
done

lemma Fake_parts_sing:
     "X \<in> synth (analz H) ==> parts{X} \<subseteq> synth (analz H) \<union> parts H";
apply (rule subset_trans) 
 apply (erule_tac [2] Fake_parts_insert)
apply (rule parts.mono, blast)
done

lemmas Fake_parts_sing_imp_Un = Fake_parts_sing [THEN [2] rev_subsetD]

text{*Relations between DM, analz and synth.*}

lemma DM_analz:
  "analz H \<subseteq> DM A H"
  apply auto
  apply (erule analz.induct)
  apply (auto intro: DM.Fst DM.Snd DM.Decrypt)
done

lemma DM_synth:
  "synth H \<subseteq> DM A H"
  apply auto
  apply (erule synth.induct)
  apply (auto intro: DM.Fst DM.Snd DM.Decrypt)
done

lemma DM_analz_synth:
  "DM A H = synth (range (Nonce A) \<union> analz H)"
proof auto
  fix A :: agent and H :: "msg set" and M :: msg
  assume "M \<in> DM A H"
  thus "M \<in> synth (range (Nonce A) \<union> analz H)"
    apply -
    apply (erule DM.induct)
    apply auto
    done
next
  fix A :: agent and H :: "msg set" and M :: msg
  assume "M \<in> synth (range (Nonce A) \<union> analz H)"
  thus "M \<in> DM A H" using DM_analz
    apply -
    apply (erule synth.induct)
    apply auto
    done
qed


text{*For proving @{text new_keys_not_used}*}
lemma keysFor_parts_insert:
     "[| K \<in> keysFor (parts (insert X G));  X \<in> synth (analz H) |] 
      ==> K \<in> keysFor (parts (G \<union> H)) | Key (invKey K) \<in> parts H"; 
by (force 
    dest!: parts_insert_subset_Un [THEN keysFor_mono, THEN [2] rev_subsetD]
           analz_subset_parts [THEN keysFor_mono, THEN [2] rev_subsetD]
    intro: analz_subset_parts [THEN subsetD] parts.mono [THEN [2] rev_subsetD])

lemma nonce_DM_nonce_subterms:
  assumes nonce_parts: "Nonce B NB \<in> subterms (DM A H)"
  and AnotB:  "A \<noteq> B"
  shows  "Nonce B NB \<in> subterms H"
proof -
  have "Nonce B NB \<in> subterms (synth (range (Nonce A) \<union> analz H))"
    using nonce_parts apply -
    by (simp add: DM_analz_synth)
  hence "Nonce B NB \<in> subterms (range (Nonce A) \<union> analz H)"
    apply -
    by (auto dest: not_subterms_not_analz)
  thus ?thesis using AnotB
    apply auto
    apply (drule subterms.singleton)
    apply auto
    done
qed

lemma nonce_DM_nonce_parts:
  assumes nonce_parts: "Nonce B NB \<in> parts (DM A H)"
  and AnotB:  "A \<noteq> B"
  shows  "Nonce B NB \<in> parts H"
proof -
  have "Nonce B NB \<in> parts (synth (range (Nonce A) \<union> analz H))"
    using nonce_parts apply -
    by (simp add: DM_analz_synth)
  hence "Nonce B NB \<in> parts (range (Nonce A) \<union> analz H)"
    apply -
    by (auto dest: not_parts_not_analz)
  thus ?thesis using AnotB
    apply auto
    apply (drule parts.singleton)
    apply auto
    done
qed

lemma key_DM_key_parts:
  assumes key_parts: "Key k \<in> parts (DM A H)"
  shows              "Key k \<in> parts H"
proof -
  have "Key k \<in> parts (synth (range (Nonce A) \<union> analz H))"
    using prems by (simp add: DM_analz_synth)
  hence "Key k \<in> parts (range (Nonce A) \<union> analz H)"
    apply -
    by (auto dest: not_parts_not_analz)
  thus ?thesis
    apply auto
    apply (drule parts.singleton)
    apply auto
    done
qed

lemma sig_DM_key_or_sig:
  assumes sig_parts: "Crypt k msig \<in> subterms (DM B H)"
  shows "Crypt k msig \<in> subterms H
         \<or> Key k \<in> parts H"
proof -
  have "Crypt k msig \<in> subterms (synth (range (Nonce B) \<union> analz H))" using prems
    by (simp add: DM_analz_synth)
  hence or: "Crypt k msig \<in> subterms (range (Nonce B) \<union> analz H)
             \<or> Crypt k msig \<in> synth (range (Nonce B) \<union> analz H)"
    apply -
    by auto
  show ?thesis
  proof cases
    assume "Crypt k msig \<in> subterms (range (Nonce B) \<union> analz H)"
    hence "Crypt k msig \<in> subterms H"
      apply auto
      apply (drule subterms.singleton)
      apply force
      done
    thus ?thesis apply auto done
  next
    assume not: "Crypt k msig \<notin> subterms (range (Nonce B) \<union> analz H)"
    hence "Crypt k msig \<in> synth (range (Nonce B) \<union> analz H)"
      using or by auto
    hence "Key k \<in> synth (range (Nonce B) \<union> analz H)"
      using not
      apply auto
      by (auto dest: not_subterms_not_analz)
    hence "Key k \<in> parts H"
      apply auto
      apply (erule rev_subsetD)
      apply (rule analz_subset_parts)
      done
    thus ?thesis by force
  qed
qed

lemma mac_DM_mac_or_key:
  assumes mac_parts: "Hash \<lbrace>Key k, m\<rbrace> \<in> subterms (DM B H)"
  shows "Hash \<lbrace>Key k, m\<rbrace> \<in> subterms H
         \<or> Key k \<in> parts H"
proof -
  have "Hash \<lbrace>Key k, m\<rbrace> \<in> subterms (synth (range (Nonce B) \<union> analz H))" using prems
    by (simp add: DM_analz_synth)
  hence or: "Hash \<lbrace>Key k, m\<rbrace> \<in> subterms (range (Nonce B) \<union> analz H)
             \<or> Hash \<lbrace>Key k, m\<rbrace> \<in> synth (range (Nonce B) \<union> analz H)"
    apply -
    by auto
  show ?thesis
  proof cases
    assume "Hash \<lbrace>Key k, m\<rbrace> \<in> subterms (range (Nonce B) \<union> analz H)"
    hence "Hash \<lbrace>Key k, m\<rbrace> \<in> subterms H"
      apply auto
      apply (drule subterms.singleton)
      apply force
      done
    thus ?thesis apply auto done
  next
    assume not: "Hash \<lbrace>Key k, m\<rbrace> \<notin> subterms (range (Nonce B) \<union> analz H)"
    hence "Hash \<lbrace>Key k, m\<rbrace> \<in> synth (range (Nonce B) \<union> analz H)"
      using or by auto
    hence "Key k \<in> synth (range (Nonce B) \<union> analz H)"
      using not
      apply auto
      by (auto dest: not_subterms_not_analz)
    hence "Key k \<in> parts H"
      apply auto
      apply (erule rev_subsetD)
      apply (rule analz_subset_parts)
      done
    thus ?thesis by force
  qed
qed

interpretation MESSAGE_DERIVATION Crypt Nonce MPair Hash Number parts subterms DM Key
  apply (unfold_locales)
  apply (erule nonce_DM_nonce_subterms, force)
  apply (erule nonce_DM_nonce_parts, force)
  apply (erule key_DM_key_parts)
  apply (erule sig_DM_key_or_sig)
  apply (erule mac_DM_mac_or_key)
done

end